/* Copyright (C) 2011 Sandia National Laboratories, Livermore CA.
 *
 * This program is part of the SST/Macroscale package at
 * https://bitbucket.org/cljanss/sstmacro.  Questions should be
 * directed to contact info found at that site for SST/Macroscale
 * related issues. Other issues related to this code may be directed
 * to the author, contact info below.
 * 
 * Author: David Evensky <evensky@sandia.gov> or <evensky@gmail.com>
 * /

/*
This program is available under BSD licensing.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

(1) Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

(2) Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

(3) Neither the name of The Ohio State University nor the names of
their contributors may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* Revision history:
 * 09/2011: evensky: Initial checkin.
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "mpi.h"
#define N_BARR   2
#ifndef HOST_NAME_MAX
#define HOST_NAME_MAX 255
#endif

#define max(a,b) ((a)>(b)?(a):(b))
#define min(a,b) ((a)<(b)?(a):(b))

static int ierr=0;
static int debug = 0;

//#define OSU_WORKS /* it doesn't yet. TODO. */

// TYPES
typedef enum {CHAR_RUN, TYPE_RUN,
#ifdef OSU_WORKS
	      OSU_RUN,
#endif
	      UNKNOWN_RUN} ping_pong_type;
typedef struct {int size; int reps; } schedule_type;


// DCL functions
void output(int size, int reps, int nsamp, double *time);
double ping_pong_types(int myid, int mypartner,
		 void *s_buf,void *r_buf,int size, int n_sample,
		 MPI_Datatype s_type,MPI_Datatype r_type,
		 MPI_Comm comm);
double ping_pong_char(int myid, int mypartner,
		 void *s_buf,void *r_buf,int size, int n_sample,
		 MPI_Datatype s_type,MPI_Datatype r_type,
		 MPI_Comm comm);
#ifdef OSU_WORKS
double ping_pong_osu(int myid, int mypartner,
		 void *s_buf,void *r_buf,int size, int n_sample,
		 MPI_Datatype s_type,MPI_Datatype r_type,
		 MPI_Comm comm);
#endif
void unexpected_send(void *buf,int unexpected_size, int nunexpected,
		     MPI_Request **unexpected_request,
		     int myid, int mypartner,MPI_Comm comm);
void unexpected_recv(void *buf,int unexpected_size, int nunexpected,
		     MPI_Request **unexpected_request,
		     MPI_Status **unexpected_status,
		     int myid, int mypartner,MPI_Comm comm);
void print_header(char *argv0,int myid, int numprocs, int mypartner,
		  int nunexpected);
void usage(char *s);
void terminate(int rtn);
schedule_type *set_schedule();




// schedule taken to match the pingpong test from Intel MPI Benchmark v3.2.2
schedule_type *set_schedule() {
  static schedule_type sched[] =
    {{   0,1000},  {  1,1000},  {  2,1000},  {  4,1000}, {8,1000},
     {  16,1000}, {  32,1000}, {  64,1000}, { 128,1000}, {256,1000},
     {512,1000},  {1024,1000}, {2048,1000}, {4096,1000}, {8192,1000},
     {  16384,1000},  {32768,1000}, { 65536,640},
     { 131072,320}, { 262144,160},  {524288,80},
     {1048576,40},  {2097152,20},  {4194304,10}, {-1,-1}};
  return sched;
}

void usage(char *s) {
  printf("Usage: %s {-u <number of unexpected messages>} {-p } {-d} {-h} {-v}\n",
	 (s?s:"argv[0]"),
#ifdef OSU_WORKS
	 "(type|char|osu)"
#else
	 "(type|char)"
#endif
	 );
}


int main(int argc,char **argv) {
  int myid = 0;
  int numprocs = 0;
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);

  schedule_type *schedule = set_schedule();
 
  int mypartner,isched;
  int verbose=0;
  int nunexpected=0;
  int unexpected_size = 100;
  ping_pong_type which_run = TYPE_RUN;
  while(1) {
    int c = getopt(argc,argv,"hvdu:U:p:");
    if (c < 0) break;
    switch(c) {
    case 'h':
      if (myid == 0)
	usage(argv[0]);
      terminate(0);
      break;
    case 'v':
      ++verbose;
      break;
    case 'd':
      ++debug;
      break;
    case 'u':
      nunexpected = atoi(optarg);
      break;
    case 'U':
      unexpected_size = atoi(optarg);
      break;
    case 'p':
      if (!strncmp("type",optarg,4))
	which_run = TYPE_RUN;
      else if (!strncmp("char",optarg,4))
	which_run = CHAR_RUN;
#ifdef OSU_WORKS
      else if (!strncmp("osu",optarg,3))
	which_run = OSU_RUN;
#endif
      else
	which_run = UNKNOWN_RUN;
      break;
    default:
      if (myid == 0)
	usage(argv[0]);
      terminate(0);
      break;
    }
  }



  //dest = myid ^ 1;?

  if (which_run == UNKNOWN_RUN) {
    usage(argv[0]);
    terminate(0);
  }

  if (numprocs % 2 != 0) {
    if (!myid)
      printf("%s: needs an even number of processes\n",argv[0]);
    terminate(-1);
  }

  if (myid % 2 == 0)
    mypartner = myid+1;
  else
    mypartner = myid-1;

  if (verbose)
    print_header(argv[0],myid,numprocs,mypartner,nunexpected);

  void *s_buf = malloc(4194304);
  void *r_buf = malloc(4194304);
  memset(s_buf,0x0d,4194304);
  memset(r_buf,0xae,4194304);

  double (*pp_fn)(int,int,void *,void *,int, int,
		  MPI_Datatype ,MPI_Datatype ,
		  MPI_Comm );
  if (which_run == TYPE_RUN) {
    pp_fn = ping_pong_types;
  } else if (which_run == CHAR_RUN) {
    pp_fn = ping_pong_char;
#ifdef OSU_WORKS
  } else if (which_run == OSU_RUN) {
    pp_fn = ping_pong_osu;
#endif
  }

  MPI_Datatype s_type,r_type;
  MPI_Request *unexpected_request;
  MPI_Status *unexpected_status;


  for (isched = 0; schedule[isched].size != -1; isched++) {
    int size = schedule[isched].size;
    int reps = schedule[isched].reps;

    if (debug && myid == 0)
      printf("Running %d: size=%d\n",isched,size);

    if (nunexpected)
      unexpected_send(s_buf,unexpected_size,nunexpected,&unexpected_request,
		      myid,mypartner,MPI_COMM_WORLD);


    if (which_run == TYPE_RUN) {
      if (debug && myid == 0)
	printf("Creating types\n");
      MPI_Type_contiguous(size, MPI_CHAR, &s_type);
      MPI_Type_commit(&s_type);
      MPI_Type_contiguous(size, MPI_CHAR, &r_type);
      MPI_Type_commit(&r_type);

      int tmp;
      MPI_Type_size(s_type,&tmp);
      if (tmp != size) {
	printf("s_type size is wrong, expected %d got %d\n",
	       size,tmp);
	terminate(0);
      }
      MPI_Type_size(r_type,&tmp);
      if (tmp != size) {
	printf("r_type size is wrong, expected %d got %d\n",
	       size,tmp);
	terminate(0);
      }
    }
    if (debug && myid == 0)
      printf("Running short warm up\n");
    pp_fn(myid,mypartner,s_buf,r_buf,size,1,
	      s_type,r_type,MPI_COMM_WORLD);
    if (debug && myid == 0)
      printf("Doing measurement\n");
    MPI_Barrier(MPI_COMM_WORLD);
    int nsamp = 50;
    int isamp;
    double *time = (double *)malloc(nsamp*sizeof(double));
    for (isamp = 0; isamp < nsamp; isamp++) {
      memset(s_buf,0xda,4194304);
      memset(r_buf,0xda,4194304);

      time[isamp] = pp_fn(myid,mypartner,s_buf,r_buf,size,reps,
			      s_type,r_type,MPI_COMM_WORLD);
      MPI_Barrier(MPI_COMM_WORLD);
    }
    if (myid == 0)
      output(size,reps,nsamp,time);
    if (which_run == TYPE_RUN) {
      MPI_Type_free(&s_type);
      MPI_Type_free(&r_type);
    }
    if (nunexpected)
      unexpected_recv(r_buf,unexpected_size,nunexpected,&unexpected_request,
		      &unexpected_status,myid,mypartner,MPI_COMM_WORLD);
  }

  terminate(0);
}

void output(int size, int reps, int nsamp, double *time) {
  double tmax, tmin, tsum,t2sum,tave,tstd;
  static int first_time = 1;

  if (first_time) {
    printf("#times are in us\n");
    printf("#%7s %4s %10s %9s %9s %9s +/- %6s %9s +/- %6s\n",
	   "size","rep","rep*size","t_min","t_max","t_ave","t_std",
	   "bw[MB/s]","bw_std");
    first_time = 0;
  }
  tmax = -1e18;
  tmin = 1e18;
  tsum = 0.0;
  t2sum = 0.0;
  int i;
  for (i=0; i < nsamp; i++) // round trip to one way, and usec.
    time[i] = time[i]/2.0*1e6;
  for (i=0; i < nsamp; i++) {
    tmax = max(tmax,time[i]);
    tmin = min(tmin,time[i]);
    tsum += time[i];
    t2sum += time[i]*time[i];
  }
  tave = tsum/(double)nsamp;
  tstd = sqrt(t2sum/(double)nsamp - tave*tave);
  double bw = size/tave;
  double bwstd = bw*(tstd/tave); // propagates as relative error.
  printf("%8d %4d %10d %9.4f %9.4f %9.4f +/- %6.4f %9.4f +/- %6.4f\n",
	 size,reps, size*reps,
	 tmin,tmax,tave,tstd, bw,bwstd);
}

double ping_pong_types(int myid, int mypartner,
		 void *s_buf,void *r_buf,int size, int n_sample,
		 MPI_Datatype s_type,MPI_Datatype r_type,
		 MPI_Comm comm) {
  int s_tag = 1;
  int r_tag = 2;
  int s_num = 1;
  int r_num = 1;
  double time,t1,t2;
  int ierr,i;
  MPI_Status stat;

#if 0
  if (size == 0) { // cheating a bit here. might be specially optimized.
    s_num = 0;
    r_num = 0;
  }
#endif
  if (debug)
    printf("ping_pong_types(%d,%d,%p,%p,%d,%d,%d,%d,%d)\n",
	   myid,mypartner,s_buf,r_buf,size,n_sample,s_type,
	   r_type,comm);

  if (myid % 2 == 0) { // EVEN
    for(i=0; i<N_BARR; i++) MPI_Barrier(comm);
    t1 = MPI_Wtime();
    for(i=0;i<n_sample;i++) {
      ierr = MPI_Send(s_buf,s_num,s_type,mypartner,s_tag,comm);
      ierr = MPI_Recv(r_buf,r_num,r_type,mypartner,r_tag,comm,&stat);
    }
    t2 = MPI_Wtime();
  } else { // ODD
    for(i=0; i<N_BARR; i++) MPI_Barrier(comm);
    t1 = MPI_Wtime();
    for(i=0;i<n_sample;i++) {
      ierr = MPI_Recv(s_buf,s_num,s_type,mypartner,s_tag,comm,&stat);
      ierr = MPI_Send(r_buf,r_num,r_type,mypartner,r_tag,comm);
    }
    t2 = MPI_Wtime();
  }

  time=(t2 - t1)/n_sample;
  return(time);
}

double ping_pong_char(int myid, int mypartner,
		      void *s_buf,void *r_buf,int size, int n_sample,
		      MPI_Datatype s_type,MPI_Datatype r_type,
		      MPI_Comm comm) {
  int s_tag = 1;
  int r_tag = 2;
  int s_num = size;
  int r_num = size;
  double time,t1,t2;
  int ierr,i;
  MPI_Status stat;

  if (debug)
    printf("ping_pong_char(%d,%d,%p,%p,%d,%d,%d,%d,%d)\n",
	   myid,mypartner,s_buf,r_buf,size,n_sample,s_type,
	   r_type,comm);

  if (myid % 2 == 0) { // EVEN
    for(i=0; i<N_BARR; i++) MPI_Barrier(comm);
    t1 = MPI_Wtime();
    for(i=0;i<n_sample;i++) {
      ierr = MPI_Send(s_buf,s_num,MPI_CHAR,mypartner,s_tag,comm);
      ierr = MPI_Recv(r_buf,r_num,MPI_CHAR,mypartner,r_tag,comm,&stat);
    }
    t2 = MPI_Wtime();
  } else { // ODD
    for(i=0; i<N_BARR; i++) MPI_Barrier(comm);
    t1 = MPI_Wtime();
    for(i=0;i<n_sample;i++) {
      ierr = MPI_Recv(s_buf,s_num,MPI_CHAR,mypartner,s_tag,comm,&stat);
      ierr = MPI_Send(r_buf,r_num,MPI_CHAR,mypartner,r_tag,comm);
    }
    t2 = MPI_Wtime();
  }

  time=(t2 - t1)/n_sample;
  return(time);
}




void unexpected_send(void *buf,int unexpected_size, int nunexpected,
		     MPI_Request **unexpected_request,
		     int myid, int mypartner,
		     MPI_Comm comm) {
  int i,j;
  int unexpected_tag = 500;

  if (nunexpected > 0) {
    *unexpected_request =
      (MPI_Request *) malloc(sizeof(MPI_Request)*nunexpected);
    if (myid % 2 == 0) { // even
      for (j=0; j < nunexpected; j++) {
	MPI_Isend(buf, unexpected_size, MPI_CHAR, mypartner,
		  unexpected_tag, comm,
		  (*unexpected_request)+j);
      }
    } // odd does nothing here, its unexpected.
  }
  //MPI_Barrier(comm);
}
void unexpected_recv(void *buf,int unexpected_size, int nunexpected,
		     MPI_Request **unexpected_request,
		     MPI_Status **unexpected_status,
		     int myid, int mypartner,MPI_Comm comm) {
  int i,j;
  int unexpected_tag = 500;

  if (nunexpected > 0) {
    *unexpected_status = (MPI_Status *)malloc(sizeof(MPI_Status)*nunexpected);
    if (myid % 2 != 0) { // ODD
      for (j=0; j < nunexpected; j++) {
	MPI_Irecv(buf, unexpected_size, MPI_CHAR, mypartner, unexpected_tag,
		  comm,(*unexpected_request)+j);
      }
    }
    MPI_Waitall(nunexpected, *unexpected_request, *unexpected_status);
  }
}

void print_header(char *argv0,int myid, int numprocs, int mypartner,
		  int nunexpected) {
  char hname[HOST_NAME_MAX+1];
  gethostname(hname,HOST_NAME_MAX);
  printf("#%s: rank %d (%s) of %d. Partner is %d\n",hname,myid,
	 (myid % 2 == 0 ? "even" : "odd"),
	 numprocs,mypartner);
  if (myid == 0) {
    printf("#Number of unexpected messages precharged %d\n",nunexpected);
  }
}
void terminate(int rtn) {
  MPI_Finalize();
  exit(rtn);
}
#ifdef OSU_WORKS
double ping_pong_osu(int myid, int mypartner,
		      void *s_buf,void *r_buf,int size, int n_sample,
		      MPI_Datatype s_type,MPI_Datatype r_type,
		      MPI_Comm comm) {
#define MAX_REQ_NUM 1000
#if 0

#define MAX_ALIGNMENT 65536
#define MAX_MSG_SIZE (1<<22)
#define MYBUFSIZE (MAX_MSG_SIZE + MAX_ALIGNMENT)


  char s_buf1[MYBUFSIZE];
  char r_buf1[MYBUFSIZE];
  align_size = getpagesize();
  assert(align_size <= MAX_ALIGNMENT);
  char *s_buf, *r_buf;
  s_buf =
    (char *) (((unsigned long) s_buf1 + (align_size - 1)) /
	      align_size * align_size);
  r_buf =
    (char *) (((unsigned long) r_buf1 + (align_size - 1)) /
	      align_size * align_size);
  /* touch the data */
  for(i = 0; i < size; i++) {
    s_buf[i] = 'a';
    r_buf[i] = 'b';
  }
#endif


  int s_tag = 1;
  int r_tag = 2;
  int s_num = size;
  int r_num = size;
  double time,t1,t2;
  int ierr,i,j;
  MPI_Status stat;
  MPI_Request req;

  if (debug)
    printf("ping_pong_osu(%d,%d,%p,%p,%d,%d,%d,%d,%d)\n",
	   myid,mypartner,s_buf,r_buf,size,n_sample,s_type,
	   r_type,comm);

  MPI_Request request[MAX_REQ_NUM];
  MPI_Status  reqstat[MAX_REQ_NUM];

  int loop = 100;
  int window_size = 64;
  int skip = 10;
  int loop_large = 20;
  int window_size_large = 64;
  int skip_large = 2;
  int large_message_size = 8192;
  int align_size;

  if(size > large_message_size) {
    loop = loop_large;
    skip = skip_large;
    window_size = window_size_large;
  }
  if(myid % 2 == 0) { // EVEN
    for(i = 0; i < loop + skip; i++) {
      if(i == skip) {
	t1 = MPI_Wtime();
      }
      for(j = 0; j < window_size; j++) {
	MPI_Isend(s_buf, size, MPI_CHAR, 1, 100, MPI_COMM_WORLD,
		  request + j);
      }
      MPI_Waitall(window_size, request, reqstat);
      MPI_Recv(r_buf, 4, MPI_CHAR, 1, 101, MPI_COMM_WORLD,
	       &reqstat[0]);
    }
    t2 = MPI_Wtime();
  }  else { // ODD
    for(i = 0; i < loop + skip; i++) {
      for(j = 0; j < window_size; j++) {
	MPI_Irecv(r_buf, size, MPI_CHAR, 0, 100, MPI_COMM_WORLD,
		  request + j);
      }
      MPI_Waitall(window_size, request, reqstat);
      MPI_Send(s_buf, 4, MPI_CHAR, 0, 101, MPI_COMM_WORLD);
    }
  }
  time=(t2 - t1)/(loop*window_size);
  return(time);

}
#endif
