/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <gtc/gtc_app.h>
#include <gtc/gtc_particle.h>
#include <gtc/gtc_point.h>
#include <gtc/gtc_particle_stubs.h>
#include <gtc/gtc_point_stubs.h>

#include <hpx/components/hpx_distributing_factory.h>
#include <hpx/lcos/promise.h>
#include <hpx/lcos/future_wait.h>
#include <iostream>
#include <fstream>

#include <sstmac/software/libraries/filesystem/openfile_parallel.h>
#include <sstmac/software/libraries/filesystem/nppfs.h>

#include <sstmac/sstmacro.h>

namespace hpx_gtc
{
  SSTRegisterApp("hpx_gtc", gtc_app);

  using namespace sstmac;
  using namespace sstmac::sw;
  using namespace hpx;

  void
  gtc_app::consume_params(const sim_parameters::ptr &params)
  {
    gtcparams_ = config_parameters::construct(params);

    gtcparams_->ihybrid_ = 1;

  }

  void
  gtc_app::register_actions()
  {

  }

  void
  gtc_app::register_components()
  {
    hpx::globalid dumm;
    threadman_->register_component < gtc_particle
        > (gtc_particle::component_name(), gtc_particle::construct(dumm));
    threadman_->register_component < gtc_point
        > (gtc_point::component_name(), gtc_point::construct(dumm));
    threadman_->register_component < hpx::components::hpx_distributing_factory
        > (hpx::components::hpx_distributing_factory::component_name(), hpx::components::hpx_distributing_factory::construct(
            dumm));
  }

  /**
   * ---------------------------------------------------------------------------------------
   * setup
   *     : sets up the simulation, modifies the parameters as necessary
   * ---------------------------------------------------------------------------------------
   */

  void
  gtc_app::setup()
  {

    // Read the input file that contains the run parameters

    read_input_params();

    sstmac::sw::environment::const_ptr env = hpx_app::get_env();
    size_t appcnt = env->get_app_count();

    gtcparams_->setup(appcnt);

    //  Changing the units of a0 and a1 from units of "a" to units of "R_0"
    gtcparams_->a0_ = gtcparams_->a0_ * gtcparams_->a_;
    gtcparams_->a1_ = gtcparams_->a1_ * gtcparams_->a_;

    // numerical constant
    gtcparams_->mstep_ = max(2, gtcparams_->mstep_);
    gtcparams_->msnap_ = min(gtcparams_->msnap_,
        gtcparams_->mstep_ / gtcparams_->ndiag_);
    int isnap = gtcparams_->mstep_ / gtcparams_->msnap_;
    gtcparams_->idiag1_ = gtcparams_->mpsi_ / 2;
    gtcparams_->idiag2_ = gtcparams_->mpsi_ / 2;
    if (gtcparams_->nonlinear_ <= 0)
      {
        gtcparams_->paranl_ = 0.0;
        gtcparams_->mode00_ = 0;
        gtcparams_->idiag1_ = 1;
        gtcparams_->idiag2_ = gtcparams_->mpsi_;
      }
    gtcparams_->rc_ = gtcparams_->rc_ * (gtcparams_->a0_ + gtcparams_->a1_);
    gtcparams_->rw_ = 1.0
        / (gtcparams_->rw_ * (gtcparams_->a1_ - gtcparams_->a0_));

    // Set up the particle decomposition within each toroidal domain

    // equilibrium unit: length (unit=cm) and time (unit=second) unit
    double ulength = gtcparams_->r0_;
    double utime = 1.0 / (9580.0 * gtcparams_->b0_); // time unit = inverse gyrofrequency of proton

    // primary ion thermal gyroradius in equilibrium unit, vthermal=sqrt(T/m)
    gtcparams_->gyroradius_ = 102.0
        * sqrt(gtcparams_->aion_ * gtcparams_->temperature_)
        / (abs(gtcparams_->qion_) * gtcparams_->b0_) / ulength;
    double tstep = gtcparams_->tstep_ * gtcparams_->aion_
        / (abs(gtcparams_->qion_) * gtcparams_->gyroradius_
            * gtcparams_->kappati_);

    // basic ion-ion collision time, Braginskii definition
    gtcparams_->do_collision_ = false;
    double r, q, zeff, tau_vth;

    if (gtcparams_->tauii_ > 0)
      {
        //tau_vth=24.0_wp-log(sqrt(edensity0)/temperature) ! Coulomb logarithm
        //tau_vth=2.09e7_wp*(temperature)**1.5_wp/(edensity0*tau_vth*2.31_wp)*sqrt(2.0_wp)/utime
        r = 0.5 * (gtcparams_->a0_ + gtcparams_->a1_);
        q = gtcparams_->q0_ + gtcparams_->q1_ * r / gtcparams_->a_
            + gtcparams_->q2_ * r * r / (gtcparams_->a_ * gtcparams_->a_);
        zeff = gtcparams_->qion_; // Set zeff to main ion species

        tau_vth = 23.0
            - log(
                sqrt(zeff * zeff * gtcparams_->edensity0_)
                    / pow(gtcparams_->temperature_, 1.5));
        tau_vth = 2.09e7 * pow(gtcparams_->temperature_, 1.5)
            * sqrt(gtcparams_->aion_)
            / (gtcparams_->edensity0_ * tau_vth * utime * zeff);
        gtcparams_->tauii_ = tau_vth * gtcparams_->tauii_; // Multiply v_thermal tau by scaling factor tauii
        gtcparams_->do_collision_ = true;

      }

    int nparam;
    if (gtcparams_->track_particles_ == 1)
      {
        // We keep track of the particles by tagging them with a number
        // We add an extra element to the particle array, which will hold
        // the particle tag, i.e. just a number
        // Each processor has its own "ptracked" array to accumulate the tracked
        // particles that may or may not reside in its subdomain.
        // The vector "ntrackp" keeps contains the number of tracked particles
        // currently residing on the processor at each time step. We write out
        // the data to file every (mstep/msnap) steps.
        gtcparams_->nparam_ = 7;

      }
    else
      {
        // No tagging of the particles
        gtcparams_->nparam_ = 6;
      }

  }

  /**
   * ---------------------------------------------------------------------------------------
   * read_input_params
   *    : models reading the parameter file from disk
   * ---------------------------------------------------------------------------------------
   */

  void
  gtc_app::read_input_params()
  {

    // model the disk access of the config file
    nppfs::ptr fs = nppfs::construct(id_);
    register_lib(fs);
    std::string filename = "config";

    openfile_parallel::ptr file =
        boost::dynamic_pointer_cast<openfile_parallel>(fs->open(filename));
    fs->read(file, file->file_->total_size_);

    unregister_lib(fs);

  }

  void
  gtc_app::hpx_main()
  {

    timestamp start = now();

    // set up the problem from the parameters
    setup();

    gtcparams_->print_thyself();

    // Create a distributing factory. This object can be used to create
    // a distributed array of components.
    hpx::components::hpx_distributing_factory::ptr factory_points =
        hpx::components::hpx_distributing_factory::construct(
            agas_->new_gid(hpx::lva::COMPONENT));

    std::vector<gtc_point::stub_ptr> point_stubs =
        factory_points->create_components<gtc_point::stubs>(
            gtc_point::component_name(), gtcparams_->ntoroidal_);

    SSTMAC_DEBUG << "gtc_app: done distributing points \n";

    std::vector<hpx::lcos::promise<int>::ptr> initial_phase;

    for (std::size_t i = 0; i < point_stubs.size(); i++)
      {
        initial_phase.push_back(
            point_stubs[i]->init_async(i, gtcparams_, point_stubs));
      }
    hpx::lcos::wait(initial_phase);

    SSTMAC_DEBUG << "gtc_app: done initializing points \n";

    bool do_collision;
    if (gtcparams_->tauii_ > 0.0)
      {
        do_collision = true;
      }
    else
      {
        do_collision = false;
      }

      { // LOAD
        std::vector<hpx::lcos::promise<int>::ptr> load_phase;
        for (std::size_t i = 0; i < gtcparams_->ntoroidal_; i++)
          {
            load_phase.push_back(point_stubs[i]->load_async());
          }
        hpx::lcos::wait(load_phase);
      }

    SSTMAC_DEBUG << "gtc_app: done loading points \n";

    std::size_t istep = 0;
      { // CHARGEI
        std::vector<hpx::lcos::promise<int>::ptr> chargei_phase;
        for (std::size_t i = 0; i < gtcparams_->ntoroidal_; i++)
          {
            chargei_phase.push_back(point_stubs[i]->chargei_async(istep));
          }
        hpx::lcos::wait(chargei_phase);
      }

    // main time loop
    for (istep = 1; istep <= gtcparams_->mstep_; istep++)
      {
        SSTMAC_DEBUG << "gtc_app: main loop: Step : " << istep << "\n";
        for (std::size_t irk = 1; irk <= 2; irk++)
          {
            // idiag=0: do time history diagnosis
            std::size_t idiag = ((irk + 1) % 2) + (istep % gtcparams_->ndiag_);

              { // SMOOTH(3)
                std::vector<hpx::lcos::promise<int>::ptr> smooth_phase;
                std::size_t iflag = 3;
                for (std::size_t i = 0; i < gtcparams_->ntoroidal_; i++)
                  {
                    smooth_phase.push_back(
                        point_stubs[i]->smooth_async(iflag, idiag));
                  }
                hpx::lcos::wait(smooth_phase);
              }

              { // FIELD
                std::vector<hpx::lcos::promise<int>::ptr> field_phase;
                for (std::size_t i = 0; i < gtcparams_->ntoroidal_; i++)
                  {
                    field_phase.push_back(point_stubs[i]->field_async());
                  }
                hpx::lcos::wait(field_phase);
              }

            // push ion
              { // PUSHI
                std::vector<hpx::lcos::promise<int>::ptr> pushi_phase;
                for (std::size_t i = 0; i < gtcparams_->ntoroidal_; i++)
                  {
                    pushi_phase.push_back(
                        point_stubs[i]->pushi_async(irk, istep, idiag));
                  }
                hpx::lcos::wait(pushi_phase);
              }

            // redistribute ion across PEs
              { // SHIFTI
                std::vector<hpx::lcos::promise<int>::ptr> shifti_phase;
                for (std::size_t i = 0; i < gtcparams_->ntoroidal_; i++)
                  {
                    shifti_phase.push_back(point_stubs[i]->shifti_async());
                  }
                hpx::lcos::wait(shifti_phase);
              }

            if (irk == 2 && do_collision)
              {
                std::cerr
                    << " Collision not implemented yet.  Not a default parameter "
                    << std::endl;
              }

            // ion perturbed density
              { // CHARGEI
                std::vector<hpx::lcos::promise<int>::ptr> chargei_phase;
                for (std::size_t i = 0; i < gtcparams_->ntoroidal_; i++)
                  {
                    chargei_phase.push_back(
                        point_stubs[i]->chargei_async(istep));
                  }
                hpx::lcos::wait(chargei_phase);
              }

            // smooth ion density
              { // SMOOTH(0)
                std::vector<hpx::lcos::promise<int>::ptr> smooth_phase;
                std::size_t iflag = 0;
                for (std::size_t i = 0; i < gtcparams_->ntoroidal_; i++)
                  {
                    smooth_phase.push_back(
                        point_stubs[i]->smooth_async(iflag, idiag));
                  }
                hpx::lcos::wait(smooth_phase);
              }

            // solve GK Poisson equation using adiabatic electron
              /*{ // POISSON(0)
                std::vector<hpx::lcos::promise<int>::ptr> poisson_phase;
                std::size_t iflag = 0;
                for (std::size_t i = 0; i < gtcparams_->ntoroidal_; i++)
                  {
                    poisson_phase.push_back(
                        point_stubs[i]->poisson_async(iflag, istep, irk));
                  }
                hpx::lcos::wait(poisson_phase);
              }*/

            for (std::size_t ihybrid = 1; ihybrid <= gtcparams_->nhybrid_; ihybrid++)
              {
                std::cerr << " par->nybrid > 0 not supported yet " << std::endl;
              }
          }
      }

    /* hpx::components::hpx_distributing_factory::ptr factory_parts =
     hpx::components::hpx_distributing_factory::construct(
     agas_->new_gid(hpx::lva::COMPONENT));

     std::vector<gtc_particle::stub_ptr> part_stubs =
     factory_parts->create_components<gtc_particle::stubs>(
     gtc_particle::component_name(), num_part_);

     SSTMAC_DEBUG << "gtc_app: done distributing particles \n";

     // Initial Data -----------------------------------------


     std::string line;
     std::string val1, val2, val3, val4;
     std::ifstream myfile;
     std::size_t i = 0;
     std::string filename = partfile_.substr(partfile_.find_last_of("//") + 1);
     SSTMAC_DEBUG << "filename: " << filename << "\n";
     openfile_parallel::ptr file = boost::dynamic_pointer_cast<openfile_parallel>(
     fs_->open(filename));
     fs_->read(file, file->file_->total_size_);

     myfile.open(partfile_.c_str());
     if (myfile.is_open())
     {
     while (myfile.good())
     {
     while (std::getline(myfile, line))
     {
     std::istringstream isstream(line);
     std::getline(isstream, val1, ' ');
     std::getline(isstream, val2, ' ');
     std::getline(isstream, val3, ' ');
     std::getline(isstream, val4, ' ');
     std::size_t node = atoi(val1.c_str());
     double posx = atof(val2.c_str());
     double posy = atof(val3.c_str());
     double posz = atof(val4.c_str());
     initial_phase.push_back(
     part_stubs[i]->init_async(node, posx, posy, posz));

     }
     i++;
     }
     myfile.close();
     }

     // for (std::size_t i = 0; i < part_stubs.size(); i++)
     //   {
     //     initial_phase.push_back(part_stubs[i]->init_async(i));
     //   }
     */
    // vector gids of the particle components
    /* std::vector<hpx::naming::id_type> particle_components;
     for (std::size_t i = 0; i < num_particles; i++)
     {
     particle_components.push_back(accu_particles[i].get_gid());
     }*/
    /*
     // We have to wait for the futures to finish before exiting.
     hpx::lcos::wait(initial_phase);

     timestamp time_init = now();

     SSTMAC_DEBUG << "gtc_app: done with initialization phase \n";

     for (int i = 0; i < num_iterations_; i++)
     {
     SSTMAC_DEBUG << "gtc_app: beginning iteration " << i << "\n";
     // Start the search/charge depositing phase
     std::vector<hpx::lcos::promise<int>::ptr> charge_phase;

     for (std::size_t i = 0; i < point_stubs.size(); i++)
     {
     charge_phase.push_back(point_stubs[i]->search_async(part_stubs));
     }
     std::vector<int> search_vector;
     hpx::lcos::wait(charge_phase);
     }

     timestamp time_end = now();

     SSTMAC_DEBUG << "initialization time: " << (time_init - start) << "\n";
     SSTMAC_DEBUG << "charge time: " << (time_end - time_init) << "\n";
     */
  }

}
