/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/backends/native/eventmap.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>

namespace sstmac
{
  namespace native
  {

    // Pointer types.

    //
    // Hello.
    //
    eventmap::eventmap(const timestamp &tv) :
        running_(false)
    {
      set_now(tv);
      last_update_ = time(NULL);
      update_interval_ = 2;
      sim_start_ = last_update_;
      last_update_sim_ = timestamp(tv);
      complete_ = false;
    }

    //
    // Hello.
    //
    eventmap::ptr
    eventmap::construct(const timestamp &tv)
    {
      eventmap::ptr retval(new eventmap(tv));
      return retval;
    }

    //
    // Goodbye.
    //
    eventmap::~eventmap() throw ()
    {
      if (running_)
        {
          SSTMAC_DEBUG
                << "Warning: eventmap (eventmanager) exiting while running\n.";
        }
    }

    //
    // Clear all events and set time back to a zero of your choice.
    //
    void
    eventmap::clear(const timestamp &zero_time)
    {
      if (running_)
        {
          throw illformederror("eventmap::clear:  Event manager is running.");
        }
      queue_.clear();
      set_now(zero_time);
    }

    //
    // Run the eventmanager.
    //
    void
    eventmap::run()
    {
      if (running_)
        {
          throw illformederror(
              "eventmap::run:  Event manager already running.");
        }
      running_ = true;
      stopped_ = false;
      while (!queue_.empty() && !stopped_)
        {
          do_one_event();
        }

      running_ = false;

      if (queue_.empty())
        {
          complete_ = true;
          SSTMAC_DEBUG << "eventmap[" << this
                << "]::run:  All out of events.\n";

          std::list<stat_logger::ptr>::iterator it, end = stats_.end();
          for (it = stats_.begin(); it != end; it++)
            {
              (*it)->simulation_finished(now());
            }
        }
    }

    void
    eventmap::step()
    {
      if (complete_)
        {
          std::cout
              << "manager: step: eventmanager is complete, need to restart. \n";
          return;
        }
      else
        {
          do_one_event();
        }

      if (queue_.empty())
        {
          complete_ = true;
          SSTMAC_DEBUG << "eventmap[" << this
                << "]::run:  All out of events.\n";

          std::list<stat_logger::ptr>::iterator it, end = stats_.end();
          for (it = stats_.begin(); it != end; it++)
            {
              (*it)->simulation_finished(now());
            }
        }

    }

    void
    eventmap::do_one_event()
    {
      queue_t::iterator it = queue_.begin();
      set_now(it->first);
      sst_event::ptr ev = it->second;
      queue_.erase(it);
      SSTMAC_DEBUG << "eventmap[" << this << "]::run:  Processing event " << ev
            << " of type " << ev->to_string() << " at time " << now() << "\n";
      ev->execute();
      SSTMAC_DEBUG << "eventmap[" << this << "]::run:  Event " << ev
            << " complete, " << queue_.size() << " events pending\n";

      time_t t = time(NULL);
      if (updates_.is_active() && t >= last_update_ + update_interval_)
        {
          double simsecpersec = (now() - last_update_sim_).sec()
              / (t - last_update_);
          updates_ << "elapsed time: " << (t - sim_start_)
              << ", simulated time: " << now() << ", simsec/sec: "
              << simsecpersec << "\n";
          last_update_ = t;
          last_update_sim_ = now();
        }
    }

    //
    // Set off the given eventhandler at the given time.
    //
    void
    eventmap::schedule(const timestamp &start_time,
        const eventhandler::ptr &handler, const sst_message::ptr &msg)
    {
      if (start_time < now())
        {
          std::cerr << "start_time " << start_time << " is before current time "
              << now() << "\n";
          throw valueerror(
              "eventmap::schedule:  Start time before current time.");
        }
      if (handler == 0)
        {
          throw nullerror("eventmap::schedule:  Null eventhandler pointer.");
        }
      sst_event::ptr ev = sst_event::construct(msg, handler);
      queue_.insert(std::make_pair(start_time, ev));
      SSTMAC_DEBUG << "eventmap[" << this << "]::schedule:  Scheduling event "
            << ev->to_string() << start_time << ", " << queue_.size()
            << " events pending.\n";
    }

  }
} // end of namespace sstmac
