/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

//#include <backends/sstmacro-native/manager-env.h>
#include <sstmac/backends/native/manager.h>
#include <sstmac/backends/native/nodelist.h>
#include <sstmac/hardware/node/simplenode.h>
#include <sstmac/hardware/nic/simplenic.h>
#include <sstmac/hardware/node/detailednode.h>

#include <sstmac/backends/native/eventmap.h>

#include <sstmac/common/counter.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/cast.h>
#include <sstmac/software/ami/environment.h>

#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>

#include <iostream>
#include <iterator>
#include <cstdlib>

namespace sstmac
{
  namespace native
  {

    using namespace sstmac::sw;
    using namespace sstmac::hw;

    //
    // Private method to indicate that a launcher is exiting.
    //
    void
    manager::launcher_exiting(const launcher::ptr &launch,
        const std::vector<nodeid::ptr> &release)
    {
      launchers_t::iterator it = launchers_.find(launch);
      if (it != launchers_.end())
        {
          /* std::vector<nodeid::ptr>::const_iterator nid;
           for (nid = release.begin(); nid != release.end(); ++nid)
           {
           allocated_.erase(allocated_.find(*nid));
           available_.insert(*nid);
           }*/
          launchers_.erase(it);
        }
    }

    //
    // Default constructor.
    //
    manager::manager() :
      eventmanager_(eventmap::construct()), nodes_(nodelist::construct()),
          running_(false), next_ppid_(0), dbg_("<debug> manager")
    //stats(statscollector::construct("stats.log"))
    {
      SSTMAC_DEBUG << "manager::manager()\n";

      logger::timer_ = eventmanager_;
    }

    //
    // Construct a manager.
    //
    manager::ptr
    manager::construct()
    {
      manager::ptr retval(new manager);

      return retval;
    }

    //
    // Goodbye.
    //
    manager::~manager() throw ()
    {
      SSTMAC_DEBUG << "manager::~manager()\n";
      if (this->running_)
        {
          std::cerr
              << "FATAL:  manager going out of scope while still running.\n";
          abort();
        }
      /*if (!launchers_.empty())
       {
       std::cerr << "WARNING: manager exiting with active launchers.\n"
       << "         This is normally caused by application deadlock.\n";
       launchers_t::const_iterator it, end = launchers_.end();
       for (it = launchers_.begin(); it != end; ++it)
       {
       if (!(*it)->completed())
       {
       std::cerr << "    Launcher " << (*it)->description()
       << " is still reported as active\n";
       }
       }
       launchers_.clear();
       nodes_ = nodelist::ptr ();
       }*/SSTMAC_DEBUG << "manager::~manager() leaving\n";
    }

    //
    // Define a network.
    //
    void
    manager::set_interconnect(const interconnect::ptr &net)
    {
      SSTMAC_DEBUG << "manager::set_interconnect(" << net << ")\n";
      if (running_)
        {
          throw illformederror(
              "manager::set_interconnect: Simulation is running.");
        }
      interconnect_ = net;

      //stats->add_stats(interconnect_->get_stats());

    }

    //
    // Access the interconnect.
    //
    interconnect::const_ptr
    manager::get_interconnect() const
    {
      return interconnect_;
    }

    //
    // Number of nodes currently defined.
    //
    int64_t
    manager::nodes() const
    {

      return nodes_->size();
    }

    //
    // Number of open network slots.
    //
    int64_t
    manager::open_slots() const
    {
      if (interconnect_ == 0)
        {
          throw nullerror("manager::open_slots:  null interconnect pointer.");
        }
      if (interconnect_->slots() <= 0)
        {
          return -1;
        }
      else
        {
          return (interconnect_->slots() - this->nodes());
        }
    }

    node::ptr
    manager::get_node(nodeid::ptr id) const
    {
      return nodes_->get(id);
    }

    //
    // Add one or more copies of the given node type to the manager.
    //
    long
    manager::add_nodes(sim_parameters::ptr params)
    {
      SSTMAC_DEBUG << "manager::add_nodes() \n";

      std::string failmodel = params->get_optional_param("failure_model",
          "none");
      if (!fail_ && failmodel != "none")
        {
          SSTMAC_DEBUG
                << "manager::add_nodes() -- constructing failure model \n";
          fail_ = failure_config::construct(params);
        }

      sstmac::native::node::ptr type;
      if (params->get_param("node_name") == "simple")
        {
          type = simplenode::construct(params);
        }
      else if (params->get_param("node_name") == "detailed")
        {
          type = detailednode::construct(params);

        }
      else
        {
          std::cout << "node type: " << params->get_param("node_name");
          throw ssterror("run: don't know how to construct node type");
        }

      sstmac::hw::networkinterface::ptr nictype;
      if (params->get_param("nic_name") == "simple")
        {
          nictype = simplenic::construct(type, interconnect_);
        }
      else
        {
          std::cout << "nic type: " << params->get_param("nic_name");
          throw ssterror("run: don't know how to construct nic type");
        }

      if (interconnect_ == 0)
        {
          throw nullerror("manager::add_nodes:  No interconnect defined yet.");
        }
      if (type == 0)
        {
          throw nullerror("manager::add_nodes:  Null node pointer\n");
        }

      long count = params->get_long_param("node_num");
      if (count != interconnect_->slots())
        {
          throw illformederror(
              "node_num parameter does not match interconnect size. "
                "node_num should never be specified in the input file.");
        }

      SSTMAC_DEBUG << "manager::add_nodes(" << type << ", " << count << ")\n";

      /** DEPRECATED CHECK: No more node_num parameter
       if ((interconnect_->slots() > 0)
       && (count > int(interconnect_->slots() - this->nodes())))
       {
       throw valueerror("manager::add_nodes:  Insufficient slots available.");
       }
       */
      if (count < 0)
        {
          std::cerr << "manager::add_nodes:  Simulator has unlimited size. "
              << " Unlimited interconnect size is no longer valid." << endl;
          throw illformederror(
              "manager::add_nodes:  Infinite interconnect size.");

          /** DEPRECATED node_num parameter
           if (interconnect_->slots() <= 0)
           {
           std::cerr << "manager::add_nodes:  Simulator has unlimited size, "
           << "so an unlimited node count is not possible\n";
           throw illformederror("manager::add_nodes:  Need infinite memory.");
           }
           count = check_cast<int>(interconnect_->slots()) - this->nodes();
           */
        }
      for (int i = 0; i < count; ++i)
        {
          node::ptr newnode = type->clone();

          sstmac::hw::networkinterface::ptr newnic = nictype->clone(newnode,
              interconnect_);
          newnode->set_nic(newnic);

          nodeid::ptr id = nodes_->add(newnode);
          SSTMAC_DEBUG << "manager: adding nodeid " << id->to_string() << "\n";
          //available_.insert(id);
        }

      return this->open_slots();
    }

    //
    // Setup the simulation.
    //
    void
    manager::setup()
    {
      SSTMAC_DEBUG << "manager::setup: entered\n";
      if (interconnect_ == 0)
        {
          throw nullerror("manager::setup:  Null interconnect pointer.");
        }
      if (running_)
        {
          throw illformederror("manager::setup:  Simulation already running.");
        }
      // Initial setup.
      if (eventmanager_ == 0)
        {
          eventmanager_ = eventmap::construct();
        }
      // Make sure the network simulator knows about our event manager.
      interconnect_->set_eventmanager(eventmanager_);

      /** In case we have an mpi app, set the event manager so mpi statistics can
       be collected */
      mpiqueue::set_eventmanager(eventmanager_);

      // Initialize the nodes.
      for (int i = 0; i < nodes_->size(); ++i)
        {
          nodeid::ptr id = nodeid::construct(i);
          node::ptr the_node = nodes_->get(id);
          the_node->initialize(id, eventmanager_);
          the_node->get_nic()->set_eventmanager(eventmanager_);
          sstmac::hw::networkinterface::set_node_map(id, the_node->get_nic());

          //the_node->add_networkinterface(simplenic::construct(the_node,
          //                                                    interconnect_));
        }

      if (fail_)
        {
          fail_->config_nodes(nodes_, eventmanager_);

          fail_->config_network(interconnect_);
        }

      // Schedule job startups.
      // for (startup_t::iterator it = startup_.begin(); it != startup_.end(); ++it)
      //  {
      //    eventmanager_->schedule(it->first, it->second, sst_message::ptr());
      //  }
      startup_.clear();
      running_ = true;
    }

    //
    // Start the simulation.
    //
    timestamp
    manager::run(timestamp until)
    {

      // setup the simulation before starting
      setup();

      SSTMAC_DEBUG << "manager::run: calling eventmanager::run\n";

      if (until.sec() > 0)
        {
          eventmanager_->schedule(until, stophandler::construct(),
              sst_message::ptr());
        }

      eventmanager_->run();

      // Now call done routine to end simulation and print Stats.
      done();

      return eventmanager_->now();

    }

    void
    manager::done()
    {
      SSTMAC_DEBUG << "manager::done: entered\n";

      running_ = false;
      /*if (allocated_.size() > 0)
       {
       std::cout << "manager no longer running but " << allocated_.size()
       << " nodes are still allocated--deallocating all nodes."
       << std::endl;
       std::copy(allocated_.begin(), allocated_.end(),
       std::inserter(available_, available_.begin()));
       allocated_.clear();
       }*/

      // stats->printStats(eventmanager_->now());
    }

  }
} // end of namespace sstmac
