/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "cycleaccurateswitch.h"

#include "network_m.h"

namespace sstmac {
namespace omnetpp {

using namespace std;

stat_time_window_sum::ptr cycleaccurateswitch::power_stat_;
double cycleaccurateswitch::static_power_ = 0;
cMessage* cycleaccurateswitch::logPowerMessage = NULL;

Define_Module(cycleaccurateswitch);

void cycleaccurateswitch::initialize() {

	//latencyPerHop = getParam("latencyPerHop");

	int myrackid = par("rackid");
	int npr = par("rack_nodes");
	int nrc = par("numRackConnections");

	omnetpp_address::ptr me = omnetpp_address::construct(-1, -1, -1, myrackid);

	int topP1 = par("topParam1");
	int topP2 = par("topParam2");
	int topP3 = par("topParam3");

	string top = par("networkTopology");

	init(me, npr, nrc, top, topP1, topP2, topP3);

	numVC_ = par("numVC");
	inBuffSize_ = par("inBuffSize");
	outBuffSize_ = par("outBuffSize");

	arbLatency_ = par("arbitrationLatency");
	switchLatency_ = par("xBarLatency");
	switchBandwidth_ = par("xBarBandwidth");
	switchBandwidth_ *= 1e9;

	mtu = par("packetMTU");

	clockFreq_ = par("clockFreq");
	clockFreq_ *= 1e6;
	clockper_ = 1.0 / clockFreq_;

	for (int i = 0; i < numRackConnections; i++) {
		inports[gate("ports$i", i)] = i;
		outports[i] = gate("ports$o", i);

		for (int j = 0; j < numVC_; j++) {
			inbuffs_[i * numVC_ + j] = new cPacketQueue();

			credits_[i * numVC_ + j] = inBuffSize_;
		}

		outbuffs_[i] = new cPacketQueue();

		inports_free_[i] = 0;
		outports_free_[i] = 0;

		pending_creds_up_[i] = new std::queue<cMessage*>();
		arb_vc_[i] = 0;
		num_in_buff_[i] = 0;

		outBuffMsgs_[i] = new cMessage("outbuffmsg");
		revOutBuffMsgs_[outBuffMsgs_[i]] = i;
	}

	clockmsg_ = new cMessage("clockmsg");

	arb_inport_ = 0;

	if (!power_stat_) {
		stat_time_window::TIME_INTERVAL = timestamp(0.0001);
		std::stringstream ss;
		ss << "network_power-" << top << topP1 << "x" << topP2 << "x" << topP3 << ".csv";
		stat_name_ = ss.str();
		power_stat_ = stat_time_window_sum::construct("<stats> router",
				stat_name_);
		logger::register_logger(power_stat_);
	}

	int technology = 32;
	int trans_type = NVT;
	double voltage = 0.8;
	double frequency = clockFreq_;

	ORION_Tech_Config *conf = new ORION_Tech_Config(technology, trans_type,
			voltage, frequency);

	// ---------- input buffer ---------------
	int buf_type = SRAM;
	int is_fifo = true;
	int n_read_port = MAX_GRANTS;
	int n_write_port = 1;
	line_width_ = 128;
	int n_entry = inBuffSize_ / line_width_;

	int outdrv = false;

	ORION_Array_Info* info = new ORION_Array_Info();
	info->init(buf_type, is_fifo, n_read_port, n_write_port, n_entry,
			line_width_, outdrv, conf);

	power_buffer_in_ = new ORION_Array();
	power_buffer_in_->init(info);

	//std::cout << "bufferin: " << power_buffer_in_->stat_static_power() << "\n";
	static_power_ += numVC_ * numRackConnections * power_buffer_in_->stat_static_power();

	// ------------ output buffer --------------
	int out_buf_type = SRAM;
	int out_is_fifo = true;
	int out_n_read_port = 1;
	int out_n_write_port = 1;

	int out_n_entry = outBuffSize_ / line_width_;

	int out_outdrv = true;

	ORION_Array_Info* out_info = new ORION_Array_Info();
	out_info->init(out_buf_type, out_is_fifo, out_n_read_port, out_n_write_port, out_n_entry,
			line_width_, out_outdrv, conf);

	power_buffer_out_ = new ORION_Array();
	power_buffer_out_->init(out_info);

	//std::cout << "bufferout: " << power_buffer_out_->stat_static_power() << "\n";
	static_power_ += numRackConnections * power_buffer_out_->stat_static_power();

	// --------------- arbiter ------------------
	int model = RR_ARBITER;
	int ff_model = NEG_DFF;
	int req_width = (numRackConnections - 1) * numVC_;
	double length = numRackConnections * n_entry
			* conf->PARM("CrsbarCellWidth");

	power_arbiter_ = new ORION_Arbiter();

	power_arbiter_->init(model, ff_model, req_width, length, conf);
	//std::cout << "arbiter: " << power_arbiter_->report_static_power() << "\n";
	static_power_ += power_arbiter_->report_static_power();

	// ----------------- crossbar -----------------
	int xmodel = MATRIX_CROSSBAR;
	int ins = numRackConnections;
	int outs = numRackConnections;
	int in_seg = 0; //used by segmented crossbar
	int out_seg = 0; //used by segmented crossbar
	int degree = 4; //used by multree crossbar
	int connect_type = TRISTATE_GATE;
	double in_len = 0; //use 0 if unknown, ORION will figure it out
	double out_len = 0;

	power_xbar_ = new ORION_Crossbar();
	power_xbar_->init(xmodel, ins, outs, in_seg, out_seg, line_width_, degree,
			connect_type, trans_type, in_len, out_len, conf);
	//std::cout << "crossbar: " << power_xbar_->report_static_power() << "\n";
	static_power_ += power_xbar_->report_static_power();

	// --------------- links -------------
	double link_len = 1000;  // in um
	link_e_per_bm_ = ORION_Link::LinkDynamicEnergyPerBitPerMeter(link_len, conf);
	static_power_ += ORION_Link::LinkLeakagePowerPerMeter(link_len, conf);
	link_length_ = link_len / 1000000; // in meters

	if (logPowerMessage == NULL) {
		logPowerMessage = new cMessage("logPowerMessage");
		scheduleAt(0, logPowerMessage);
	}

}

void cycleaccurateswitch::finish() {

	for (int i = 0; i < numRackConnections; i++) {

		for (int j = 0; j < numVC_; j++) {

			int index = i * numVC_ + j;

			if (credits_[index] != inBuffSize_) {
				std::cout << "cycleaccurateswitch[" << myAddr->rack()
						<< "]: credits for VC " << j << " on inport " << i
						<< " (" << credits_[index]
						<< ") do not equal buff size(" << inBuffSize_ << ")"
						<< std::endl;
			}

			if (!inbuffs_[index]->empty()) {
				std::cout << "cycleaccurateswitch[" << myAddr->rack() << "]: "
						<< inbuffs_[index]->length()
						<< " messages still in inbuff " << i << ", vc = " << j
						<< std::endl;

				NetworkAccurateMsg * msg =
						(NetworkAccurateMsg*) inbuffs_[index]->front();
				omnetpp_address::ptr addr = msg->getToAddr();
				omnetpp_address::ptr from = msg->getFromAddr();
				bool isDimChange = false;
				int outport = router_->route(msg, addr, i, isDimChange);

				std::cout << "cycleaccurateswitch[" << myAddr->rack()
						<< "]: front message size " << msg->getByteLength()
						<< ", outport: " << outport << ", outport free: "
						<< outports_free_[outport] << ", outport buff size: "
						<< outbuffs_[outport]->getByteLength() << ", credits: "
						<< credits_[outport * numVC_ + j] << std::endl;

			}
		}

		if (!outbuffs_[i]->empty()) {
			std::cout << "cycleaccurateswitch[" << myAddr->rack() << "]: "
					<< outbuffs_[i]->length() << " messages still in outbuff "
					<< i << std::endl;
		}
	}

	for (int i = 0; i < numRackConnections; i++) {

		for (int j = 0; j < numVC_; j++) {
			delete inbuffs_[i * numVC_ + j];

		}

		delete outbuffs_[i];
		delete pending_creds_up_[i];
		delete outBuffMsgs_[i];
	}

	delete power_arbiter_;
	delete power_xbar_;
	delete power_buffer_in_;
	delete power_buffer_out_;

	delete clockmsg_;
}

void cycleaccurateswitch::handleMessage(cMessage* cmsg) {

	if (cmsg->isSelfMessage()) {
		if (cmsg == logPowerMessage) {
			std::cout << stat_name_ << ": Router static power: " << static_power_ << "\n";

			delete logPowerMessage;
			logPowerMessage = NULL;
		} else if (cmsg == clockmsg_) {
			arbitrate();
		} else if (revOutBuffMsgs_.find(cmsg) != revOutBuffMsgs_.end()) {
			int outport = revOutBuffMsgs_[cmsg];

			SSTMAC_DEBUG << "switch[" << myAddr->rack()
						<< "]: sending on outport " << outport << "\n";

			///send out the out buffer
			NetworkAccurateMsg* n =
					(NetworkAccurateMsg*) outbuffs_[outport]->front();
			outbuffs_[outport]->pop();
			messageInports_.erase(n);

			power_stat_->record(
					now(),
					power_buffer_out_->stat_energy(
							n->getByteLength() / line_width_, 0, 0));

			power_stat_->record(now(), link_e_per_bm_ * link_length_ * line_width_ * n->getByteLength() * 8);

			send(n, outports[outport]);

			if (!outbuffs_[outport]->empty()) {
				SSTMAC_DEBUG
							<<
							"switch["
							<< myAddr->rack()
							<< "]: scheduling outport selfmessage, queue size is "
							<< outports[outport]->size() << " \n";
				scheduleAt(
						outports[outport]->getTransmissionChannel()->getTransmissionFinishTime(),
						cmsg);
			}

			if (!clockmsg_->isScheduled()) {
				scheduleAt(simTime() + clockper_, clockmsg_);
			}

		} else { //it's a data message arriving to the outbuff
			NetworkAccurateMsg* n = check_and_cast<NetworkAccurateMsg*>(cmsg);
			omnetpp_address::ptr addr = n->getToAddr();

			bool isDimChange = false;
			unsigned int outport = router_->route(n, addr, 0, isDimChange);

			SSTMAC_DEBUG << "switch[" << myAddr->rack()
						<< "]: data message arriving to outbuff "
						<< (int) outport << "\n";

			outbuffs_[outport]->insert(n);

			power_stat_->record(
					now(),
					power_buffer_out_->stat_energy(0,
							n->getByteLength() / line_width_, 0));

			double initial = power_xbar_->report();
			power_xbar_->record(0, messageInports_[n], outport,
					n->getByteLength());
			power_stat_->record(now(), power_xbar_->report() - initial);

			if (!outBuffMsgs_[outport]->isScheduled()) {
				simtime_t atTime =
						max(
								simTime() + clockper_,
								outports[outport]->getTransmissionChannel()->getTransmissionFinishTime());
				scheduleAt(atTime, outBuffMsgs_[outport]);

			}

			if (!clockmsg_->isScheduled()) {
				scheduleAt(simTime() + clockper_, clockmsg_);
			}
		}
	} else {
		NetworkAccurateMsg* n = check_and_cast<NetworkAccurateMsg*>(cmsg);
		omnetpp_address::ptr addr = n->getToAddr();
		omnetpp_address::ptr from = n->getFromAddr();

		cGate* ingate = cmsg->getArrivalGate();

		int inport = inports[ingate];
		int inbuffindex = inport * numVC_ + n->getVc();
		int msgtype = n->getCamsgtype();

		SSTMAC_DEBUG << "switch[" << myAddr->rack()
					<< "]: message arrived on inport " << inport << " on VC "
					<< n->getVc() << " ...";

		if (msgtype == CAMSG_CREDITUP) {
			int creds = n->getCredits();

			credits_[inbuffindex] += creds;

			SSTMAC_DEBUG << "credits up by " << creds << " to "
						<< credits_[inbuffindex] << "\n";

			if (!clockmsg_->isScheduled()) {
				scheduleAt(simTime() + clockper_, clockmsg_);
			}

			delete cmsg;
			return;
		} else { //it's data
			SSTMAC_DEBUG << "data \n";

			inbuffs_[inbuffindex]->insert(n);
			num_in_buff_[inport]++;
			messageInports_[n] = inport;

			power_stat_->record(
					now(),
					power_buffer_in_->stat_energy(0,
							n->getByteLength() / line_width_, 0));

			if (!clockmsg_->isScheduled()) {
				scheduleAt(simTime() + clockper_, clockmsg_);
			}

		}
	}

}

void cycleaccurateswitch::arbitrate() {

	//iterate over input ports, starting with the last one that went
	int maxGrants = MAX_GRANTS;
	int numGrants = 0;

	int begininport = (arb_inport_ + 1) % numRackConnections;
	bool firstinport = true;
	for (int i = begininport;
			firstinport || (i != begininport && numGrants < maxGrants);
			i = (i + 1) % numRackConnections) {
		SSTMAC_DEBUG << "switch[" << myAddr->rack() << "]: examining inport "
					<< i << ", num_in_buff = " << num_in_buff_[i] << ", time = "
					<< simTime().dbl() << ", inport_free = "
					<< inports_free_[i].dbl() << "\n";
		firstinport = false;
		if (num_in_buff_[i] > 0) {
			if (inports_free_[i] <= simTime()) {

				int beginVC = (arb_vc_[i] + 1) % numVC_;
				bool firstVC = true;
				for (int vc = beginVC; firstVC || vc != beginVC;
						vc = (vc + 1) % numVC_) {

					firstVC = false;
					int index = i * numVC_ + vc;

					SSTMAC_DEBUG << "switch[" << myAddr->rack()
								<< "]: examining vc " << vc << ", queue size = "
								<< inbuffs_[index]->length() << "";

					if (!inbuffs_[index]->empty()) {
						NetworkAccurateMsg * msg =
								(NetworkAccurateMsg*) inbuffs_[index]->front();
						omnetpp_address::ptr addr = msg->getToAddr();
						omnetpp_address::ptr from = msg->getFromAddr();
						bool isDimChange = false;
						int outport = router_->route(msg, addr, i, isDimChange);

						if (msg->getVc() != vc) {
							opp_error(
									"cycleaccurate switch: vcs don't match for some reason");
						}

						SSTMAC_DEBUG << ", outports_free = "
									<< outports_free_[outport].dbl()
									<< ", outbuff occupancy = "
									<< outbuffs_[outport]->getByteLength()
									<< ", credits = "
									<< credits_[outport * numVC_ + vc] << "\n";

						if (outports_free_[outport] <= simTime()
								&& (outBuffSize_
										- outbuffs_[outport]->getByteLength())
										>= msg->getByteLength()
								&& credits_[outport * numVC_ + vc]
										>= (isDimChange ? mtu : 0)
												+ msg->getByteLength()) {
							SSTMAC_DEBUG << "switch[" << myAddr->rack()
										<< "]: arbitration good from " << i
										<< " to " << outport << " \n";

							inbuffs_[i * numVC_ + vc]->pop();
							//now we're ok to send through the crossbar
							//  sendDelayed(msg, arbLatency_, outports[outport]);
							simtime_t outbuffarrival = simTime()
									+ switchLatency_
									+ msg->getByteLength() * 8
											/ switchBandwidth_;
							scheduleAt(outbuffarrival, msg);

							power_stat_->record(
									now(),
									power_buffer_in_->stat_energy(
											msg->getByteLength() / line_width_,
											0, 0));

							double initial = power_xbar_->report();
							power_xbar_->record(1, messageInports_[msg],
									outport, msg->getByteLength());
							power_stat_->record(now(),
									power_xbar_->report() - initial);

							inports_free_[i] = outbuffarrival;
							outports_free_[outport] = outbuffarrival;
							arb_vc_[i] = vc;
							numGrants++;
							arb_inport_ = i;
							num_in_buff_[i]--;

							int inbuffindex = outport * numVC_ + vc;
							credits_[inbuffindex] -= msg->getByteLength();

							//send credit ack back

							NetworkAccurateMsg* ack = new NetworkAccurateMsg();
							ack->setByteLength(8);
							ack->setCredits(msg->getByteLength());
							ack->setCamsgtype(CAMSG_CREDITUP);
							ack->setVc(vc);
							// outbuffs_[i]->insert(ack);

							if (!outBuffMsgs_[i]->isScheduled()) {
								if (outbuffs_[i]->length() > 0)
									outbuffs_[i]->insertBefore(
											outbuffs_[i]->front(), ack);
								else
									outbuffs_[i]->insert(ack);
								simtime_t atTime =
										max(
												simTime() + clockper_,
												outports[i]->getTransmissionChannel()->getTransmissionFinishTime());
								scheduleAt(atTime, outBuffMsgs_[i]);

							} else {

								outbuffs_[i]->insertAfter(outbuffs_[i]->front(),
										ack);

							}

							break;
						}
					}
				}
			}
		}

	}

	if (numGrants > 0) {
		double initial = power_arbiter_->report();
		power_arbiter_->record(numGrants, numGrants);
		power_stat_->record(now(), power_arbiter_->report() - initial);

		scheduleAt(simTime() + clockper_, clockmsg_);
	}
}

timestamp cycleaccurateswitch::now() {
	timestamp t(simTime().dbl());
	return t;
}

/*

 void cycleaccurateswitch::tryOutQ() {
 bool successful = false;
 simtime_t closestFinish = MAX_SIM_TIME;
 for (int i = 0; i < numRackConnections; i++) {

 if (!outQs[i]->empty()) {
 simtime_t
 txFinishTime =
 outports[i]->getTransmissionChannel()->getTransmissionFinishTime();
 if (txFinishTime <= simTime()) { //channel free
 //std::cout << "NIC: sending message " << endl;
 cPacket* msg = outQs[i]->pop();
 send(msg, outports[i]);

 successful = true;
 } else if (!transFinish->isScheduled()) {
 if (txFinishTime < closestFinish)
 closestFinish = txFinishTime;
 }
 }
 }

 if (successful)
 tryOutQ();
 else if (closestFinish < MAX_SIM_TIME)
 scheduleAt(closestFinish, transFinish);

 }
 */
}
}
