/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "cycleaccuratenic.h"

#include <network_m.h>

namespace sstmac
{
  namespace omnetpp
  {

    Define_Module(cycleaccuratenic);

    void
    cycleaccuratenic::initialize()
    {
      fromProc = gate("toProcessor$i", 0);
      toProc = gate("toProcessor$o", 0);
      fromNet = gate("toNetwork$i", 0);
      toNet = gate("toNetwork$o", 0);

      mtu = par("packetMTU");

      numVC_ = par("numVC");

      nextVC_ = rand() % numVC_;

      for (int vc = 0; vc < numVC_; vc++)
        {
          credits_[vc] = par("inBuffSize");
        // credits_[vc] *= 0.5;
        }

      sent = 0;
      currentMsg = NULL;

      nextPacketMsg = new cMessage("nextPacketMsg");
    }

    void
    cycleaccuratenic::finish()
    {

      if (!credAckQ_.empty())
        {
          std::cout << "Warning: NIC credackQ is not empty" << std::endl;
        }

      if(!netQ.empty()){
          std::cout << "NIC: netQ still contains " << netQ.length() << " messages " << std::endl;
      }
      delete nextPacketMsg;
    }

    void
    cycleaccuratenic::handleMessage(cMessage *cmsg)
    {
      cGate* arrival = cmsg->getArrivalGate();

      if (cmsg == nextPacketMsg)
        {
          if (currentMsg == NULL)
            try_net_Q();
          else
            next_packet();
        }
      else if (arrival == fromProc)
        {
          CoreMsg* c = check_and_cast<CoreMsg*>(cmsg);

          netQ.insert(c);
          if (currentMsg == NULL)
            try_net_Q();

        }
      else if (arrival == fromNet)
        {
          NetworkAccurateMsg* net = check_and_cast<NetworkAccurateMsg*>(cmsg);

          if (net->getCamsgtype() == CAMSG_CREDITUP)
            {

              credits_[net->getVc()] += net->getCredits();

              SSTMAC_DEBUG << "NIC: credits up by " << net->getCredits()
                      << " to " << credits_[net->getVc()] << "\n";

              if (currentMsg == NULL)
                try_net_Q();
              else
                next_packet();

            }
          else if (net->getIsLast())
            {
              SSTMAC_DEBUG
              << "NIC: Last packet arrived, forwarding to Processor \n";

              //send credit back upstream
              NetworkAccurateMsg* ack = new NetworkAccurateMsg();
              ack->setByteLength(8);
              ack->setCredits(net->getByteLength());
              ack->setCamsgtype(CAMSG_CREDITUP);
              ack->setVc(net->getVc());
              credAckQ_.insert(ack);
              next_packet();

              net->setByteLength(0x7FFFFFFF); //do this so omnet won't complain about decapsulating the whole app msg at the end
              CoreMsg* m = (CoreMsg*) net->decapsulate();
              send(m, toProc);

            }
          else
            {
              //send credit back upstream
              NetworkAccurateMsg* ack = new NetworkAccurateMsg();
              ack->setByteLength(8);
              ack->setCredits(net->getByteLength());
              ack->setCamsgtype(CAMSG_CREDITUP);
              ack->setVc(net->getVc());
              credAckQ_.insert(ack);
              next_packet();
            }
          delete net;
        }
    }

    void
    cycleaccuratenic::next_packet()
    {

      if (!nextPacketMsg->isScheduled())
        {

          if (!credAckQ_.empty()) //credit acks get priority
            {
              double delay = 2e-9;
              cPacket* ack = credAckQ_.front();
              credAckQ_.pop();

              send(ack, toNet);

              simtime_t txFinishTime =
                  toNet->getTransmissionChannel()->getTransmissionFinishTime();
              scheduleAt(txFinishTime, nextPacketMsg);
            }
          else
            {

              if (currentMsg == NULL)
                {
                  opp_error(
                      "cycleaccuratenic: in next_packet(), but currentMsg is null. shouldn't be here.");
                }
              bool last = (mtu < 0) || ((sent + mtu) > currentMsg->getByteLength());

              int packsize = std::min(
                  (mtu > 0) ? mtu : 0x7FFFFFFF,
                      (mtu < 0) ?
                          (int) currentMsg->getByteLength() :
                          (int) (currentMsg->getByteLength() - sent));

              if (packsize <= credits_[nextVC_])
                {

                  NetworkAccurateMsg* pack = new NetworkAccurateMsg();
                  pack->setToAddr(currentMsg->getToAddr());
                  pack->setFromAddr(currentMsg->getFromAddr());
                  pack->setIsLast(last);
                  pack->setVc(nextVC_);
                  pack->setCamsgtype(CAMSG_DATA);
                  pack->setKind(currentMsg->getKind());
                  pack->setRoute(currentRoute_);

                  if (last)
                    {
                      // int diff = currentMsg->getByteLength() - sent;
                      pack->encapsulate(currentMsg);
                      pack->setByteLength(packsize);
                      sent = 0;
                    }
                  else
                    {
                      pack->setByteLength(packsize);
                      sent += mtu;
                    }
                  SSTMAC_DEBUG << "NIC: Sending a packet of length "
                      << (int) pack->getByteLength() << "\n";

                  send(pack, toNet);

                  credits_[nextVC_] -= packsize;

                  if (last)
                    {
                      currentMsg = NULL;
                      nextVC_ = (nextVC_ + 1) % numVC_;
                    }

                  simtime_t txFinishTime =
                      toNet->getTransmissionChannel()->getTransmissionFinishTime();
                  scheduleAt(txFinishTime, nextPacketMsg);

                }
              else
                {
                  SSTMAC_DEBUG << "NIC: Not enough credits, stalling \n";
                }

            }
        }

    }

    void
    cycleaccuratenic::try_net_Q()
    {
      if (!credAckQ_.empty())
        {
          next_packet();
        }
      else if (!netQ.empty())
        {
          CoreMsg* msg = check_and_cast<CoreMsg*>(netQ.pop());
          currentMsg = msg;
          currentRoute_ = rand() % 2;

          //ack on start of transmission
          if (msg->getMsgType() == APP)
            {
              ApplicationMsg* appmsg = (ApplicationMsg*) msg;
              send_proc_ack(appmsg->dup(), toProc);

            }

          next_packet();

        }

    }

  }
}
