/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "abstractnodeprocessor.h"

#include <sstmac/hardware/node/simplenode.h>
#include <sstmac/hardware/node/degradednode.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/software/services/launch/instantlaunch.h>

#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>

#include <sstmac/backends/omnetpp/omnetpp_params.h>

#include <sstmac/hardware/network/topology/hdtorus.h>

namespace sstmac
{
  namespace omnetpp
  {

    Define_Module(sstmac::omnetpp::abstractnodeprocessor);

    using namespace std;

    debug abstractnodeprocessor::dbg_;

    abstractnodeprocessor::abstractnodeprocessor()
    {
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> omnetnode");
          logger::register_logger(&dbg_);
        }

    }

    abstractnodeprocessor::~abstractnodeprocessor()
    {

    }

    void
    abstractnodeprocessor::initialize()
    {

      int c = 0;

      int p = 0;

      int n = par("nodeid");

      int r = par("rackid");

      myid = omnetpp_address::construct(c, p, n, r);

      omnetpp_params::ptr params = omnetpp_params::construct(this);

      coresPerProc = par("node_cores");
      numOfNodes = par("node_num");
      procsPerNode = par("node_procs");
      nodesPerRack = par("rack_nodes");
      numRacks = par("rack_num");

      omnetpp_address::set_system_params(coresPerProc, procsPerNode,
          nodesPerRack, numRacks);

      outgate = gate("toNetwork$o");
      ingate = gate("toNetwork$i");

      boost::intrusive_ptr<fakeeventmanager> fem = fakeeventmanager::construct(
          this);

      string temp = par("logger_params");
      bool color = par("debug_coloring");
      logger::set_debug_coloring(color);
      logger::set_user_param(temp);
      logger::timer_ = fem;

      // ---- get parameters and such
      string app_p_fname = par("app_paramfile");
      sim_parameters::ptr appparams = file_parameters::construct(app_p_fname);
      params->add_param_set(appparams);

      string man = par("manager");

      if (man == "simple")
        {
          node_ = sstmac::hw::simplenode::construct(params);
        }
      else if (man == "degraded")
        {
          bool found = false;
          string deg = par("manager_degraded_nodes");

          if (deg == "evens")
            {
              if (myid->unique_id() % 2 == 0)
                {
                  found = true;
                  node_ = sstmac::hw::degradednode::construct(params);
                }
            }
          else if (deg == "odds")
            {
              if (myid->unique_id() % 2 == 1)
                {
                  found = true;
                  node_ = sstmac::hw::degradednode::construct(params);
                }
            }
          else
            {
              bool error = false;
              std::vector<int> degraded = get_intvec(deg.c_str(), error);

              for (size_t i = 0; i < degraded.size(); i++)
                {
                  if (degraded[i] == myid->unique_id())
                    {
                      found = true;
                      node_ = sstmac::hw::degradednode::construct(params);
                      break;
                    }
                }

            }

          if (!found)
            {
              node_ = sstmac::hw::simplenode::construct(params);
            }
        }
      else
        {
          opp_error("unknown manager name");
        }
      nic_ = fakenic::construct(node_, this, numOfNodes);
      node_->set_nic(nic_);
      nic_->set_eventmanager(fem);
      mpiqueue::set_eventmanager(fem);

      node_->initialize(myid, fem);

    }

    void
    abstractnodeprocessor::finish()
    {

    }

    bool
    abstractnodeprocessor::is_master_core()
    {
      return ((myid->rack() == 0 || myid->rack() < 0) && myid->node() == 0);
    }

    void
    abstractnodeprocessor::handleMessage(cMessage* cmsg)
    {

      if (cmsg->isSelfMessage())
        {

          EventWrapperMsg* emsg = check_and_cast<EventWrapperMsg*>(cmsg);
          eventhandlerpod epod = emsg->getEvhandler();
          eventhandler::ptr evh = epod.load;
          sst_message::ptr m = emsg->getMsgcontent().load;
          evh->handle(m);

          delete emsg;

        }
      else if (cmsg->getArrivalGate() == ingate)
        { //message arriving in from the NoC
          CoreMsg *m = check_and_cast<CoreMsg*>(cmsg);

          if (m->getToAddr()->rack() != myid->rack()
              || m->getToAddr()->node() != myid->node())
            {
              std::cout << "My ID: " << myid->to_string() << endl;
              std::cout << "Msg Dest: " << m->getToAddr()->to_string() << endl;
              opp_error("Application: got a message destined for NOT me");
            }

          switch (m->getMsgType())
            {

          case APP:
            {
              SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc "
                    << myid->to_string() << ": app message arrived \n";
              ApplicationMsg* appmsg = check_and_cast<ApplicationMsg*>(m);
              payloadpod ppod = appmsg->getMsgcontent();
              sst_message::ptr bsstm = ppod.load;

              node_->handle(bsstm);

              break;
            }

          case NIC_SEND_ACK:
            {
              SSTMAC_DEBUG << "[t=" << now().sec() << "] -- Proc "
                    << myid->to_string() << ": NIC send ack arrived ";
              ApplicationMsg* appmsg = check_and_cast<ApplicationMsg*>(
                  m->decapsulate());

              payloadpod ppod = appmsg->getMsgcontent();
              sst_message::ptr bsstm = ppod.load;

              SSTMAC_DEBUG << bsstm->to_string() << "\n";
              node_->handle(bsstm);

              delete appmsg;
              //}
              break;
            }

          case OS_KILL:
            //delete app;
            // app = NULL;

            break;

          default:
            std::cout << "Proc id: " << myid->to_string() << "\n";
            opp_error("Proc: unknown OS message type received.");

            }

          delete cmsg;

        }
      else
        {
          opp_error("abstractnodeprocessor: unexpected message");
        }
    }

    timestamp
    abstractnodeprocessor::now()
    {
      timestamp t(simTime().dbl());
      return t;
    }

  // ------------------------------------ //

  }
}
