/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_OMNETPP_OMNETPP_PARAMS_H_INCLUDED
#define SSTMAC_BACKENDS_OMNETPP_OMNETPP_PARAMS_H_INCLUDED

#include <sstmac/common/sim_parameters.h>
#include <sstmac/common/errors.h>
#include <omnetpp.h>

namespace sstmac
{
  namespace omnetpp
  {

    class omnetpp_params : public sim_parameters
    {
    protected:
      omnetpp_params(cSimpleModule* p) :
          parent_(p)
      {

      }

    public:
      typedef boost::intrusive_ptr<omnetpp_params> ptr;

      static ptr
      construct(cSimpleModule* p)
      {
        return ptr(new omnetpp_params(p));
      }

      virtual std::string
      get_param(std::string key) const
      {
        if (parent_->hasPar(key.c_str()))
          {
            return parent_->par(key.c_str());
          }
        else
          {
            std::list<sim_parameters::ptr>::const_iterator it, end =
                other_params_.end();
            for (it = other_params_.begin(); it != end; it++)
              {
                if ((*it)->has_param(key))
                  {
                    return (*it)->get_param(key);
                  }
              }
          }

        throw valueerror("omnetpp_params: does not contain key " + key);
      }

      virtual int
      get_int_param(std::string key) const
      {
        if (parent_->hasPar(key.c_str()))
          {
            return parent_->par(key.c_str());
          }
        else
          {
            std::list<sim_parameters::ptr>::const_iterator it, end =
                other_params_.end();
            for (it = other_params_.begin(); it != end; it++)
              {
                if ((*it)->has_param(key))
                  {
                    return (*it)->get_int_param(key);
                  }
              }
          }

        throw valueerror("omnetpp_params: does not contain key " + key);
      }

      virtual long
      get_long_param(std::string key) const
      {
        if (parent_->hasPar(key.c_str()))
          {
            return parent_->par(key.c_str());
          }
        else
          {
            std::list<sim_parameters::ptr>::const_iterator it, end =
                other_params_.end();
            for (it = other_params_.begin(); it != end; it++)
              {
                if ((*it)->has_param(key))
                  {
                    return (*it)->get_long_param(key);
                  }
              }
          }

        throw valueerror("omnetpp_params: does not contain key " + key);
      }

      virtual double
      get_double_param(std::string key) const
      {
        if (parent_->hasPar(key.c_str()))
          {
            return parent_->par(key.c_str());
          }
        else
          {
            std::list<sim_parameters::ptr>::const_iterator it, end =
                other_params_.end();
            for (it = other_params_.begin(); it != end; it++)
              {
                if ((*it)->has_param(key))
                  {
                    return (*it)->get_double_param(key);
                  }
              }
          }

        throw valueerror("omnetpp_params: does not contain key " + key);
      }

      virtual timestamp
      get_time_param(std::string key) const
      {
        SSTMAC_DEBUG << "omnetpp_params: getting time param called " << key << "\n";
        if (parent_->hasPar(key.c_str()))
          {
            SSTMAC_DEBUG << "omnetpp_params: param in ned file \n";
            double d = parent_->par(key.c_str());
            return timestamp(d);
          }
        else
          {
            std::list<sim_parameters::ptr>::const_iterator it, end =
                other_params_.end();
            for (it = other_params_.begin(); it != end; it++)
              {
                if ((*it)->has_param(key))
                  {
                    return (*it)->get_time_param(key);
                  }
              }
          }

        throw valueerror("omnetpp_params: does not contain key " + key);
      }

      virtual bool
      has_param(std::string key) const
      {
        bool has = parent_->hasPar(key.c_str());

        std::list<sim_parameters::ptr>::const_iterator it, end =
            other_params_.end();
        for (it = other_params_.begin(); it != end && !has; it++)
          {
            has = has | (*it)->has_param(key);
          }
        return has;
      }

      virtual void
      add_param(std::string key, std::string val)
      {
        throw unimplementederror("omnetpp_params::add_param - don't call this");
      }

      virtual void
      add_param_override(std::string key, std::string val)
      {
        throw unimplementederror(
            "omnetpp_params::add_param_override - don't call this");

      }

      virtual void
      add_param_set(sim_parameters::ptr sim)
      {
        other_params_.push_back(sim);
      }

      virtual void
      combine_into(sim_parameters::ptr sp)
      {
        throw unimplementederror(
            "omnetpp_params::combine_into - don't call this");
      }

      virtual std::string
      to_string() const
      {
        return "omnetpp_parameters";
      }

      virtual
      ~omnetpp_params()
      {
      }

    protected:
      cSimpleModule* parent_;
      std::list<sim_parameters::ptr> other_params_;

    };

  }
}

#endif
