/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_VALUEPAYLOAD_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_VALUEPAYLOAD_H_INCLUDED

#include <sstmac/common/messages/payload.h>

namespace sstmac
{

  /**
   * Network payload consisting of a value with sensible copy semantics.
   *
   * Don't use this for pointer types, since you will break assumptions
   * about const-ness or protection of underlying data.
   */
  template<typename T>
    class valuepayload : public payload
    {
      /// The data we keep.
      mutable T data_;
      size_t size_;

      /// Construction time.
      valuepayload(const T &inval, size_t size = sizeof(T)) :
          data_(inval), size_(size)
      {
      }

      /// Construction with type conversion.
      template<typename Other>
        valuepayload(const Other &inval, size_t size = sizeof(T)) :
            data_(inval), size_(size)
        {
        }

      /// Private copy constructor.
      valuepayload(const valuepayload<T> &other, size_t size = sizeof(T)) :
          data_(other.data_), size_(size)
      {
      }

    public:
      typedef boost::intrusive_ptr<valuepayload<T> > ptr;
      typedef boost::intrusive_ptr<const valuepayload<T> > const_ptr;

      virtual
      ~valuepayload()
      {
      }

      /// Construct a new valuepayload.
      static typename valuepayload::const_ptr
      construct(const T &inval)
      {
        // Results in ICE in gcc-4.0.
        //return typename boost::shared_ptr< valuepayload<T> >
        //  (new valuepayload(inval));
        return typename valuepayload::const_ptr(new valuepayload<T>(inval));
      }

      /// Construct a new valuepayload with type conversion.
      template<typename Other>
        static typename valuepayload::const_ptr
        construct(const Other &inval)
        {
          // Results in ICE in gcc-4.0
          //return typename boost::shared_ptr< valuepayload<T> >
          //  (new valuepayload(inval));
          return typename valuepayload::const_ptr(new valuepayload<T>(inval));
        }

      /// Clone this object.
      payload::const_ptr
      clone() const
      {
        // Results in ICE in gcc-4.0
        //return typename boost::shared_ptr< valuepayload<T> >
        //  (new valuepayload(*this));
        return typename valuepayload::const_ptr(new valuepayload<T>(*this));
      }

      /// Access the underlying data in a non-const context.
      T&
      data_unconst()
      {
        return data_;
      }

      /// Access the underlying data in a const context.
      const T&
      data() const
      {
        return data_;
      }



      virtual long
      get_byte_length() const
      {
        return size_;
      }

      /**
       * Add operator
       * @param other
       * @return a smart pointer to a new valuepayload object
       */
      virtual const payload::const_ptr
      add(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<T> >(
            other);
        //std::cout << "adding " << casted->data() << " to " << data() << std::endl;
        T newval = casted->data() + data();
        payload::const_ptr ret = construct(newval);
        return ret;
      }

      /**
       * Less than comparator
       * @param other
       * @return
       */
      virtual bool
      less_than(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<T> >(
            other);
        //std::cout << "adding " << casted->data() << " to " << data() << std::endl;
        return data() < casted->data();

      }

      /**
       * Greater than comparator
       * @param other
       * @return
       */
      virtual bool
      greater_than(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<T> >(
            other);
        //std::cout << "adding " << casted->data() << " to " << data() << std::endl;
        return data() > casted->data();

      }

      /**
       * Equals comparator
       * @param other
       * @return
       */
      virtual bool
      equals(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<const valuepayload<T> >(
            other);
        //std::cout << "adding " << casted->data() << " to " << data() << std::endl;
        return data() == casted->data();

      }

      /**
       * Serialize this object into the sst_serializer, for
       * running in parallel simulation
       * @param ser the serializer to use
       */
      virtual void
      serialize(sst_serializer* ser) const
      {
        SSTMAC_DEBUG << "valuepayload<" << typeid(T).name()
              << ">: serializing \n";
        ser->pack(get_deser_name());
        ser->pack(data_);
      }

      /**
       * Deserialize a valuepayload out of the serializer
       * @param ser the serializer that contains a valuepayload
       * @return a smart pointer to a new valuepayload object
       */
      static payload::const_ptr
      deserialize(sst_serializer* ser)
      {
        SSTMAC_DEBUG << "valuepayload<" << typeid(T).name()
              << ">: deserializing \n";

        T v;
        ser->unpack(v);

        return construct(v);
      }

      /**
       * Strinfier
       * @return a string description
       */
      virtual std::string
      to_string() const
      {
        std::stringstream ss;
        ss << "valuepayload(" << data_ << ")";
        return ss.str();
      }

      /**
       * The deserialization name, so sst_serializer can reconstruct this
       * @return a string name for this class
       */
      static std::string
      get_deser_name()
      {
        std::stringstream ss;
        ss << "valuepayload<" << typeid(T).name() << ">";
        return ss.str();
      }
    };

} // end of namespace sstmac
#endif
