/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_VECTORPAYLOAD_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_VECTORPAYLOAD_H_INCLUDED

#include <sstmac/common/config.h>
#include <sstmac/common/messages/payload.h>
#ifdef HAVE_BOOST_1_48
#include <boost/multi_array.hpp>
#endif

namespace sstmac
{

  /**
   * Network payload consisting of a boost multiarray.
   *
   */
  template<typename Type, typename VectorType = std::vector<Type> >
    class vector1payload : public payload
    {
      /// The data we keep.
      typedef VectorType arraytype;

      arraytype data_;

      std::vector<int> dims_;

      /// Construction time.
      vector1payload(const arraytype &inval, int dim1) :
          data_(inval)
      {
        dims_.push_back(dim1);
      }

      vector1payload(const Type* buf, int dim1)
      {
        dims_.push_back(dim1);
        for (int i = 0; i < dim1; i++)
          {
            std::cout << "constructing value: " << buf[i] << "\n";
            data_.push_back(buf[i]);
          }
      }

    public:
      typedef boost::intrusive_ptr<vector1payload<Type, VectorType> > ptr;
      typedef boost::intrusive_ptr<const vector1payload<Type, VectorType> > const_ptr;

      virtual
      ~vector1payload()
      {
      }

      static const_ptr
      construct(const arraytype &inval, int dim1)
      {
        return const_ptr(new vector1payload<Type, VectorType>(inval, dim1));
      }

      static const_ptr
      construct(const Type* inval, int dim1)
      {
        return const_ptr(new vector1payload<Type, VectorType>(inval, dim1));
      }

      /// Clone this object.
      payload::const_ptr
      clone() const
      {
        return const_ptr(new vector1payload<Type, VectorType>(data_, dims_[0]));
      }

      /// Access the underlying data in a const context.
      const arraytype&
      data() const
      {
        return data_;
      }

      virtual long
      get_byte_length() const
      {
        int dimsize = 1;
        for (int i = 0; i < dims_.size(); i++)
          dimsize *= dims_[i];
        return sizeof(Type) * dimsize;
      }

      /**
       * Add operator
       * @param other
       * @return a smart pointer to a new valuepayload object
       */
      virtual const payload::const_ptr
      add(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<
            const vector1payload<Type, VectorType> >(other);

        arraytype newval(data_);

        for (int i = 0; i < dims_[0] && i < data_.size(); i++)
          {

            newval[i] = data_[i] + casted->data_[i];

          }

        return construct(newval, dims_[0]);
      }

      /**
       * Less than comparator
       * @param other
       * @return
       */
      virtual bool
      less_than(const payload::const_ptr &other) const
      {
        throw unimplementederror("vectorpayload: less_than");
      }

      /**
       * Greater than comparator
       * @param other
       * @return
       */
      virtual bool
      greater_than(const payload::const_ptr &other) const
      {
        throw unimplementederror("vectorpayload: greater_than");

      }

      /**
       * Equals comparator
       * @param other
       * @return
       */
      virtual bool
      equals(const payload::const_ptr &other) const
      {
        const_ptr casted = boost::dynamic_pointer_cast<
            const vector1payload<Type, VectorType> >(other);

        for (int i = 0; i < dims_[0] && i < data_.size(); i++)
          {
            if (data_[i] != casted->data_[i])
              return false;
          }

        return true;
      }

      /**
       * Serialize this object into the sst_serializer, for
       * running in parallel simulation
       * @param ser the serializer to use
       */
      virtual void
      serialize(sst_serializer* ser) const
      {
        throw unimplementederror("vectorpayload: serialize");
      }

      /**
       * Deserialize a valuepayload out of the serializer
       * @param ser the serializer that contains a valuepayload
       * @return a smart pointer to a new valuepayload object
       */
      static payload::const_ptr
      deserialize(sst_serializer* ser)
      {
        throw unimplementederror("vectorpayload: deserialize");
      }

      /**
       * Strinfier
       * @return a string description
       */
      virtual std::string
      to_string() const
      {
        std::stringstream ss;
        ss << "vector1payload(" << typeid(Type).name() << ", " << dims_[0]
            << ")";
        return ss.str();
      }

      /**
       * The deserialization name, so sst_serializer can reconstruct this
       * @return a string name for this class
       */
      static std::string
      get_deser_name()
      {
        std::stringstream ss;
        ss << "vector1payload(" << typeid(Type).name() << ")";
        return ss.str();
      }
    };



} // end of namespace sstmac
#endif
