/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_HARDWARE_FAILURE_FAILABLE_H_INCLUDED
#define SSTMAC_HARDWARE_FAILURE_FAILABLE_H_INCLUDED

#include <sstmac/common/stats/stats_common.h>

#include <sstmac/hardware/failure/failure_event.h>
#include <math.h>
#include <sstmac/hardware/failure/failure_model.h>

namespace sstmac
{
  class eventmodule;

  namespace hw
  {

    class failable : public virtual ptr_type
    {
    protected:
      /**
       * constructor - blank so inheritance doesn't require that we use the failure model
       */
      failable()
      {
        uid_ = failable_count_++;
        has_failed_ = false;
      }

    public:
      typedef boost::intrusive_ptr<failable> ptr;

      virtual std::string
      to_string() const
      {
        return "failable";
      }

      virtual
      ~failable()
      {
      }

      virtual void
      init_failable(failure_model::ptr mo,
          const boost::intrusive_ptr<eventmodule> &parentmodel);

      virtual void
      failure_occurred(const sst_message::ptr &msg,
          const boost::intrusive_ptr<eventmodule> &parentmodel)
      {
        has_failed_ = true;

        SSTMAC_DEBUG << "component: " << to_string() << " has failed \n";

        model_->propagate();

        model_->next_failure(parentmodel);
      }

      /**
       * use_me - call this to indicate that this thing is being used.
       *   Can be used to update the probability of failure, or to
       *   check if the object has failed
       */
      virtual void
      use_me()
      {

        if (has_failed_)
          {
            uses_while_failed_->record(1);
          }
      }

      /**
       * unique_id - used to hash into the children map
       */
      long
      unique_id() const
      {
        return uid_;
      }

      /**
       * Can be used to get an identifier which can be coupled with a name
       */
      virtual long
      failure_id() const {
        return -1;
      }

      bool
      has_failed()
      {
        return has_failed_;
      }

      void
      reset_failed()
      {
        has_failed_ = false;
      }

      /**
       * parent_failed - this function is called when a parent has failed
       * and it's child has definitely failed
       */
      void
      parent_failed()
      {
        has_failed_ = true;

        SSTMAC_DEBUG << "component: " << to_string()
              << " has failed because of its parent \n";

        model_->propagate();
      }

    protected:
      bool has_failed_;

      //ptr self_;
      long uid_;


      static stat_sum::ptr uses_while_failed_;

      static debug dbg_;
      static long failable_count_;

      failure_model::ptr model_;

    };

  }
}

#endif
