/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_BANDWIDTH_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_BANDWIDTH_H_INCLUDED

#include <sstmac/common/timestamp.h>
#include <sstmac/common/logger.h>
#include <sstmac/hardware/network/congestion/data.h>
#include <iosfwd>
#include <string>
#include <map>

namespace sstmac { namespace hw 
  {

  /**
   * A container to store a fixed-point representation of bandwidth.
   *
   * By default, the bandwidth is stored as a 64-bit integer
   * rather than an arbitrary precision value.  This implementation
   * detail not fully opaque to the external caller although it may be
   * subject to change.   To support builds where the bandwidth is stored
   * as an arbitrary precision values, you should always use bps_t
   * instead of assuming a specific built-in integer type.
   */
  class bandwidth
  {
  public:
#ifndef SSTMAC_USE_GMPXX
    /// The type used to represent bandwidth in bits per second.
    typedef uint64_t bps_t;
#else
    typedef timestamp::tick_t bps_t;
#endif

    /// The units recognized for bandwidth values.
    ///
    /// The units are identical to data units, with
    /// 'ps' appended as an abbreviation for "per second"
    enum units
    {
      bps,
      Kbps,
      Kibps,
      Mbps,
      Mibps,
      Gbps,
      Gibps,
      Tbps,
      Tibps,
      Pbps,
      Pibps,
      Ebps,
      Eibps,
      Bps,
      KBps,
      KiBps,
      MBps,
      MiBps,
      GBps,
      GiBps,
      TBps,
      TiBps,
      PBps,
      PiBps,
      EBps,
      EiBps
    };

  private:
    /// Map string names to bandwidth units.
    typedef std::map<std::string, units> unitmap_t;
    static unitmap_t *unitmap_;

    /// Initialize the unitmap if the unitmap_ pointer is NULL.
    static void init_unitmap();

    /// The input bandwidth (stored in bits per second).
    bps_t bits_per_second_;

    /// A rational representation of bits-per-tick
    /// where the tick frequency is obtained using timestamp::tick_interval().
    /// This value is computed at construction time.
    struct bpt_t
    {
      bps_t numerator;
      bps_t denominator;
    } bits_per_tick_;

    /// Given two bandwidth value types, compute the greatest common divisor.
    inline bps_t gcd(const bps_t &a, const bps_t &b) const;

    /// Private initialization method used by constructors.
    template<typename T>
    void set_bw(T bw, units unit);

  public:
    /// Default constructor.  Creates an object with zero bandwidth.
    bandwidth();

    /// Build a bandwidth object with exactly the given bits per second.
    bandwidth(const bps_t &bitspersecond);

    /// Build a new bandwidth object.
    bandwidth(double bw, units unit);

    /// Alternate convenience way of defining the units (as a string).
    bandwidth(double bw, const std::string &unit);

    /// Define a bandwidth based on a number of bits transferred
    /// and the time it takes to transfer them.
    bandwidth(const data &dt, const timestamp &ts);

    /// Return the exact bandwidth in bits per second.
    const bps_t& bits_per_second() const
    {
      return bits_per_second_;
    }

    /// Return the bandwidth using the given units.
    double get_bandwidth(units unit) const;

    /// A convenience function to specify units using a string value.
    double get_bandwidth(const std::string &unit) const;

    /// Given a quantity of data to transfer, return a timestamp
    /// corresponding to how long the transfer would take.
    timestamp transfer_time(data quantity) const;

    /// Given a tansfer time, report how much data would be transferred
    /// at this bandwidth.
    data transferred_data(const timestamp &ts) const;

    /// Increment bandwidth in place.
    bandwidth& operator+=(const bandwidth &bw);

    /// Decrement bandwidth in place.
    bandwidth& operator-=(const bandwidth &bw);

    /// Scale bandwidth in place.
    bandwidth& operator*=(double scaling);

    /// Divide bandwidth in place.
    bandwidth& operator/=(double scaling);

  };

  bandwidth operator+(const bandwidth &a, const bandwidth &b);
  bandwidth operator-(const bandwidth &a, const bandwidth &b);
  bandwidth operator*(const bandwidth &t, double scaling);
  bandwidth operator*(double scaling, const bandwidth &t);
  bandwidth operator/(const bandwidth &t, double scaling);

  std::ostream& operator<<(std::ostream &os, const bandwidth &t);
  logger& operator<<(logger &os, const bandwidth &t);

  inline bool operator<(const bandwidth &a, const bandwidth &b)
  {
    return (a.bits_per_second() < b.bits_per_second());
  }

  inline bool operator<=(const bandwidth &a, const bandwidth &b)
    {
    return (a.bits_per_second() <= b.bits_per_second());
    }

  inline bool operator>(const bandwidth &a, const bandwidth &b)
  {
    return (a.bits_per_second() > b.bits_per_second());
  }

  inline bool operator>=(const bandwidth &a, const bandwidth &b)
    {
    return (a.bits_per_second() >= b.bits_per_second());
    }

  inline bool operator==(const bandwidth &a, const bandwidth &b)
    {
    return (a.bits_per_second() == b.bits_per_second());
    }

  inline bool operator!=(const bandwidth &a, const bandwidth &b)
    {
    return (a.bits_per_second() != b.bits_per_second());
    }

  inline data operator*(const bandwidth &bw, const timestamp &ts)
  {
    return bw.transferred_data(ts);
  }

  inline data operator*(const timestamp &ts, const bandwidth &bw)
  {
    return bw.transferred_data(ts);
  }

  inline bandwidth operator/(const data &dt, const timestamp &ts)
  {
    return bandwidth(dt, ts);
  }

  inline timestamp operator/(const data &dt, const bandwidth &bw)
  {
    return bw.transfer_time(dt);
  }

  } } // end of namespace sstmac.

#endif
