/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_CIRCUIT_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_CIRCUIT_H_INCLUDED

#include <sstmac/hardware/network/congestion/interconnect.h>
#include <sstmac/hardware/network/congestion/bandwidth.h>
#include <sstmac/hardware/network/congestion/latencymodel.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/stats/stats_common.h>
#include <vector>

#include <sstmac/common/sim_parameters.h>
#include <sstmac/common/units.h>

namespace sstmac { namespace hw 
  {

  // Forward declaration of important types.


  /**
   * Model network traffic by requiring that all network links
   * between the source and destination be simultaneously available.
   * An idealized network performance model is assumed where the
   * total time a message will keep the circuit open is the latency
   * plus the message size divided by the bandwidth.  The network
   * topology and routes are provided by specializations of the
   * network class.
   *
   * This is roughly equivalent to a circuit network with a separate
   * signaling channel.  The signaling channel (managed through circuit::edge)
   * handles requests for a given edge using a FIFO queue; a message is
   * initiated by placing a reservation for all edges connecting a sender
   * and receiver (using circuit::edge::reserve), and upon completion the edge
   * is explicitly released (using circuit::edge::release).
   */
  class circuit: public interconnect
  {
  public:
    /// Timing model.  This may need to be re-thought a bit.
    enum timingmodel
    {
      ORIGINAL, SPLIT
    };

  private:
    /// Private class to keep track of whether an edge is available or not
    /// and to signal a sender when the entire circuit is available.
    class edge;

    /// Private class to handle the actual transmission of a message;
    /// place the request, wait for confirmation from the edges,
    /// "transfer data" (set up completion events for sender and receiver),
    /// and release the edges.
    friend class transfer;
    class transfer;



    /// Reservation queues for each of the edges in our network.
    /// It is the topology's job to determine whether edges
    /// are shared in both directions (half duplex) or not (full duplex).
    std::vector<edge> edge_;



    /// The latency passed to the ctor. This and the bandwidth
    /// are used to construct timing models.
    latencymodel::ptr latency_;

    /// The bandwidth passed to the ctor. This and the latency
    /// are used to construct timing models.
    bandwidth bandwidth_;

    /// The timing model used.  Defaults to 'original'.
    timingmodel timing_model_;




  protected:
    /// Construct a circuit with a custom latency model.
    circuit(const latencymodel::ptr &latency, bandwidth bw,
        const topology::ptr &net,
        const power_config::ptr &pwrcfg);



    /// Use the original timing model.
    void original_timing_model(data payload, timestamp &start_time,
        timestamp &send_time, timestamp &recv_time, timestamp &tran_time);

    /// Use the split timing model.
    void split_timing_model(data payload, timestamp &start_time,
        timestamp &send_time, timestamp &recv_time, timestamp &tran_time);
  public:
    typedef boost::intrusive_ptr<circuit> ptr;

    /// New circuit network with a constant latency.
    /// Legacy interface with double-precision value for bandwidth.
    static circuit::ptr
    construct(const timestamp latency, double bandwidth,
        const topology::ptr &net);

    /// New circuit network with a constant latency.
    static circuit::ptr
    construct(const timestamp latency, bandwidth bw,
        const topology::ptr &net);

    /// New circuit network with a caller-defined latency model - no power config
    static circuit::ptr
    construct(const latencymodel::ptr &latency, bandwidth bw,
        const topology::ptr &net);

    /// New circuit network with a caller-defined latency model.
    static circuit::ptr
    construct(const latencymodel::ptr &latency, bandwidth bw,
        const topology::ptr &net,
        const power_config::ptr &pwrcfg);

    static ptr
    construct(sim_parameters::ptr sim_params, topology::ptr top){
      bool errorflag = false;
      timestamp lat = sim_params->get_time_param("network_latency");
      double bw = get_bandwidth(
          sim_params->get_param("network_bandwidth").c_str(), errorflag);

      return construct(lat, bw, top);
    }

    /// Goodbye.
    virtual ~circuit() throw ();

    /// Required method from parent.
    virtual int64_t slots() const;


    /// Send the given message at the given time.
    /// Notify the sender node when the send is complete
    /// and notify the receiver node when receive needs to be performed.
    virtual void send(const timestamp &start_time,
        const sst_message::ptr &msg,
        const eventhandler::ptr &sendhandler,
        const eventhandler::ptr &recvhandler);

    /// Request original timing model.
    void original_timing_model()
    {
      timing_model_ = ORIGINAL;
    }

    /// Request split timing model.
    void split_timing_model()
    {
      timing_model_ = SPLIT;
    }

    /// Get current timing model.
    timingmodel timing_model() const
    {
      return timing_model_;
    }
  };

  // Debug -- gcc-4.5 is having some real issues with trying to
  // use lexical_cast on circuit.
  inline std::ostream& operator<<(std::ostream &os, const circuit&)
  {
    os << "[class circuit]";
    return os;
  }

  } } // end of namespace sstmac

#endif
