/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_FULLNETWORK_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_FULLNETWORK_H_INCLUDED

#include <sstmac/common/nodeaddress.h>
#include <sstmac/hardware/network/congestion/interconnect.h>
#include <sstmac/hardware/network/congestion/bandwidth.h>
#include <sstmac/hardware/network/congestion/latencymodel.h>

#include <sstmac/common/units.h>
#include <sstmac/common/sim_parameters.h>
#include <map>

namespace sstmac { namespace hw 
  {


  /** The fullnetwork uses an idealized network model that only
 considers the latency and bandwidth. The network is
 completely congestion free. The input for fullnetwork
 can specify the following attributes:

 <table border="1">

 <tr><td>Keyword<td>Type<td>Default<td>Description

 <tr><td><tt>bandwidth</tt><td>double<td>1000000000<td>The network
 bandwidth in bytes per second.

 <tr><td><tt>latency</tt><td>double<td>0.000001<td>The network
 latency in seconds.

 <tr><td><tt>nnode</tt><td>long<td>0<td>The number of nodes in
 the network.  The default of zero means that any number of nodes
 is supported.

 </table>
   */

#define DEFAULT_FULL_NETWORK_SIZE 100

  class fullnetwork: public interconnect
  {
  private:
    /// The latency of the network.
    latencymodel::ptr latency_;
    /// The bandwidth of the network in bytes per second (why not bits?).
    bandwidth bandwidth_;
    /// The maximum size of the network.
    long size_;



  protected:
    /// Construct a fullnetwork object with the given properties
    /// \param latency is the network latency model
    /// \param bandwidth is the network bandwidth (bytes/second)
    /// \param maxsize is the maximum network size (<= 0 for unlimited size).
    fullnetwork(const latencymodel::ptr &latency, bandwidth bw,
        int maxsize, const power_config::ptr &pwrcfg);



  public:
    typedef boost::intrusive_ptr<fullnetwork> ptr;

    /// Construct a full network with fixed latency (legacy interface).
    static fullnetwork::ptr
    construct(const timestamp &latency, double bandwidth, int maxsize=DEFAULT_FULL_NETWORK_SIZE);

    /// Construct a full network with caller-defined latency.
    static fullnetwork::ptr
    construct(const latencymodel::ptr &latency, bandwidth bw, int maxsize=DEFAULT_FULL_NETWORK_SIZE);

    static ptr
    construct(sim_parameters::ptr params){
      bool error = false;
      timestamp lat = params->get_time_param("network_latency");
      double bw = get_bandwidth(params->get_param("network_bandwidth").c_str(), error);

      /** We are no longer allowing infinite interconnects because
          we want to rid ourselves of the evil node_num parameter */
      long size = params->get_optional_long_param("network_size", DEFAULT_FULL_NETWORK_SIZE);

      return construct(lat, bw, size);
    }

    /// Exit stage left.
    virtual ~fullnetwork() throw ();

    /// The maximum number of nodes possible in this simulator.
    /// A value of <= 0 indicates that any number of nodes is possible.
    virtual int64_t slots() const;

    virtual std::string
    to_string() const{
      return "fullnetwork";
    }

    /// Send the given message at the given time.
    /// Notify the sender node when the send is complete
    /// and notify the receiver node when receive needs to be performed.
    virtual void send(const timestamp &start_time,
        const sst_message::ptr &msg,
        const eventhandler::ptr &sendhandler,
        const eventhandler::ptr &recvhandler);
  };

  } } // end of namespace sstmac

#endif
