/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_INTERCONNECT_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_INTERCONNECT_H_INCLUDED

#include <sstmac/common/timestamp.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/ptr_type.h>

#include <sstmac/common/stats/power_config.h>

#include <sstmac/common/debug.h>

#include <sstmac/common/eventmodule.h>

#include <sstmac/hardware/network/topology/topology.h>

#include <sstmac/common/stats/stats_common.h>
#include <sstmac/common/stats/stat_time_window.h>

#include <sstmac/hardware/failure/failable.h>
#include <sstmac/hardware/failure/generic_failure_component.h>

#include <list>

namespace sstmac
{
  namespace hw
  {

    /**
     * Base class for network congestion models.
     */
    class interconnect : virtual public nocopy, public eventmodule
    {
    public:
      typedef boost::intrusive_ptr<interconnect> ptr;
      typedef boost::intrusive_ptr<const interconnect> const_ptr;

      virtual std::string
      to_string() const
      {
        return "interconnect";
      }

      virtual
      ~interconnect()
      {

      }

      /// The maximum number of nodes on this network.
      /// Return a negative (really <= 0) value to indicate an infinite capacity.
      virtual int64_t
      slots() const = 0;

      /// Send the given message at the given time.
      /// Notify the sendhandler when the send is complete.
      /// Notify the recvhandler when the recv is complete.
      virtual void
      send(const timestamp &start_time, const sst_message::ptr &msg,
          const eventhandler::ptr &sendhandler,
          const eventhandler::ptr &recvhandler) = 0;

      virtual topology::ptr
      get_topology()
      {
        return topology_;
      }

      virtual void
      init_failure_model(std::map<vertex_index, failable::ptr> c)
      {
        failuremodel_ = true;
        fail_switches_ = c;
      }

    protected:

      interconnect(topology::ptr t);

      static debug dbg_;

      // stats
      //netstat::ptr nstat;
      // powerstat::ptr pwrstat;

      power_config::ptr pwr_config_;

      static bool registered_stats_;
      static stat_sum::ptr pwrstat_;
      static stat_time_window_sum_duration::ptr energystat_;
      static stat_time_window_sum::ptr bandwidthstat_;
      static stat_time_window_sum::ptr msg_count_;

      std::map<vertex_index, failable::ptr> fail_switches_;
      bool failuremodel_;

      /// The network topology.  This object should eventually take over the
      /// ownership of per-link latencies and bandwidths.
      topology::ptr topology_;

    };

  }
} // end of namespace sstmac

#endif
