/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_NOISYLATENCY_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_NOISYLATENCY_H_INCLUDED

#include <sstmac/hardware/network/congestion/latencymodel.h>
#include <sstmac/common/normaldist.h>
#include <stdint.h>

namespace sstmac { namespace hw  {

  /**
   * A latency model where random (normal or half-normal) noise is added to
   * the default (mean or best-case) latency distribution.
   *
   * Note that this class does some floating point math (somewhat unavoidable).
   */
  class noisylatency : public latencymodel {
    /// The mean or best-case latency.
    timestamp mean_;

    /// The latency variance.
    timestamp variance_;

    /// Are we using a noise (half normal distribution) or jitter model
    /// (normal distribution)?
    bool using_noise_model_;

    /// The random number generator we use to decide the next latency.
    RNG::normaldist::ptr rng_;

    /// \param mean       The best-case latency for noise modeling, or mean
    ///                   latency for jitter models.
    /// \param variance   The variance for half-normal (noise) or normal (jitter)
    ///                   distributed noise.
    /// \param noisemodel Set to true for noise modeling, false for jitter.
    noisylatency(const timestamp &mean, const timestamp &variance,
                 uint64_t rngseed, bool noisemodel);

  public:
    typedef boost::intrusive_ptr<noisylatency> ptr;

    /// \param mean       The best-case latency for noise modeling, or mean
    ///                   latency for jitter models.
    /// \param variance   The variance for half-normal (noise) or normal (jitter)
    ///                   distributed noise.
    /// \param noisemodel Set to true for noise modeling, false for jitter.
    static noisylatency::ptr
    construct(const timestamp &mean, const timestamp &variance,
              uint64_t rngseed, bool noisemodel = true);
    
    /// Done.
    virtual ~noisylatency();

    /// Provide a latency estimate for the next transfer.
    virtual timestamp latency(data) const;
  };

} } // end of namespace sstmac.

#endif
