/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/sharedbuffer.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-arrival.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-departure.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-flowregistration.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-bufferchange.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-bandwidthchange.h>

#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace hw  {

  

  // Make me.
  sharedbuffer::arrival::arrival(sharedbuffer::ptr parent,
                                 flow_id_t fid,
                                 vertex_index v,
                                 double input_bw,
                                 double bytes) :
    parent_(parent), fid_(fid), v_(v), input_bw_(input_bw), bytes_(bytes)
  {}

  // Construct-o-saurus.
  sharedbuffer::arrival::ptr
  sharedbuffer::arrival::construct(sharedbuffer::ptr parent,
                                   flow_id_t fid,
                                   vertex_index v,
                                   double input_bw,
                                   double bytes)
  {
    return arrival::ptr
      (new arrival(parent, fid, v, input_bw, bytes));
  }

  // Kill me.
  sharedbuffer::arrival::~arrival() throw() 
  {}

  // Handle me.
  void sharedbuffer::arrival::handle(const sst_message::ptr &msg)
  {

    // See if we've already arrived at this vertex.
    if(parent_->flow_registrations_[v_.value()].find(fid_) != 
       parent_->flow_registrations_[v_.value()].end()){
        if (parent_->dbg_.is_active())
                        parent_->dbg_ << "arrival::handle: flow " << fid_ << " tring to arrive at vertex " << v_.value() << " twice\n";
      return;
      //throw illformederror("trying to arrive at a link twice");
    }

    if (parent_->dbg_.is_active())
                parent_->dbg_ << "arrival::handle: flow " << fid_ << " has arrived at vertex " << v_.value() << " input bw " << input_bw_ << "\n";

    // construct new flow registration
    flowregistration::ptr fr = flowregistration::construct();
    parent_->flow_registrations_[v_.value()].insert(
        std::pair<flow_id_t, flowregistration::ptr >(fid_, fr));

    fr->buffer_occupancy_ = 0.0;
    fr->last_buffer_update_time_ = parent_->now();
    fr->input_bw_ = input_bw_;
    // timestamp(-1) is a special value indicating that the flow has only just arrived and
    // no estimate has been made as to when it will depart
    fr->expected_departure_time_ = timestamp(-1);
    fr->output_bw_ = 0.0;

    int i;
    const std::vector<edge_index> &edges = parent_->flow_edges_[fid_];

    // handle injection/ejection specially
    if(parent_->topology_->is_node(v_)){
      // injection
      if(parent_->topology_->g().index_to_edge(edges[0]).from().value() == v_.value()){
        fr->output_edge_ = edges[0];
        fr->tail_ = true;
        fr->injection_ = true;

        // assume that entire message is buffered at the nic
        fr->buffer_occupancy_ = bytes_;
      }
      else{ // ejection
        fr->input_edge_ = edges[edges.size() - 1];
        fr->output_edge_ = -1;
        fr->output_bw_ = fr->input_bw_;
        fr->ejection_ = true;
        return; 
      }
    }
    else{
      // find what edge this flow took to this vertex
      for(i = 0; i < edges.size(); i++){
        
        if(parent_->topology_->g().index_to_edge(edges[i]).to() == v_)
          break;
      }

      if(i == edges.size()) {
        throw illformederror("at edge not on path");
      }

      fr->input_edge_ = edges[i];
      fr->output_edge_ = edges[i+1];
    
    }

    // can only proceed if there's a vacant virtual channel on the output port
    if(parent_->available_vcs_[fr->output_edge_.value()] > 0){

      // we're taking one
      parent_->available_vcs_[fr->output_edge_.value()]--;
      fr->has_vc_ = true;

      // schedule bandwidth change to find bw assignment
      bandwidthchange::ptr bwc = bandwidthchange::construct(parent_,
                                                                   v_,
                                                                   fr->output_edge_);
      parent_->schedule(parent_->now(), bwc);

    }
    else{

        if (parent_->dbg_.is_active())
                        parent_->dbg_ << "arrival::handle: flow " << fid_ << " has blocked waiting for a vc\n";

      // record that we are in fact waiting on a virtual channel
      parent_->flows_blocked_on_vc_[fr->output_edge_.value()].push(fid_);

      // if input_bw is positive, this flow will fill its buffer at this router
      // (equivalent to having 0 output bw)
      // add 1e-12 (ie 1 tick) to ensure we go beyond full
      if(fr->input_bw_ > parent_->bw_epsilon_){
        timestamp fill_time(parent_->now() +
            timestamp((parent_->buffer_capacity_ / fr->input_bw_) + 1e-12));
        
        bufferchange::ptr bc = bufferchange::construct(parent_,
                                                              v_,
                                                              fid_);
        parent_->schedule(fill_time, bc);
      }

    }
    
  }

} } //end of namespace sstmac

