/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_SHAREDBUFFER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_SHAREDBUFFER_H_INCLUDED

#include <sstmac/hardware/network/congestion/interconnect.h>

#include <sstmac/hardware/network/congestion/bandwidth.h>
#include <sstmac/hardware/network/topology/edge.h>
#include <map>
#include <vector>
#include <set>
#include <queue>

#include <sstmac/common/units.h>

#include <sstmac/common/sim_parameters.h>

namespace sstmac
{
  namespace hw
  {

    // Important types in the sstmac namespace
    class topology;

    /**
     * A specialization of the interconnect that uses buffers
     * to emulate a packet-like motion of flows through the network.
     */
    class sharedbuffer : public interconnect
    {

    protected:

      /// Nested event type handling arrival of a flow at a particular vertex.
      class arrival;
      /// Nested event type handling departure of a flow from a particular vertex.
      class departure;
      /// Nested type containing all state information about a flow at a particular vertex.
      class flowregistration;
      /// Nested event type handling a flow filling or draining a buffer at a particular vertex.
      class bufferchange;
      /// Nested event type handling any state changes that could affect the bandwidth allocations on a particular edge.
      class bandwidthchange;

      /// Flow ID counter
      long fid_counter_;

      /// Size of each buffer for each link in bytes.
      double buffer_capacity_;

      /// Link bandwidth between two routers in bytes per second.
      double router_router_bw_;
      /// Link bandwidth between a node and a router in bytes per second.
      double node_router_bw_;

      /// Endpoint latency in seconds.
      timestamp node_lat_;
      /// Link latency between two vertices in seconds.
      timestamp link_lat_;
      /// Router latency in seconds.
      timestamp router_lat_;

      /// Link bandwidth epsilon for floating point comparisons.
      double bw_epsilon_;

      /// Buffer epsilon for floating point comparisons.
      double buffer_epsilon_;

      typedef int64_t flow_id_t;

      /// Number of virtual channels per link.
      int num_vcs_;

      typedef std::vector<int> available_vcs_t;
      /// Number of available (ie unoccupied) virtual channelss per link.
      available_vcs_t available_vcs_;

      typedef std::vector<std::queue<flow_id_t> > flows_blocked_on_vc_t;
      /// Queues holding all flows waiting for a virtual channel at each edge.
      flows_blocked_on_vc_t flows_blocked_on_vc_;

      typedef std::vector<
          std::map<flow_id_t, boost::intrusive_ptr<flowregistration> > > flow_registrations_t;
      /// Flow registration map per vertex.
      flow_registrations_t flow_registrations_;

      typedef std::vector<double> edge_bw_t;
      /// Link available bandwidths.
      edge_bw_t edge_bw_;

      /// Link maximum bandwidths.
      edge_bw_t edge_max_bw_;

      typedef std::map<flow_id_t, std::vector<edge_index> > flow_edges_t;
      /// Structure holding flow routes.
      flow_edges_t flow_edges_;

      typedef std::map<flow_id_t, eventhandler::ptr> flow_handlers_t;
      /// Structure holding flow receive handlers.
      flow_handlers_t recv_handlers_;
      /// Structure holding flow send handlers.
      flow_handlers_t send_handlers_;

      typedef std::map<flow_id_t, sst_message::ptr> recv_messages_t;
      recv_messages_t recv_messages_;

      typedef std::vector<std::set<flow_id_t> > vertices_departed_t;
      /// Set maintaining which flows have departed which vertices.
      vertices_departed_t vertices_departed_;

      /// Construct a new buffered interconnect.
      sharedbuffer(double router_router_bw, double node_router_bw,
          const timestamp node_lat, const timestamp link_lat,
          const timestamp router_lat, double buffer_capacity, int num_vcs,
          const topology::ptr &net);

      /// Update the buffer occupancy of the flow with registration fr.
      /// Occupancy is affected by the difference between input and output rates
      /// and last time since the buffer occupancy was updated.
      void
      update_buffer_usage(boost::intrusive_ptr<flowregistration> fr,
          const timestamp &now);

      /// Tests to see if two doubles are equal to within the buffer epsilon.
      bool
      are_equal(double a, double b);

    public:
      typedef boost::intrusive_ptr<sharedbuffer> ptr;

      /// Construct a new buffered interconnect for real.
      static sharedbuffer::ptr
      construct(double router_router_bw, double node_router_bw,
          const timestamp node_lat, const timestamp link_lat,
          const timestamp router_lat, double buffer_capacity, int num_vcs,
          const topology::ptr &net);

      static ptr
      construct(sim_parameters::ptr params, topology::ptr top)
      {
        bool errorflag = false;
        double router_router_bw = get_bandwidth(
            params->get_param("network_bandwidth_r2r").c_str(), errorflag);
        double node_router_bw = get_bandwidth(
            params->get_param("network_bandwidth_n2r").c_str(), errorflag);
        timestamp node_lat = params->get_time_param("network_latency_node");
        timestamp link_lat = params->get_time_param("network_latency_link");
        timestamp router_lat = params->get_time_param("network_latency_router");
        double buffer_capacity = params->get_double_param("network_buffer");
        int num_vcs = params->get_int_param("network_vcs");
        return sharedbuffer::construct(router_router_bw, node_router_bw,
            node_lat, link_lat, router_lat, buffer_capacity, num_vcs, top);
      }

      /// Destructinator.
      virtual
      ~sharedbuffer() throw ();

      /// Required method from parent.
      virtual int64_t
      slots() const;

      /// Send the given message at the given time.
      /// Notify the sender node when the message is put on the line
      /// and notify the receiver when the message is ready to be picked up.
      virtual void
      send(const timestamp &start_time, const sst_message::ptr &msg,
          const eventhandler::ptr &sendhandler,
          const eventhandler::ptr &recvhandler);

      virtual std::string
      to_string() const
      {
        return "sharedbuffer";
      }

    };

  }
} // end of namespace sstmac

#endif
