/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_SHAREDCIRCUIT_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_SHAREDCIRCUIT_H_INCLUDED

#include <sstmac/hardware/network/congestion/interconnect.h>
#include <sstmac/hardware/network/congestion/bandwidth.h>

#include <sstmac/common/stats/power_config.h>
#include <sstmac/hardware/network/congestion/flower.h>
#include <sstmac/hardware/network/congestion/latencymodel.h>

#include <sstmac/common/sim_parameters.h>
#include <sstmac/common/units.h>

namespace sstmac
{
  namespace hw
  {

    /**
     * A specialization of circuit that shares circuit bandwidth between
     * concurrent transfer requests.
     * */

    class sharedcircuit : public interconnect
    {
      /// Private nested type to handle stopping flows and notifying listeners.
      class stopper;
      /// Private nested type to schedule starting a flow.
      class starter;

      /// The flower handles the actual sharing of the circuit.
      flower::ptr flower_;
      /// The sharedcircuit has exactly one stopper.
      boost::intrusive_ptr<stopper> stopper_;
      /// Network latency.
      latencymodel::ptr latency_;
      /// Network bandwidth.
      bandwidth bandwidth_;

    protected:
      /// Construct a new circuit.
      sharedcircuit(const latencymodel::ptr &latency, bandwidth bw,
          const topology::ptr &net, const flower::ptr &,
          const power_config::ptr &pwrcfg);

    public:
      typedef boost::intrusive_ptr<sharedcircuit> ptr;

      /// Construct a new circuit with constant latency - no power config
      static sharedcircuit::ptr
      construct(const timestamp latency, double bandwidth,
          const topology::ptr &net, const flower::ptr &,
          const power_config::ptr &pwrcfg);

      /// Construct a new circuit with constant latency.
      static sharedcircuit::ptr
      construct(const timestamp latency, double bandwidth,
          const topology::ptr &net, const flower::ptr &);

      /// Construct a new circuit with user-defined latency modeling.
      static sharedcircuit::ptr
      construct(const latencymodel::ptr &latency, bandwidth bw,
          const topology::ptr &net, const flower::ptr &,
          const power_config::ptr &pwrcfg);

      static ptr
      construct(sim_parameters::ptr sim_params, topology::ptr top);

      /// So long, Mr. Despair.
      virtual
      ~sharedcircuit() throw ();

      /// Required method from parent.
      virtual int64_t
      slots() const;

      /// Send the given message at the given time.
      /// Notify the sender node when the send is complete
      /// and notify the receiver node when receive needs to be performed.
      virtual void
      send(const timestamp &start_time, const sst_message::ptr &msg,
          const eventhandler::ptr &sendhandler,
          const eventhandler::ptr &recvhandler);

    };

  }
} // end of namespace sstmac

#endif
