/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/typical_flower.h>
#include <sstmac/common/errors.h>

#include <boost/lexical_cast.hpp>
#include <iterator>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <limits>
#include <stdlib.h>
#include <cmath>

namespace sstmac { namespace hw 
{


using boost::lexical_cast;
template<typename T>
std::string strfy(T value)
{
        return lexical_cast<std::string>(value);
}

typical_flower::typical_flower(const topology::ptr &net,
                double node_router_bw, double router_router_bw) :
                flower(net), next_flow_id_(flower::initial_flow_id), stopping_flow_(0), last_tval_(0), last_bw_(
                                0), aggregate_bw_(0)
{
        long nedge = net->g().n_edges();
        edge_bw_.resize(nedge);
        edge_flows_.resize(nedge);
        edge_bw_utilization_.resize(nedge);

        for (int i = 0; i < nedge; i++)
        {
                edge_index ei(i);
                const lwedge &e = net->g().index_to_edge(ei);
                if (net->is_node(e.from()) || net->is_node(e.to()))
                {
                        edge_bw_[i] = node_router_bw;
                }
                else
                {
                        edge_bw_[i] = router_router_bw;
                }
                edge_bw_utilization_[i] = 0.0;
        }

        max_bw_ = std::max(node_router_bw, router_router_bw);

        bw_epsilon_ = max_bw_ * 10 * std::numeric_limits<double>::epsilon();
}

timestamp typical_flower::next_stopping_time()
{
        if (stopping_flow_ == 0)
        {
                throw illformederror("no next stopping flow");
        }
        return stopping_time_;
}

flow_id_t typical_flower::next_stopping_flow()
{
        return stopping_flow_;
}

void typical_flower::advance_time(timestamp tnew)
{
        timestamp deltat = tnew - t_;
        t_ = tnew;

        for (flow_bw_t::iterator i = flow_bw_.begin(); i != flow_bw_.end(); i++)
        {
                flow_id_t f = i->first;
                double bw = i->second;
                double &nbyte = flow_nbyte_[f];
                nbyte = nbyte - deltat.sec() * bw;
                if (nbyte < 0)
                        nbyte = 0;
        }
}

void typical_flower::find_next_stop()
{
        SSTMAC_DEBUG << "typical_flower::find_next_stop: entered\n";
        stopping_flow_ = 0;
        for (flow_bw_t::iterator i = flow_bw_.begin(); i != flow_bw_.end(); i++)
        {
                flow_id_t f = i->first;
                double bw = i->second;
                SSTMAC_DEBUG << "  f = " << f << ", bw = " << bw;
                if (bw > 0)
                {
                        double nbyte = flow_nbyte_[f];
                        timestamp stopping_time = t_ + timestamp(nbyte / bw);
                        SSTMAC_DEBUG << ", stopping_time = " << stopping_time;
                        if (stopping_flow_ == 0 || stopping_time_ > stopping_time)
                        {
                                stopping_flow_ = f;
                                stopping_time_ = stopping_time;
                        }
                }
                SSTMAC_DEBUG << "\n";
        }
        if (stopping_flow_ == 0 && flow_bw().size() > 0)
        {
                throw illformederror("flows exist but next stopping flow not found");
        }
        SSTMAC_DEBUG << "  stopping_flow = " << stopping_flow_ << "\n";
}

#ifdef TYPICAL_FLOWER_USE_OLD_EDGE_BW_UTIL
double
typical_flower::edge_bw_utilization(edge_id_t e)
{
        double bw = 0.0;
        double max_edge_bw = edge_bw_[e];
        const std::set<flow_id_t> &flows = edge_flows_[e];
        for (std::set<flow_id_t>::const_iterator i = flows.begin();
                        i != flows.end(); i++)
        {
                bw += flow_bw_[*i];
        }
        if (bw<edge_bw_utilization_[e] - bw_epsilon_
                        ||bw>edge_bw_utilization_[e] + bw_epsilon_)
        {
                std::cout << "computed bw = " << bw
                << " stored bw = " << edge_bw_utilization_[e]
                << std::endl;
                for (std::set<flow_id_t>::const_iterator i = flows.begin();
                                i != flows.end(); i++)
                {
                        std::cout << "flow " << *i << " is " << flow_bw_[*i] << std::endl;
                }
                throw illformederror("edge_bw_utilization inconsistent");
        }
        return bw;
}
#endif // TYPICAL_FLOWER_USE_OLD_EDGE_BW_UTIL
void typical_flower::set_flow_bw(flow_id_t fid, double bw)
{
        // Change the bw for flow fid, creating a new flow_bw_ entry, if needed
        flow_bw_t::iterator current_flow_bw = flow_bw_.find(fid);
        double delta_bw;
        if (current_flow_bw == flow_bw_.end())
        {
                delta_bw = bw;
                flow_bw_[fid] = bw;
        }
        else
        {
                delta_bw = bw - current_flow_bw->second;
                current_flow_bw->second = bw;
        }

        if (fabs(delta_bw) <= bw_epsilon_)
                return;

        /*
         // hhack -- track changes in bandwidth usage.
         aggregate_bw_ += delta_bw;
         if(last_tval_ < t_) {
         std::cout << t_ << "\t" << last_bw_ << " # previous aggregate bw\n";
         std::cout << t_ << "\t" << aggregate_bw_ << " # aggregate bandwidth after adding " << delta_bw << "\n";
         last_tval_ = t_;
         last_bw_ = aggregate_bw_;
         }
         */

        bool edge_bw_exceeded = false;

        // Apply delta_bw to all the edges being used by the flow.
        const std::vector<edge_id_t> &edges = flow_edges_.at(fid);
        for (int i = 0; i < edges.size(); i++)
        {
                edge_bw_utilization_[edges[i]] += delta_bw;
                if (edge_bw_utilization_[edges[i]] > edge_bw_[edges[i]] + bw_epsilon_)
                {
                        edge_bw_exceeded = true;
                }
        }

        if (edge_bw_exceeded)
        {
                throw illformederror("edge_bw_ exceeded in set_flow_bw");
        }
}

void typical_flower::add_to_flow_bw(flow_id_t fid, double delta_bw)
{
        flow_bw_.at(fid) += delta_bw;

        // Apply delta_bw to all the edges being used by the flow.
        const std::vector<edge_id_t> &edges = flow_edges_.at(fid);
        for (int i = 0; i < edges.size(); i++)
        {
                edge_bw_utilization_[edges[i]] += delta_bw;
        }
}

void typical_flower::scale_flow_bw(flow_id_t fid, double scaling_factor)
{
        double &bw = flow_bw_.at(fid);
        double newbw = bw * scaling_factor;
        double delta_bw = newbw - bw;
        bw = newbw;

        bool edge_bw_exceeded = false;

        // Apply delta_bw to all the edges being used by the flow.
        const std::vector<edge_id_t> &edges = flow_edges_.at(fid);
        for (int i = 0; i < edges.size(); i++)
        {
                edge_bw_utilization_[edges[i]] += delta_bw;
                if (edge_bw_utilization_[edges[i]] > edge_bw_[edges[i]] + bw_epsilon_)
                {
                        edge_bw_exceeded = true;
                }
        }

        if (edge_bw_exceeded)
        {
                throw illformederror("edge_bw_ exceeded in set_flow_bw");
        }
}

double typical_flower::get_flow_bw(flow_id_t fid)
{
        return flow_bw_.at(fid);
}

void typical_flower::erase_flow_bw(flow_id_t fid)
{
        set_flow_bw(fid, 0.0);
        flow_bw_.erase(fid);
        SSTMAC_DEBUG << "erase_flow_bw: erasing flow " << fid << "\n";
}

double typical_flower::edge_bw_equal_and_higher_priority_utilization(
                edge_id_t e, flow_id_t f)
{
        SSTMAC_DEBUG
                                << "typical_flower::edge_bw_equal_and_higher_priority_utilization: e = "
                                << e << ", f = " << f << ", nflow = " << edge_flows_[e].size()
                                << ", bw = ";
        double bw = 0.0;
        const std::set<flow_id_t> &flows = edge_flows_[e];
        for (std::set<flow_id_t>::const_iterator i = flows.begin();
                        i != flows.end(); i++)
                        {
                // See if this (and, thus, all remaining) flows are lower
                // priority (higher flow_id). If so, break out of loop.
                if (*i > f)
                        break;
                SSTMAC_DEBUG << "+" << get_flow_bw(*i) << "(" << *i << ")";
                bw += get_flow_bw(*i);
        }
        SSTMAC_DEBUG << "\n";
        return bw;
}

flow_id_t typical_flower::start_flow(timestamp tstart,
                const std::vector<edge_id_t> &edges, long nbyte)
{
        if (tstart < t_)
        {
                throw illformederror("inserting a flow at an earlier time");
        }

        advance_time(tstart);

        flow_id_t current_flow_id = next_flow_id_++;

        flow_edges_[current_flow_id] = edges;
        if (flow_edges_.find(current_flow_id) == flow_edges_.end())
        {
                throw illformederror(
                                "flow::start_flow: flow_edges_ missing current_flow_id");
        }

        for (unsigned long i = 0; i < edges.size(); i++)
        {
                SSTMAC_DEBUG << "adding flow " << current_flow_id << " to edge "
                                        << edges[i] << "\n";
                edge_flows_[edges[i]].insert(current_flow_id);
        }

        flow_nbyte_[current_flow_id] = nbyte;

        set_flow_bw(current_flow_id, 0.0);

        rebalance_flow_start(current_flow_id);

        find_next_stop();

        SSTMAC_DEBUG << "Started flow " << current_flow_id << "\n";
        return current_flow_id;
}

void typical_flower::stop_flow(timestamp tstop, flow_id_t f)
{
        SSTMAC_DEBUG << "Stopping flow " << f << "\n";
        if (tstop < t_)
        {
                throw illformederror("stopping a flow at an earlier time");
        }
        if (flow_edges_.find(f) == flow_edges_.end())
        {
                throw illformederror(
                                std::string("stopping flow number ") + strfy(f)
                                                + " that does not exist");
        }

        advance_time(tstop);

        const std::vector<edge_id_t> &edges = flow_edges_[f];

        for (unsigned long i = 0; i < edges.size(); i++)
        {
                SSTMAC_DEBUG << "removing flow " << f << " from edge " << edges[i]
                                        << "\n";
                edge_flows_[edges[i]].erase(f);
                SSTMAC_DEBUG << "  edge " << edges[i] << " erasing flow " << f
                                        << "\n";
        }

        erase_flow_bw(f);
        flow_edges_.erase(f);
        flow_nbyte_.erase(f);
        SSTMAC_DEBUG << "erasing flow " << f << "\n";

        rebalance_flow_stop(f);

        find_next_stop();

}

} } // end of namespace sstmac
