/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// graph.cc: Implementation of generic graphs.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <iostream>
#include <stdexcept>
#include <iomanip>
#include <stdlib.h>
#include <sstmac/hardware/network/topology/graph.h>

using std::logic_error;
using std::ostream;
using std::cout;
using std::endl;
using std::set;
using std::vector;
using std::map;

namespace sstmac { namespace hw  {

graph::graph()
{
}

graph::graph(const std::set<vertex> &v, const std::set<edge> &e)
{
  init(v,e);
}

void
graph::init(const std::set<vertex> &v, const std::set<edge> &e)
{
  int index;

  in_.resize(v.size());
  out_.resize(v.size());
  ie_.resize(e.size());
  iv_.resize(v.size());
  uei_.clear();
  uvi_.clear();

  // initialize the mapping of vertices to integers
  index = 0;
  for (set<vertex>::iterator i = v.begin();
       i != v.end();
       i++, index++) {
      uvi_[*i] = index;
      iv_[index] = *i;
    }

  // initialize the mapping of edges to integers,
  // converting to lwedge objects in the process
  index = 0;
  for (set<edge>::iterator i = e.begin();
       i != e.end();
       i++, index++) {
      vertex_index v1 = uvi_[i->from()];
      vertex_index v2 = uvi_[i->to()];
      lwedge lwe(v1,v2,i->id());
      uei_[lwe] = index;
      ie_[index] = lwe;
  }

  // initialize the adjacency information
  for (edge_map::iterator i = uei_.begin();
       i != uei_.end();
       i++) {
      out_[i->first.from().value()].push_back(i->first.to());
      in_[i->first.to().value()].push_back(i->first.from());
    }
}

int64_t
graph::in_degree() const
{
  int64_t d = 0;
  for (int64_t i=0; i<in_.size(); i++) {
      if (in_[i].size() > d) d = in_[i].size();
    }
  return d;
}

int64_t
graph::out_degree() const
{
  int64_t d = 0;
  for (int64_t i=0; i<out_.size(); i++) {
      if (out_[i].size() > d) d = out_[i].size();
    }
  return d;
}

int64_t
graph::degree() const
{
  int64_t in = in_degree();
  int64_t out = out_degree();
  return in>out?in:out;
}

void
graph::print(ostream &o) const
{
  o << "order = " << order() << endl;
  o << "n_edge = " << uei_.size() << endl;
}

void
graph::print_all(ostream &o, bool verbose) const
{
  // Outputting sequentially is a little more expensive
  // now that we're using unorder maps

  int index;
  o << "Vertices:" << endl;
  index = 0;
  for (std::vector<vertex>::const_iterator i = iv_.begin();
       i != iv_.end();
       i++, index++) {
      if (verbose) {
        o << "  " << std::setw(4) << index << ":";
      }
      o << "  " << *i << endl;
    }

  o << "Edges:" << endl;
  index = 0;
  for (std::vector<lwedge>::const_iterator i = ie_.begin();
       i != ie_.end();
       i++, index++) {
      o << "  ";
      if (verbose) {
        o << std::setw(4) << index;
        o << ":  ";
      }
      else {
        o << *i << endl;
      }
  }
}

void
graph::print_dot(ostream &o) const
{
  // Outputting sequentially is a little more expensive
  // now that we're using unorder maps
  o << "digraph gamma {" << endl;
  for (std::vector<lwedge>::const_iterator i = ie_.begin();
       i != ie_.end();
       i++) {
      o << "n" << i->from().value()
        << " -> "
        << "n" << i->to().value() << ";" << endl;
  }
  o << "}" << endl;
}

void
graph::check_walk(const vertex &v0, const vertex &v1, const walk &w) const
{
  if (w.length() == 0) {
      if (v0 != v1) {
          cout << "bad empty walk from " << v0 << " to " << v1 << endl;
          throw std::runtime_error("bad empty walk");
        }
      return;
    }

  for (int i=0; i<w.length(); i++) {
      if (uei_.find(w.e(i)) == uei_.end()) {
          cout << "bad edge " << i << " in walk from "
               << v0 << " to " << v1 << ": ";
          if (w.evec().size() > 0) {
            cout << index_to_vertex(w.evec()[0].from());
            for (int i=0; i<w.evec().size(); i++) {
              cout << " (" << w.evec()[i].id()
                   << ")-> " << index_to_vertex(w.evec()[i].to());
            }
          }
          cout << endl;
          throw std::runtime_error("bad edge in walk");
        }
    }
}

int64_t
graph::word_length() const
{
  if (iv_.size() == 0) return 0;
  return (*iv_.begin()).v().size();
}

vertex_index
graph::vertex_to_index(const vertex &v) const
{
  vertex_map::const_iterator i = uvi_.find(v);
  if (i==uvi_.end()) {
      throw logic_error("graph::vertex_to_index: vertex not in map");
    }
  return (*i).second;
}

edge_index
graph::edge_to_index(const lwedge &e) const
{
  edge_map::const_iterator i = uei_.find(e);
  if (i==uei_.end()) {
      throw logic_error("graph::edge_to_index: edge not in map");
    }
  return (*i).second;
}

} } //end of namespace sstmac
