/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// hypercube.cc: Implementation of hypercube networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#include <stdexcept>
#include <sstream>
#include <sstmac/hardware/network/topology/hypercube.h>

#include <sstmac/hardware/network/topology/walk.h>

using std::set;
using std::vector;
using std::string;

namespace sstmac { namespace hw  {

  static void
  compute_vertices(int d, set<vertex> &vertices)
  {
    vertex v(1);
    int n = 1<<d;
    for (int i=0; i<n; i++) {
      v[0] = i;
      vertices.insert(v);
    }
  }

  static void
  compute_edges(int d, set<edge> &edges)
  {
    int n = 1<<d;
    vertex vi(1), vo(1);
    for (int i=0; i<n; i++) {
      vi[0] = i;
      for (int j=0; j<d; j++) {
        int o = i ^ (1<<j);
        vo[0] = o;
        edges.insert(edge(vi,vo));
      }
    }
  }

  // static sstmac::xmlclass::register_class reg("sstmac::hypercube",
  //                               sstmac::xmlclass::ctor_template<hypercube>);

  // hypercube::hypercube(TiXmlHandle h)
  // {
  //   int d;
  //   h.ToElement()->QueryIntAttribute("d",&d);
  //   init_hypercube(d);
  // }

  hypercube::hypercube(int d)
  {
    init_hypercube(d);
  }

  void
  hypercube::init_hypercube(int d)
  {
    d_ = d;
    compute_vertices(d_, nodes_);

    set<edge> edges;
    compute_edges(d_, edges);

    g_.init(nodes_, edges);
    init();
  }

  hypercube::ptr hypercube::construct(int dimension) {
    return hypercube::ptr(new hypercube(dimension));
  }

  hypercube::~hypercube() {
  }

  void
  hypercube::shortest_path(const vertex &v1, const vertex &v2, walk &w) const
  {
    vector<lwedge> edges;

    vertex current(1), next(1);
    current[0] = v1[0];
    for (int i=0; i<d_; i++) {
      if ((current[0] & (1<<i)) != (v2[0] & (1<<i))) {
        next[0] = current[0] ^ (1<<i);
        edges.push_back(g().generate_lwedge(current,next));
        current = next;
      }
    }

    w.init(edges);
  }

  int64_t
  hypercube::diameter() const
  {
    return d_;
  }

  double
  hypercube::a() const
  {
    return d_/2.0;
  }

  int64_t
  hypercube::b() const
  {
    return 1<<d_;
  }

  bool
  hypercube::vertex_symmetric() const
  {
    return true;
  }

  string
  hypercube::name() const
  {
    std::ostringstream ostr;
    ostr << "Hypercube" << "(" << d_ << ")";
    return ostr.str();
  }

} } //end of namespace sstmac
