/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// torus.h: Interface for torus networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TORUS_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_TORUS_H_INCLUDED

#include <sstmac/hardware/network/topology/product.h>
#include <vector>

namespace sstmac { namespace hw  {

  /** Implements a 2D torus network as the product of two rings and
      a 3D torus as the product of 2D torus and ring.

      The XML input for the gamma network uses the following attributes:

      <table border="1">

      <tr><td>Keyword<td>Type<td>Default<td>Description

      <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
      dimension.

      <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
      dimension.

      <tr><td><tt>n3</tt><td>int<td>0<td>The number of nodes in the third
      dimension.  If this is zero, then a 2D torus will be constructed using
      only n1 and n2.

      </table>
  */
  class torus: public product {
  public:
    int width;

  private:
    std::string name_;
    bool direct_;
    int nndpersw_;
    // Helgi hack.
    std::vector<int> dimensions_;

    void init_torus_2D(int x, int y, int nndpersw, int widtharg);
    void init_torus_3D(int x, int y, int z, int nndpersw, int widtharg);

    bool is_switch(const vertex_index &vi) const {
      if (direct_) return true;
      const vertex &v = g().index_to_vertex(vi);
      if (v[1] == nndpersw_) return true;
      return false;
    }
  protected:
    torus();

  public:
    typedef boost::intrusive_ptr<torus> ptr;

    /** Construct a 2D torus.
        @param n1 The number of switches in the first dimension.
        @param n2 The number of switches in the second dimension.
        @param nndpersw The number of nodes per switch.
        @param widtharg The width of edges connecting two switches
        If the number of nodes per switch is zero, then a direct
        network is used. Otherwise, the given number of nodes
        is attached to the switch, each through their own link.
        The width of edges between a node and a switch is always one.
    */
    static torus::ptr construct_2D(int n1=4, int n2=4,
                                                 int nndpersw=0,
                                                 int widtharg=1);
    /** Construct a 3D torus.
        @param n1 The number of switches in the first dimension.
        @param n2 The number of switches in the second dimension.
        @param n3 The number of switches in the third dimension.
        @param nndpersw The number of nodes per switch.
        @param widtharg The width of edges connecting two switches
        If the number of nodes per switch is zero, then a direct
        network is used. Otherwise, the given number of nodes
        is attached to the switch, each through their own link.
        The width of edges between a node and a switch is always one.
    */
    static torus::ptr construct_3D(int n1=4, int n2=4, int n3=4,
                                                 int nndpersw=0,
                                                 int widtharg=1);
    /** Return the width of the given edge. Edges involving nodes
        always have a width of one. The width of edges between two
        switches is specified in the constructor or by setting the
        public width member. */
    int edge_width(edge_index e) const;

    virtual ~torus();
    std::string name() const;

    /* Functions added by Helgi (hacks) 08/10/09. */

    /// Get the dimensions of the torus.
    const std::vector<int>& dimensions() const {
      return dimensions_;
    }
  };

} } //end of namespace sstmac

#endif
