/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// walk.h: Interface for walks through a graph.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_WALK_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_WALK_H_INCLUDED

#include <sstmac/hardware/network/topology/edge.h>

namespace sstmac { namespace hw  {

/// A walk is an ordered collection of edges representing a walk through a
/// graph.
class walk {
    std::vector<lwedge> e_;
  public:
    walk() {}
    walk(const std::vector<vertex_index> &v) { init(v); }
    walk(const std::vector<lwedge> &e) { init(e); }
    /// Initialize the walk by forming edges connecting each
    /// of the given vectors, in order.
    void init(const std::vector<vertex_index> &v);
    /// Intialize the walk to be the given edges.
    void init(const std::vector<lwedge> &e) { e_ = e; }
    /// Return the i'th vertex in the walk. The number of vertices is the
    /// one plus the number of edges (as returned by the length()
    /// member), unless there are no edges, in which case there are no
    /// vertices.
    const vertex_index &v(int i) const;
    /// Return the i'th edge in the walk.
    const lwedge &e(int64_t i) const { return e_[i]; }
    /// Return the i'th edge in the walk.
    lwedge &e(int64_t i) { return e_[i]; }
    /// Return the vector of edges that form this walk.
    const std::vector<lwedge> &evec() const { return e_; }
    /// Return the vector of edges that form this walk.
    std::vector<lwedge> &evec() { return e_; }
    /// Return the number of edges in the walk.
    int64_t length() const { return e_.size(); }
    /// Make this the empty walk.
    void clear() { e_.clear(); }
    /// Append the given edge to the list of edges in this walk.
    void append(const lwedge &e) { e_.push_back(e); }
    /// Print this walk. */
    void print(std::ostream &o = std::cout) const;
};

/// Print the walk to the output stream.
std::ostream& operator<<(std::ostream &o, const walk&w);

} } //end of namespace sstmac

#endif
