/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/node/node.h>
#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/common/driver_util.h>

#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/services/servdescr.h>

namespace sstmac
{
  namespace hw
  {

    debug node::dbg_;
    sstmac::sw::allocationstrategy::ptr node::allocation_;
      
      bool node::has_failed() {
          return nic_->has_failed();
      }
      void node::reset_failed(){
          nic_->reset_failed();
      }

    node::node(const sim_parameters::ptr &params) :
        params_(params)
    {
      sstmac::sw::hardware_descriptor::ptr desc =
          sstmac::sw::hardware_descriptor::construct(1);
      os_ = sstmac::sw::operatingsystem::construct(desc);

      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> node");
          logger::register_logger(&dbg_);
        }
    }

    void
    node::set_nic(const networkinterface::ptr &n)
    {
      nic_ = n;
    }

    boost::intrusive_ptr<networkinterface>
    node::get_nic() const
    {
      return nic_;
    }

    void
    node::initialize(nodeaddress::ptr my_id, eventmanager::ptr man)
    {
      SSTMAC_DEBUG << "starting node initialization \n";
      eventmodule::set_eventmanager(man);

      nodeid_ = my_id;
      msg_handler_ = message_handler::construct(ptr(this));

      std::list<sstmac::sw::launchinfo::ptr> linfos = construct_launchers(
          params_);
      SSTMAC_DEBUG << "launchers constructed \n";

      if (!nic_)
        {
          throw valueerror("node::initialize: nic has not been set yet");
        }
      topology::ptr top = nic_->get_topology();
      launcher_ = instantiate_launcher(params_, top, linfos, allocation_);
      os_->register_ser(launcher_);

      ptr self(this);
      SSTServDesc_build::start_services(self);

      SSTMAC_DEBUG << "done with node initialize \n";

    }

    void
    node::handle(const sst_message::ptr &msg)
    {
      msg_handler_->handle(msg);
    }

    void
    node::add_service(sstmac::sw::service::ptr serv){
      SSTMAC_DEBUG << "adding service called " << serv->to_string() << " at node " << nodeid_->to_string() << "\n";
      os_->register_ser(serv);
      serv->start();
    }

  }
} // end of namespace sstmac

