/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/errors.h>
#include <sstmac/hardware/processor/instructionprocessor.h>
#include <sstmac/common/messages/eventdata.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>

#include <iostream>

namespace sstmac
{
  namespace hw
  {

    //
    // Constructor
    //
    instructionprocessor::instructionprocessor(
        const instructionprocessor_config::ptr& cfg,
        const eventhandler::ptr &completion, computescheduler::ptr sched) :
        processor(cfg, completion, sched), config_(cfg)
    {
      flop_label_ = config_->flop_label();
      memop_label_ = config_->memop_label();
      memop_local_label_ = config_->memop_local_label();
      memop_rand_label_ = config_->memop_rand_label();
      tflop_ = config_->tflop();
      tmem_ = config_->tmemop();
      tmemloc_ = config_->tmemop_local();
      tmemrnd_ = config_->tmemop_rand();

      if(stat_pwr_->is_active())
        stat_pwr_->record(config_->get_pwr_static());

      // debug report
      if (dbg_.is_active())
        {
          dbg_ << "instructionprocessor: initializing\n";
          dbg_ << "  average flop service time:   " << tflop_ << " ns\n";
          dbg_ << "  average memory service time: " << tmem_ << " ns\n";
        }
    }

    //
    // Construct a new processor with the given average service times
    //
    instructionprocessor::ptr
    instructionprocessor::construct(const instructionprocessor_config::ptr& cfg,
        const eventhandler::ptr &completion, computescheduler::ptr sched)
    {
      instructionprocessor::ptr retval(
          new instructionprocessor(cfg, completion, sched));

      return retval;
    }

    processor::ptr
    instructionprocessor::clone() const
    {

      return construct(config_, completion_, csched_);

    }

    void
    instructionprocessor::compute(const sst_message::ptr& inst, long thread_id)
    {
      SSTMAC_DEBUG << "instructionprocessor::compute(" << inst << ")\n";
    eventdata::ptr data = boost::dynamic_pointer_cast < eventdata > (inst);

    // compute execution time in seconds
    timestamp time, zero(0);
    if (tflop_ > zero)
      {
        if (flop_label_.size() == 0)
          throw valueerror("missing flop data label");
        uint64_t nop = data->get_event(flop_label_);
        timestamp thistime = nop * tflop_;
        if (thistime.sec() > 0 && stat_energy_->is_active()){
          stat_energy_->record(csched_->now(), nop * config_->get_pwr_flop(), thistime);
          stat_proc_->record(csched_->now(), nop);
        }
        time += thistime;
      }
    if (tmem_ > zero)
      {
        if (memop_label_.size() == 0)
          throw valueerror("missing memory op data label");
        uint64_t nop = data->get_event(memop_label_);
        timestamp thistime = nop * tmem_;
        if (thistime.sec() > 0 && stat_energy_->is_active())
          stat_energy_->record(csched_->now(), nop * config_->get_pwr_memop(), thistime);
        time += thistime;
      }
    if (tmemloc_ > zero)
      {
        if (memop_local_label_.size() == 0)
          throw valueerror("missing local memory op data label");
        uint64_t nop = data->get_event(memop_local_label_);
        timestamp thistime = nop * tmemloc_;
        if (thistime.sec() > 0 && stat_energy_->is_active())
          stat_energy_->record(csched_->now(), nop * config_->get_pwr_memop_local(), thistime);
        time += thistime;
      }
    if (tmemrnd_ > zero)
      {
        if (memop_rand_label_.size() == 0)
          throw valueerror("missing random memory op data label");
        uint64_t nop = data->get_event(memop_rand_label_);
        timestamp thistime = nop * tmemrnd_;
        if (thistime.sec() > 0 && stat_energy_->is_active())
          stat_energy_->record(csched_->now(), nop * config_->get_pwr_memop_rand(), thistime);
        time += thistime;
      }

    SSTMAC_DEBUG << "instructionprocessor: execution time: " << time.sec() * 1e6
          << "us\n";
    request_compute(time, inst, thread_id);

  }

}
} // end of namespace sstmac

