/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_ACTOR_SIMPLEMAILSERVER_MBOX_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_ACTOR_SIMPLEMAILSERVER_MBOX_H_INCLUDED

#include <sstmac/software/libraries/actor/simplemailserver.h>
#include <sstmac/software/libraries/actor/simplembox.h>
#include <sstmac/software/process/key.h>
#include <sstmac/software/libraries/actor/simplemboxapp.h>
#include <utility>
#include <deque>
#include <list>

namespace sstmac
{
  namespace sw
  {

// Forward declaration of important types.

    /**
     * A concrete implementation of a simplembox.
     */
    class simplemailserver::mbox : public simplembox
    {
    public:
      typedef simplembox::tagmap_t tagmap_t;

    private:

      /// The process identifier for this object.
      actorid id_;
      softwareid sid_;

      /// The parent object that "owns" this mbox.
      simplemailserver::ptr parent_;

      static tagmap_t tag_map_;

      /// Track whether we are in between init and finalize.
      bool active_;

      /// Blocked receive calls are held here.
      typedef std::pair<actorid, actorpattern::ptr> actorp_t;
      typedef std::pair<actorp_t, key::ptr> pend_t;
      typedef std::list<pend_t> blocked_t;
      blocked_t blocked_;

      /// Un-handled received messages are held here.
      typedef std::pair<actorid, actormsg::ptr> msgpair_t;
      typedef std::list<msgpair_t> unhandled_t;
      unhandled_t unhandled_;

      bool initialized_;

      /// Constructor.
      mbox(softwareid id);

    public:
      typedef boost::intrusive_ptr<mbox> ptr;

      virtual bool
      is_init() const
      {
        return initialized_;
      }

      /// Get us a new mbox.
      static simplemailserver::mbox::ptr
      construct(softwareid sid){
        return simplemailserver::mbox::ptr(new simplemailserver::mbox(sid));
      }

      /// Done.
      virtual
      ~mbox() { }

      /// Allow the parent to send us a message to handle.
      virtual void
      message_arrived(const actorid &sender, const actormsg::ptr& payload);

      /* Methods from here on down are part of the simplembox interface */

      /// What's my pid.
      virtual actorid
      mypid() const;

      /// What nodes are known to the current system?
      // virtual const std::vector<nodeaddress::ptr >& nodes() const;

      /// All peer identifiers (magically, we know about everybody).
      virtual const world_t
      world() const;

      /// Register the given tag mapping in the registry.
      /// A key cannot be redefined unless it is first erased.
      virtual void
      set_tag(int key, actorid value);

      /// Test whether the given key is defined.
      virtual bool
      has_tag(int key) const;

      /// Access the value of the given key.
      virtual actorid
      get_tag(int key) const;

      /// Access the tag mapping in a const context.
      virtual const tagmap_t&
      tag_map() const;

      /// Initialize.  This method is collective and "magic" -- it ensures
      /// that all nodes have joined the party before anybody starts.
      virtual timestamp
      init();

      /// Finalize.  This is a magic operation.
      virtual timestamp
      finalize();

      /// Spawn the given task on the given node and return its pid.
      /// This can only be called between 'init()' and 'finalize()'.
      virtual actorid
      spawn(nodeaddress::ptr &target,
          const boost::intrusive_ptr<simplemboxapp>& entity,
          int64_t communication_size);

      /// Send a message to the given pid.
      /// This call always succeeds except for invalid pid.
      /// \throw valueerror if the pid is not valid.
      /// \throw nullerror if sending a NULL message (reserved for recv timeout).
      virtual void
      send(const actorid &target, actormsg::ptr msg, int64_t size);

      /// Test whether there is any message that matches the given pattern.
      /// This has the same effect as calling recv with a zero-length timeout.
      /// \param pat     The pattern we are matching (may be actorpattern::any).
      virtual bool
      test(const actorid &source, const actorpattern::ptr& pat) const;

      /// Block until we get a message matching the given pattern.
      /// \param source  Our desired source (may be actorid::any).
      /// \param pat     The pattern we are matching (may be actorpattern::any).
      /// \return        The received object (sorry about the dynamic casts)
      ///                or std::make_pair(actorid::invalid, NULL) if timed out.
      virtual std::pair<actorid, actormsg::ptr>
      recv(const actorid &source, const actorpattern::ptr& pat);

      virtual std::string
      to_string() const{
        return "simplemailserver-mbox";
      }

    };

  }
} // end of namespace sstmac

#endif
