/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_ACTOR_SIMPLEMBOX_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_ACTOR_SIMPLEMBOX_H_INCLUDED

#include <sstmac/software/libraries/actor/actorid.h>
#include <sstmac/software/libraries/actor/actormsg.h>
#include <sstmac/software/libraries/actor/actorpattern.h>
#include <sstmac/software/libraries/library.h>
#include <sstmac/software/libraries/actor/actorworld.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/sstmac-config.h>
#include <sstmac/software/process/waiter.h>
#include <utility>
#include <vector>
#include <map>
#include <set>

namespace sstmac { namespace sw {

  // Forward declaration of important types.
  class simplemboxapp;

  /**
   * A very simplified version of a distributed a Kahn process network.
   *
   * Perform non-blocking sends and blocking receives.
   */
  class simplembox: public library {
  public:
    typedef boost::intrusive_ptr<simplembox> ptr;

    /// Required support for knowing about all peer identifiers in our world.
    //typedef std::set<actorid> world_t;
    typedef actorworld::ptr world_t;

    world_t world_;



    /// Required support for globally mapping integer tags to actorid
    /// objects.
    typedef std::map<int, actorid> tagmap_t;

    /// Return type for recv calls.
    typedef std::pair<actorid, actormsg::ptr > recvresult_t;

    simplembox(){

      if(!dbg_.is_real()){
           dbg_ = debug("<debug> mbox | actor");
           logger::register_logger(&dbg_);
       }


    }

    /// Goodbye.
    virtual ~simplembox() {
    }

    virtual bool
    is_init() const = 0;

    /// What's my pid.
    virtual actorid
    mypid() const = 0;

    /// What nodes are known to the current system?
    //virtual const std::vector<nodeaddress::ptr >&
    //nodes() const = 0;

    /// All peer identifiers (magically, we know about everybody).
    virtual const world_t
    world() const = 0;

    /// Register the given tag mapping in the registry.
    /// A key cannot be redefined unless it is first erased.
    virtual void
    set_tag(int key, actorid value) = 0;

    /// Test whether the given key is defined.
    virtual bool
    has_tag(int key) const = 0;

    /// Access the value of the given key.
    virtual actorid
    get_tag(int key) const = 0;

    /// Access the tag mapping in a const context.
    virtual const tagmap_t&
    tag_map() const = 0;


    /// Initialize.  This method is collective and "magic" -- it ensures
    /// that all nodes have joined the party before anybody starts.
    virtual timestamp
    init() = 0;

    /// Finalize.  This is a magic operation.
    virtual timestamp
    finalize() = 0;

    /// Spawn the given task on the given node and return its pid.
    /// This can only be called between 'init()' and 'finalize()'.
    virtual actorid
    spawn(nodeaddress::ptr &target,
        const boost::intrusive_ptr<simplemboxapp>& entity,
        int64_t communication_size) = 0;

    /// Send a message to the given pid.
    /// This call always succeeds except for invalid pid.
    /// \throw valueerror if the pid is not valid.
    /// \throw nullerror if sending a NULL message (reserved for recv timeout).
    virtual void
    send(const actorid &target, actormsg::ptr msg,
        int64_t size) = 0;

    /// Test whether there is any message that matches the given pattern.
    /// This has the same effect as calling recv with a zero-length timeout.
    /// \param source  Our desired source (may be actorid::any).
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    virtual bool
    test(const actorid &source,
        const actorpattern::ptr& pat) const = 0;

    /// Block until we get a message matching the given pattern.
    /// \param source  Our desired source (may be actorid::any).
    /// \param pat     The pattern we are matching (may be actorpattern::any).
    /// \return        The received object (sorry about the dynamic casts)
    ///                or std::make_pair(actorid::invalid, NULL) if timed out.
    virtual recvresult_t
    recv(const actorid &source,
        const actorpattern::ptr& pat) = 0;


  protected:
    static debug dbg_;

  };


} } // end of namespace sstmac

#endif
