/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/filesystem/nppfs.h>
#include <sstmac/software/libraries/filesystem/file_metadata.h>
#include <sstmac/software/libraries/filesystem/disk_message.h>
#include <math.h>

namespace sstmac
{
  namespace sw
  {

    debug nppfs::dbg_;

    openfile::ptr
    nppfs::process_openfile_meta(sst_message::ptr msg)
    {
      file_metadata::ptr arrived = boost::dynamic_pointer_cast<file_metadata>(
          msg);

      file_descriptor::ptr fdes = boost::dynamic_pointer_cast<file_descriptor>(
          arrived->data_);

      openfile_parallel::ptr ret = openfile_parallel::construct(fdes);

      return ret;
    }

    openfile::ptr
    nppfs::create(std::string name)
    {
      SSTMAC_DEBUG << "nppfs[" << os_->get_my_addr()->to_string()
            << "]::create - creating file " << name << "\n";
      file_metadata::ptr meta = file_metadata::construct("nppfs_server", name,
          file_metadata::CREATE);
      meta->bytes_ = 8;
      meta->from_lib_ = get_lib_name();

      os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, meta);

      sst_message::ptr arr = wait_type(sst_message::FILEDATA);

      if (arr)
        {
          return process_openfile_meta(arr);
        }
      else
        {
          return openfile::ptr();
        }

    }

    openfile::ptr
    nppfs::open(std::string name)
    {
      SSTMAC_DEBUG << "nppfs[" << os_->get_my_addr()->to_string()
            << "]::open - opening file " << name << "\n";
      file_metadata::ptr meta = file_metadata::construct("nppfs_server", name,
          file_metadata::OPEN);
      meta->bytes_ = 8;
      meta->from_lib_ = get_lib_name();

      os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, meta);

      sst_message::ptr arr = wait_type(sst_message::FILEDATA);
      if (arr)
        {
          return process_openfile_meta(arr);
        }
      else
        {
          return openfile::ptr();
        }

    }

    void
    nppfs::read(openfile::ptr file, long bytes)
    {
      openfile_parallel::ptr parfile = boost::dynamic_pointer_cast<
          openfile_parallel>(file);
      std::vector<node_access::ptr> nodes = get_access_nodes(parfile, bytes);

      SSTMAC_DEBUG << "nppfs[" << os_->get_my_addr()->to_string()
            << "]: reading from " << nodes.size() << " nodes \n";
      for (int i = 0; i < nodes.size(); i++)
        {
          file_metadata::ptr meta = file_metadata::construct("nppfs_server",
              parfile->name(), file_metadata::READ);
          meta->toaddr_ = nodes[i]->addr_;
          meta->bytes_ = 8;
          meta->access_size_ = nodes[i]->bytes_;
          meta->from_lib_ = get_lib_name();

          openfile::ptr of = openfile::construct(parfile->name(),
              nodes[i]->sector_);
          of->seek(nodes[i]->offset_);
          meta->data_ = of;

          SSTMAC_DEBUG << "nppfs[" << os_->get_my_addr()->to_string()
                << "]: read request: " << meta->to_string() << "\n";

          os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, meta);
        }

      sst_message::ptr arr = wait_num_type(sst_message::FILEDATA, nodes.size());

    }

    void
    nppfs::write(openfile::ptr file, long bytes, bool wait_complete)
    {
      openfile_parallel::ptr parfile = boost::dynamic_pointer_cast<
          openfile_parallel>(file);
      parfile->update_max_access(parfile->get_offset() + bytes);
      std::vector<node_access::ptr> nodes = get_access_nodes(parfile, bytes);

      SSTMAC_DEBUG << "nppfs[" << os_->get_my_addr()->to_string()
            << "]: writing to " << nodes.size() << " nodes \n";
      for (int i = 0; i < nodes.size(); i++)
        {
          file_metadata::ptr meta = file_metadata::construct("nppfs_server",
              parfile->name(), file_metadata::WRITE);
          meta->toaddr_ = nodes[i]->addr_;
          meta->bytes_ = 8;
          meta->access_size_ = nodes[i]->bytes_;
          meta->from_lib_ = get_lib_name();

          openfile::ptr of = openfile::construct(parfile->name(),
              nodes[i]->sector_);
          of->seek(nodes[i]->offset_);
          meta->data_ = of;
          meta->write_resp_ = wait_complete;

          os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, meta);
        }

      if (wait_complete)
        {
          sst_message::ptr arr = wait_num_type(sst_message::FILEDATA,
              nodes.size());
        }

    }

    void
    nppfs::seek(openfile::ptr file, long bytes)
    {
      file->seek(bytes);

    }

    std::vector<nppfs::node_access::ptr>
    nppfs::get_access_nodes(openfile_parallel::ptr file, long bytes)
    {
      std::vector<node_access::ptr> ret;

      if (bytes < 0 || bytes > file->file_->total_size_)
        {
          std::cout << "BYTES: " << bytes << "\n";
          throw valueerror(
              "nppfs: trying to access a number of bytes that isn't valid for this file");
        }

      long totaloffset = 0; // in the file as a whole
      bool first = true;
      long accounted = 0;
      for (int i = 0; i < file->file_->nodes_.size() && totaloffset
          < file->get_offset() + bytes; i++)
        {
          if (file->get_offset() < totaloffset + file->file_->sizes_[i])
            {
              long off = 0;
              if (first)
                off = file->get_offset() - totaloffset;
              long by = std::min(file->file_->sizes_[i] - off,
                  bytes - accounted);

              node_access::ptr na = node_access::construct(
                  file->file_->nodes_[i], off, file->file_->sectors_[i], by);
              ret.push_back(na);
              SSTMAC_DEBUG << "nppfs[" << os_->get_my_addr()->to_string()
                    << "]::get_access_nodes - accessing node "
                    << na->addr_->to_string() << ", offset " << off
                    << ", bytes " << by << "\n";
              first = false;
              accounted += by;
            }
          totaloffset += file->file_->sizes_[i];

        }
      return ret;
    }

    void
    nppfs::close(openfile::ptr file)
    {
      file_metadata::ptr meta = file_metadata::construct("nppfs_server",
          file->name(), file_metadata::CLOSE);

      os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, meta);

      file->close();

    }

    bool
    nppfs::supported()
    {
      bool cansend = os_->kernel_supported(AMI::AMI_COMM, AMI::AMI_COMM_SEND);
      bool hasdisk = os_->kernel_supported(AMI::AMI_COMPUTE,
          AMI::AMI_COMP_DISKACCESS);
      return cansend && hasdisk;
    }

  }
} /* namespace sstmac */
