/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_QUEUE_MPIQUEUE_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_QUEUE_MPIQUEUE_H_INCLUDED

#include <sstmac/common/sstmac-config.h>
#include <sstmac/software/libraries/mpi/mpimessage.h>
#include <sstmac/software/libraries/mpi/mpirequest.h>
#include <sstmac/software/libraries/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitype.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitag.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpiid.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/common/messages/messageid.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/libraries/mpi/mpiserver.h>
#include <list>
#include <map>
#include <set>

namespace sstmac { namespace sw
{

  // Important types.
  class mpistatus;
  class nodeinterface;
  class networkinterface;

  //class mpisend;
  //class mpirecv;
  //class mpiprobe;
  //class payload;
  //class eventhandler;
  class appid;

  ///* A one-per-peer queue that handles inbound and outbound messages.
  ///*
  ///*

  class mpiqueue : virtual public nocopy, public ptr_type
  {
    /// The mpiserver is allowed to access the private incoming_message method.
    //// friend class mpiserver;
  public:
    typedef boost::intrusive_ptr<mpiqueue> ptr;

  private:

    /// Nested types.
    class recvrequest;
    class sendrequest;
    class proberequest;

    /// Utility class to sort arriving messages in arrival order.
    struct sortbyseqnum
    {
      bool
      operator()(const mpimessage::ptr &a,
          const mpimessage::ptr &b) const;
    };



    /// The shared backend (currently likely to result in a memory leak).
    mpiserver::ptr server_;

    /// The node.
    //boost::shared_ptr<nodeinterface> node_;

    /// The cutoff at which we switch between direct (ready send)
    /// and handshake.
    int64_t handsize_;

    /// Extent of envelope information.
    int64_t envelope_;

    /// Minimum transmission size.
    int64_t mintrans_;

    taskid myid_;

    /// The sequence number for our next outbound transmission.
    std::map<taskid, int> next_outbound_;

    /// The sequence number expected for our next inbound transmission.
    std::map<taskid, int> next_inbound_;

    /// Hold messages that arrived out of order.
    typedef std::set<mpimessage::ptr, sortbyseqnum> hold_list_t;
    std::map<taskid, hold_list_t> held_;

    /// The (locally unique) id that will be given to the next message.
    messageid next_id_;

    /// Receive requests waiting for a matching message (or envelope).
    typedef std::list<boost::intrusive_ptr<recvrequest> > pending_message_t;
    pending_message_t pending_message_;

    /// Inbound messages waiting for a matching receive request.
    typedef std::list<mpimessage::ptr > need_recv_t;
    need_recv_t need_recv_;

    // Save all sends so we can match up the nic acks that come back
    typedef std::list<boost::intrusive_ptr<sendrequest> > send_needs_ack_t;
    send_needs_ack_t send_needs_ack_;

    // nics that arrived  that we might need to save in case a wait has been called
    //typedef std::list<mpimessage::ptr > send_ack_needs_send_t;
    /// send_ack_needs_send_t send_ack_needs_send_;

    /// Inbound messages that arrived out-of-order and are waiting for
    /// the message sequence to be complete.
    typedef std::map<int, mpimessage::ptr > reorderlist_t;
    reorderlist_t reorderlist_;

    /// Requests waiting for an ack.
    typedef std::map<messageid, boost::intrusive_ptr<sendrequest> > ack_needed_t;
    ack_needed_t ack_needed_;

    /// Requests that sent an ack and are waiting for the rest of their data.
    typedef std::map<messageid, boost::intrusive_ptr<recvrequest> > ack_sent_t;
    ack_sent_t ack_sent_;

    /// Probe requests watching for a given envelope.
    typedef std::list<boost::intrusive_ptr<proberequest> > probelist_t;
    probelist_t probelist_;

    static debug dbg_;

    static stat_spyplot::ptr spy_;

    static stat_spyplot::ptr spy_bytes_;

  protected:
    /// Hi there.
    mpiqueue(const mpiserver::ptr& server, taskid myid,
        // const boost::shared_ptr<nodeinterface>& node,
        int64_t handshake_cutoff, int64_t envelope_bytes,
        int64_t smallest_transmission_unit);



  public:
    virtual std::string to_string() const{
            return "mpiqueue";
          }

    /// \param server:   The MPI server through which we route messages.
    /// \param node:     The node to which we post copy and compute requests.
    /// \param envelope: Number of bytes prefixed to each MPI message
    /// \param mintrans: Smallest MPI message.
    static mpiqueue::ptr
    construct(const mpiserver::ptr& server, taskid myid,
        //const boost::shared_ptr<nodeinterface>& node,
        int64_t handshake_cutoff = 1 << 16, int64_t envelope = 24,
        int64_t mintrans = 0);

    /// Goodbye.
    virtual
    ~mpiqueue() throw ();

    /// Send data.  All sends are triggered at time nodeinterface::now().
    /// Upon completion, the completion object have an updated
    /// status object relating to the send operation.
    void
    send(const mpirequest_t key, int count, const mpitype &type,
        const mpiid &dest, const mpitag &tag, const mpicomm &comm, bool rsend,
        bool ssend, mpimessage::category message_type,
        const eventhandler::ptr &completion,
        const payload::const_ptr& content = payload::null());

    /// Receive data.  All receives are triggered at time nodeinterface::now().
    /// Upon completion, the receiver object will be updated to
    /// hold a copy of the inbound mpimessage with updated status information
    /// including any inbound content (null or not).
    void
    recv(const mpirequest_t key, int count, const mpitype &type,
        const mpiid &source, const mpitag &tag, const mpicomm &comm,
        mpimessage::category message_type,
         eventhandler::ptr &completion);

    /// Ask for a notification when a message with the given signature arrives.
    /// Probing is only enabled for message_type==mpimessage::user.
    /// Upon completion, the status object held by the completion object
    /// will be updated to reflect the envelope information for the incoming
    /// message, and the completion object will be notified 
    /// (using eventhandler::complete).
    void
    probe(const mpirequest_t key, const mpicomm &comm, const mpiid &source, const mpitag &tag,
        eventhandler::ptr &completion);

    /// Immediate-mode probe for whether a message with the given signature
    /// has arrived.
    /// Probing is only enabled for message_type==mpimessage::user.
    /// \return null if no such message is pending.
    mpistatus::ptr
    iprobe(const mpicomm &comm, const mpiid &source, const mpitag &tag);

    /// CALLBACK used by mpiserver when this object has a message.
    void
    incoming_message(const mpimessage::ptr& message);

    appid
    get_appid();

    static void set_eventmanager(const eventmanager::ptr &m);

  protected:
    /// The incoming_message method forwards to here when the message is
    /// a response to a handshake request.
    void
    incoming_ack_response(const mpimessage::ptr &message);

    /// The incoming_message method forwards to here when the message is
    /// data delivered following a handshake request this queue ack-ed.
    void
    incoming_data_payload(const mpimessage::ptr &message);

    /// The incoming_message method forwards to here when the message is
    /// a new message, whehter it's an eager send or a new handshake request.
    void
    incoming_new_message(const mpimessage::ptr &message);

    /// Called by the mpiserver when it figured out that the nic is sending
    /// back an ack that indicates that the message was sent
    void
    handle_send_ack(const mpimessage::ptr &message);

    /// The incoming_new_message method calls here to complete an
    /// inbound message.  At this point, arrival ordering has been ensured.
    void
    handle_new_message(const mpimessage::ptr &message);

    /// Used when a receive operation sends an ack (either ready-to-receive
    /// or message received).
    /// The ack is keyed to the sequence number associated with the mpimessage.
    void
    send_ack(const mpimessage::ptr& message,
        const boost::intrusive_ptr<recvrequest>& operation);

    /// Used when a send operation requests an ack for the given message.
    /// The ack is keyed to the sequence number associated with the mpimessage.
    /// When an ack arrives for this message_id, the sendrequest::ack
    /// method will be called.
    void
    request_ack(const mpimessage::ptr& message,
        const boost::intrusive_ptr<sendrequest>& operation);

  };

} } // end of namespace sstmac


#endif
