/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpibtreegather.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>

#include <sstmac/common/eventhandler.h>

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_iterator/btscatteriterator.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <deque>

namespace sstmac { namespace sw
{

  // Pointer types.
  
  
  using boost::dynamic_pointer_cast;

  //
  // Given an absolute rank, return the relative rank.
  //
  int
  mpibtreegather::relrank(const mpiid &therank) const
  {
    return ((therank.id - root_.id + comm_.size().id) % comm_.size().id);
  }

  //
  // Given a relative rank, return the absolute rank.
  //
  mpiid
  mpibtreegather::absrank(int therank) const
  {
    return mpiid((therank + root_.id) % comm_.size().id);
  }

  //
  // Fire off any and all receives concurrently.
  //
  void
  mpibtreegather::init_recvs()
  {
    SSTMAC_DEBUG << "mpibtreegather[" << comm_.rank().id << "]::init_recvs()\n";
    const int relative_rank = relrank(comm_.rank());
    const int size = comm_.size().id;
    int keymask = 1;
    if (relative_rank == 0)
      {
        int v = size;
        int shift = 0;
        while (v)
          {
            ++shift;
            v >>= 1;
          }
        keymask = 1 << shift;
      }
    else
      {
        while (!(relative_rank & keymask))
          {
            keymask <<= 1;
          }
      }
    // Figure out the sources for this node.
    int localsrc = 1;
    std::vector<mpisource> sources;
    while (localsrc < keymask)
      {
        const int source = relative_rank + localsrc;
        if (source >= size)
          break;
        const int entries = ((source + localsrc <= size) ? localsrc : size
            - source);
        const int length = entries * recvcount_;
        ++pending_recvs_;
        sources.push_back(mpisource(absrank(source), length));
        localsrc <<= 1;
      }
    if (pending_recvs_)
      {
        // Fire off all receives at once.
        for (size_t i = 0; i < sources.size(); ++i)
          {
            this->start_recv(sources[i].count, recvtype_, tag_,
                sources[i].source);
          }
      }
    else
      {
        // Leaf node -- no receives, just sends.
        init_send();
      }
  }

  //
  // Fire off the send operation (if any).  Can be called from root.
  //
  void
  mpibtreegather::init_send()
  {
    SSTMAC_DEBUG << "mpibtreegather[" << comm_.rank().id << "]::init_send()\n";
    // We should only be here when all receives are done.
    if (pending_recvs_)
      {
        throw illformederror("mpibtreegather::init_send:  Pending receives.");
      }
    const int relative_rank = relrank(comm_.rank());
    const int size = comm_.size().id;
    if (relative_rank != 0)
      {
        // We need to do a send.
        int keymask = 1;
        while (!(relative_rank & keymask))
          keymask <<= 1;
        const int localdest = relative_rank - keymask;
        const int entries = ((relative_rank + keymask) <= size ? keymask : size
            - relative_rank);
        const int length = entries * sendcount_;
        ++pending_sends_;
        this->start_send(length, sendtype_, tag_, absrank(localdest), content_);
      }
    else if (pending_sends_ <= 0 && pending_recvs_ <= 0)
      {

        mpicollective::complete(content_);
      }
  }

  //
  // Hi.
  //
  mpibtreegather::mpibtreegather(const mpirequest::ptr& thekey,
      const mpiqueue::ptr& queue, int sendcount,
      const mpitype &sendtype, int recvcount, const mpitype &recvtype,
      const mpiid &root, const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content,
      eventhandler::ptr completion) :
    mpicollective(thekey, queue, comm, completion), sendcount_(sendcount),
        sendtype_(sendtype), recvcount_(recvcount), recvtype_(recvtype), root_(
            root), tag_(tag), comm_(comm), pending_sends_(0), pending_recvs_(0)
  {
    // bool isroot = comm_.rank() == root_;
    // Clone the content if it is non-null.
    // magic_->add_collective_payload(typeid(*this), tag_, comm_,
    //                               (content ? content->clone() : content),
    //                               isroot);

    std::vector<payload::const_ptr > vals;
    for (int i = 0; i < comm.size().id; i++)
      {
        if (i == comm.rank().id)
          vals.push_back(content);
        else
          vals.push_back(payload::const_ptr ());
      }

    content_ = mpicollpayload::construct(vals);

    // content_ = boost::dynamic_pointer_cast<mpicollpayload>(content->clone());
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void
  mpibtreegather::send_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &dest)
  {
    SSTMAC_DEBUG << "mpibtreegather[" << comm_.rank().id << "]::send_complete(" << ", " << count << ", "
          << type.id << ", " << tag.id << ", " << dest.id << ")\n";
    --pending_sends_;
    if (!pending_sends_)
      {

        mpicollective::complete(content_);
      }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void
  mpibtreegather::recv_complete(int count,
      const mpistatus_t& stat)
  {
    SSTMAC_DEBUG << "mpibtreegather[" << comm_.rank().id << "]::recv_ complete(" << ", " << stat << ")\n";
    // this->start_funneled_copy(count, stat->recv_type(), stat->tag(),
    //     stat->source());

    payload::const_ptr pay = stat->content();
    if (pay)
      {
        mpicollpayload::const_ptr other = mpicollpayload::construct(
            boost::dynamic_pointer_cast<const mpicollpayload> (stat->content()->clone()));

        std::vector<payload::const_ptr > realload =
            other->get_content();

        std::vector<payload::const_ptr >::iterator it, end =
            realload.end();

        int i = 0;
        for (it = realload.begin(); it != end; it++)
          {
            if (*it)
              {
                if (dbg_.is_active())
                  dbg_ << "mpibtreegather[" << comm_.rank().id << "]: setting content for position " << i << "\n";

                content_->set_content(*it, i);
              }
            i++;
          }
      }

    --pending_recvs_;
    if (!pending_recvs_)
      {
        this->init_send();
      }
  }

  //
  /// Callback for the copier to signal completion.
  //
  void
  mpibtreegather::copy_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &source)
  {
    SSTMAC_DEBUG << "mpibtreegather[" << comm_.rank().id << "]::copy_done(" << ", " << type.id << ", " << tag.id
          << ", " << source.id << ")\n";
    --pending_recvs_;
    if (!pending_recvs_)
      {
        this->init_send();
      }
  }

  //
  // Hello.
  //
  mpibtreegather::ptr
  mpibtreegather::construct(const mpirequest::ptr& thekey,
      const mpiqueue::ptr& queue, int sendcount,
      const mpitype &sendtype, int recvcount, const mpitype &recvtype,
      const mpiid &root, const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content,
      eventhandler::ptr completion)
  {
    mpibtreegather::ptr retval(new mpibtreegather(thekey, queue,
        sendcount, sendtype, recvcount, recvtype, root, tag, comm, content,
        completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpibtreegather::~mpibtreegather() throw ()
  {
  }

  //
  // Start this kernel.
  //
  void
  mpibtreegather::start()
  {
    SSTMAC_DEBUG << "mpibtreegather[" << comm_.to_string() << "]::start(" << ")\n";
    mpicollective::start();
    this->init_recvs();
  }

} } // end of namespace sstmac

