/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpirabenseifner.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpirankmap.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw
{

  

  // Nested sendrecv container.
  mpirabenseifner::sendrecv::sendrecv() :
    partner(-1), send_count(0), recv_count(0)
  {
  }
  mpirabenseifner::sendrecv::sendrecv(mpiid pt, int sc, int rc, bool rd) :
    partner(pt), send_count(sc), recv_count(rc), recv_reduce(rd)
  {
    // if(debugger::active())
    // SSTMAC_DEBUG << "mpirbenseifner::sendrecv::sendrecv(" << pt.id << ", "
    //       << sc << ", " << rc << ", " << rd << ")\n";
  }

  //
  // Private method to run send/receive operations.
  //
  void
  mpirabenseifner::do_sendrecv()
  {

      SSTMAC_DEBUG << "Enter mpirabenseifner[" << comm_.to_string()
          << "]::do_sendrecv: " << "queue_.size()=" << queue_.size()
          << ", pending_sends=" << pending_sends_ << ", pending_recvs="
          << pending_recvs_ << "\n";
    while ((pending_recvs_ <= 0) && (!queue_.empty()))
      {
        // Pop of at least one sendreceive, and then keep popping them off
        // until we hit a receive.  Receives must be completed before next-level
        // send operations.
        sendrecv sr = queue_.front();
        if ((sr.partner.id < 0) || (sr.partner >= comm_.size()))
          {
            std::cerr << "Invalid partner " << sr.partner << "\n";
            throw illformederror("mpirabenseifner::do_sendrecv:  Partner "
              "out of range.");
          }
        queue_.pop_front();
        if (sr.send_count >= 0)
          {

              SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                  << "]::do_sendrecv:  send " << sr.send_count << " to "
                  << sr.partner.id << "\n";
            ++pending_sends_;
            this->start_send(sr.send_count, type_, tag_, sr.partner, content_);
          }
        if (sr.recv_count >= 0)
          {

              SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                  << "]::do_sendrecv  recv " << sr.recv_count << " from "
                  << sr.partner.id << "\n";
            ++pending_recvs_;
            reduce_this_recv_ = sr.recv_reduce;
            this->start_recv(sr.recv_count, type_, tag_, sr.partner);
          }

          SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string() << "]::do_sendrecv: "
              << "queue size is now " << queue_.size() << "\n";
      }

      SSTMAC_DEBUG << "Testing mpirabenseifner[" << comm_.to_string()
          << "]::do_sendrecv: " << "queue_.size()=" << queue_.size()
          << ", pending_sends=" << pending_sends_ << ", pending_recvs="
          << pending_recvs_ << ", got_result_=" << got_result_ << "\n";
    if (queue_.empty() && pending_sends_ == 0 && pending_recvs_ == 0)
      {
        payload::const_ptr result = content_;
        if (!got_result_)
          {
            if (get_result_)
              {

                  SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                      << "]:  Getting payload\n";
                // if (!magic_->have_all_collective_payload(typeid(*this), tag_,
                //     comm_))
                //   throw illformederror(
                //       "mpirabenseifner::do_sendrecv:  Consumer "
                //         "node done, but payload result unavailable.");
                // result = magic_->get_collective_payload(typeid(*this), tag_,
                //     comm_);
              }
            got_result_ = true;

              SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                  << "]::do_sendrecv:  done\n";
            mpicollective::complete(result);
          }
      }

      SSTMAC_DEBUG << "Return from mpirabenseifner[" << comm_.to_string()
          << "]::do_sendrecv: " << "queue_.size()=" << queue_.size()
          << ", pending_sends=" << pending_sends_ << ", pending_recvs="
          << pending_recvs_ << "\n";
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void
  mpirabenseifner::send_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &dest)
  {

    (void) count;
    (void) type;
    (void) tag;
    (void) dest;

      SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string() << "]: send_complete \n";
    --pending_sends_;
    this->do_sendrecv();
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void
  mpirabenseifner::recv_complete(int count,
      const mpistatus_t& stat)
  {

      SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string() << "]: recv_complete \n";

      --pending_recvs_;

    payload::const_ptr load = stat->content();

    if (reduce_this_recv_)
      {
        if (load)
          {
            SSTMAC_DEBUG << "combining content " << content_->to_string() << ", with " << load->to_string() << "\n";
            content_ = combine_content(content_, load, op_);
          }
      }else{
          SSTMAC_DEBUG << "not combining this turn \n";

          if(pending_recvs_ == 0){
              content_ = load;
          }
      }
    //if (reduce_this_recv_)
    //  start_funneled_copy(count, type_, tag_, stat->source());
    //else
    //  {

    this->do_sendrecv();
    //}
  }

  //
  // Callback method to indicate that a copy operation has completed.
  //
  void
  mpirabenseifner::copy_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &source)
  {

    (void) count;
    (void) type;
    (void) tag;
    (void) source;
    --pending_recvs_;
    this->do_sendrecv();
  }

  //
  // Hello.
  //
  mpirabenseifner::mpirabenseifner(bool allreduce,
      const mpirequest::ptr& thekey, const mpiqueue::ptr& queue,
      int count, const mpitype &datatype, const mpiop &op, const mpiid &root,
      const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content,
      const eventhandler::ptr& completion) :
    mpicollective(thekey, queue, comm, completion), allreduce_(allreduce),
        count_(count), type_(datatype), op_(op), root_(root), tag_(tag), comm_(
            comm), pending_sends_(0), pending_recvs_(0), reduce_this_recv_(0)
  {
    //  if(debugger::active())
    //    debug() << "mpirabenseifner[" << comm_ << "]::mpirabenseifner("
    //          << allreduce << ", "
    //          << thekey << ", " << queue << ", " << count << ", "
    //           << datatype << ", op, " << root << ", " << tag << ", "
    //           << comm << ", " << content << ", " << magic << ")\n";
    if ((!allreduce) && (root < mpiid(0) || root >= comm_.size()))
      throw valueerror("mpirabenseifner:  Reduce operation requested with "
        "a root index that's out of range.");
    if (count < 0)
      throw valueerror("mpirabenseifner:  Won't send negative element counts.");
    if (allreduce)
      root_ = mpiid(-1);
    get_result_ = (allreduce ? true : (comm_.rank() == root));
    got_result_ = false;
    // Clone the content if it is non-null
    //  magic_->add_collective_payload(typeid(*this), tag, comm,
    //     (content ? content->clone() : content), get_result_);
    content_ = content;
  }

  //
  // Hi.
  //
  mpirabenseifner::ptr
  mpirabenseifner::construct(bool allreduce,
      const mpirequest::ptr& thekey, const mpiqueue::ptr& queue,
      int count, const mpitype &datatype, const mpiop &op, const mpiid &root,
      const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content,
      const eventhandler::ptr& completion)
  {
    mpirabenseifner::ptr retval(new mpirabenseifner(allreduce, thekey,
        queue, count, datatype, op, root, tag, comm, content, completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpirabenseifner::~mpirabenseifner() throw ()
  {
  }

  //
  // Get busy child.
  //
  void
  mpirabenseifner::start()
  {

      SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string() << "]::start\n";
    // Precompute the queue.
    queue_.clear();
    mpirankmap rmap(root_, comm_);
    // Donor/acceptor exchange to shrink data.
    if (rmap.donor() || rmap.acceptor())
      {
        int lowerhalf = count_ / 2;
        int upperhalf = count_ - lowerhalf;
        mpiid partner = rmap.exchangewith();
        if (rmap.donor())
          {

              SSTMAC_DEBUG << "mpirbenseifner[" << comm_.to_string()
                  << "]::start:  Node is donor\n";
            queue_.push_back(sendrecv(partner, lowerhalf, upperhalf, true));
            queue_.push_back(sendrecv(partner, upperhalf, -1, false));
          }
        else if (rmap.acceptor())
          {

              SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                  << "]::start:  Node is acceptor\n";
            queue_.push_back(sendrecv(partner, lowerhalf, upperhalf, true));
            queue_.push_back(sendrecv(partner, -1, upperhalf, false));
          }
      }
    // Recursive exchanges with progressively larger stride and less data.
    std::vector<int> evensize(rmap.log2(), -1);
    std::vector<int> oddsize(rmap.log2(), -1);
    int effrank = rmap.effective_rank();
    if (!rmap.donor())
      {
        int x_count = count_;
        for (int logdist = 0; logdist < rmap.log2(); ++logdist)
          {
            evensize.at(logdist) = x_count / 2;
            oddsize.at(logdist) = x_count - evensize[logdist];
            int dist = 1 << logdist;

              SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string() << "]::start: "
                  << "Recursive fan-out with range=" << dist << "\n";
            mpiid partner;
            int sendcount = -1, recvcount = -1;
            if ((effrank & dist) == 0)
              { // "even" rank w.r.t. current stride
                partner = rmap.real_rank(effrank + dist);
                x_count = evensize.at(logdist);
                sendcount = oddsize.at(logdist);
                recvcount = evensize.at(logdist);
              }
            else
              { // "odd" rank w.r.t. current strie
                partner = rmap.real_rank(effrank - dist);
                x_count = oddsize.at(logdist);
                sendcount = evensize.at(logdist);
                recvcount = oddsize.at(logdist);
              }
            queue_.push_back(sendrecv(partner, sendcount, recvcount, true));
          }
      }
    // Recursive exchanges with progressively smaller stride and more data.
    // This is where we do things differently for reduce or allreduce.
    if (allreduce_)
      {
        // Reduce data onto all nodes except donor nodes
        if (!rmap.donor())
          {
            for (int logdist = rmap.log2() - 1; logdist >= 0; --logdist)
              {
                int dist = 1 << logdist;

                  SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                      << "]::start: " << "Allreduce: Fan-in with range="
                      << dist << "\n";
                mpiid partner(-1);
                int sendcount = -1, recvcount = -1;
                if ((effrank & dist) == 0)
                  { // "even" w.r.t. current stride

                      SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                          << "]::start: " << "Even node with relative partner "
                          << effrank + dist << "\n";
                    partner = rmap.real_rank(effrank + dist);
                    sendcount = evensize.at(logdist);
                    recvcount = oddsize.at(logdist);
                  }
                else
                  { // "odd" rank w.r.t. current stride.

                      SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                          << "]::start: " << "Even node with relative partner "
                          << effrank - dist << "\n";
                    partner = rmap.real_rank(effrank - dist);
                    sendcount = oddsize.at(logdist);
                    recvcount = evensize.at(logdist);
                  }
                queue_.push_back(sendrecv(partner, sendcount, recvcount, false));
              }
          }
        // Update donor nodes
        if (rmap.donor())
          {

              SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string() << "]::start: "
                  << "Donor node receiving update data\n";
            queue_.push_back(sendrecv(rmap.exchangewith(), -1, count_, false));
          }
        else if (rmap.acceptor())
          {

              SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string() << "]::start: "
                  << "Acceptor node sending updated data.\n";
            queue_.push_back(sendrecv(rmap.exchangewith(), count_, -1, false));
          }
      }
    else
      { // not MPI_Allreduce -- just a regular MPI_Reduce
        bool remapped_root = false;
        int newroot = rmap.effective_rank(root_);
        if (rmap.donor(root_))
          {

              SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                  << "]::start: Remap root\n";
            remapped_root = true;
            newroot = 0;
            if (comm_.rank() == root_)
              {
                // Compute the receive indices that the old node 0 has used to date.
                effrank = newroot;
                int x_count = count_;
                for (int logdist = 0; logdist < rmap.log2(); ++logdist)
                  {
                    evensize.at(logdist) = int(x_count) / 2;
                    oddsize.at(logdist) = x_count - evensize.at(logdist);
                    x_count /= 2;
                  }
                // Set up to receive all data from the old node 0.
                queue_.push_back(sendrecv(mpiid(0), -1, count_, false));
              }
            else if (comm_.rank() == mpiid(0))
              {
                // This node is swapping places with the remapped root.
                queue_.push_back(sendrecv(root_, count_, -1, false));
              }
          }
        // Now it's mostly a normal exchange, except we are excluding some nodes
        bool include_me = (comm_.rank() == root_ || !(rmap.donor()
            || (remapped_root && comm_.rank() == mpiid(0))));
        if (include_me)
          {
            for (int logdist = rmap.log2() - 1; logdist >= 0; --logdist)
              {
                int dist = 1 << logdist;

                  SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                      << "]::start: " << "Reduce: Fan-in with range " << dist
                      << "\n" << "    effective rank: " << effrank << "\n"
                      << "    root: " << root_.id << "\n";
                int relative_partner;
                int transfer = -1;
                if ((effrank & dist) != (newroot & dist))
                  { // sender
                    if ((effrank & dist) == 0)
                      { // even
                        relative_partner = effrank + dist;
                        transfer = evensize.at(logdist);
                      }
                    else
                      { // odd
                        relative_partner = effrank - dist;
                        transfer = oddsize.at(logdist);
                      }
                    mpiid partner(-1);
                    if (remapped_root && relative_partner == 0)
                      partner = root_;
                    else
                      partner = rmap.real_rank(relative_partner);
                    queue_.push_back(sendrecv(partner, transfer, -1, false));
                  }
                else
                  { // receiver
                    if ((effrank & dist) == 0)
                      { // even
                        relative_partner = effrank + dist;
                        transfer = oddsize.at(logdist);

                          SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                              << "]::start: " << "relative partner: "
                              << relative_partner << "\n";
                      }
                    else
                      { // odd
                        relative_partner = effrank - dist;
                        transfer = evensize.at(logdist);

                          SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                              << "]::start:  " << "relative partner: "
                              << relative_partner << "\n";
                      }
                    mpiid partner = rmap.real_rank(relative_partner);

                      SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
                          << "]::start:  " << "real partner: " << partner.id
                          << "\n";
                    queue_.push_back(sendrecv(partner, -1, transfer, false));
                  }
              }
          }
      }

      SSTMAC_DEBUG << "mpirabenseifner[" << comm_.to_string()
          << "]::start:  Start collective\n";
    mpicollective::start();
    do_sendrecv();
  }

} } // end of namespace sstmac
