/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollective.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollective-sender.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollective-recver.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollective-funneledcopy.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollective-immediatecopy.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>
//#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw
{

  // Pointer types.
  
  debug mpicollective::dbg_;

  //
  // Hi.
  //
  mpicollective::mpicollective(const mpirequest::ptr& the_key,
      const mpiqueue::ptr& queue, const mpicomm &comm,
      eventhandler::ptr completion) :
    req_(the_key), status_(mpistatus::construct()), queue_(queue), comm_(comm),
        completion_(completion)
  {
    if(!dbg_.is_real()){
        dbg_ = debug("<debug> mpicollective | mpi | library", logger::cyan);
        logger::register_logger(&dbg_);
    }
  }



  //
  // Request a send.
  //
  void
  mpicollective::start_send(int count, const mpitype &type, const mpitag &tag,
      const mpiid &dest, const payload::const_ptr& content)
  {
    if (dbg_.is_active())
      dbg_ << "mpicollective[" << comm_.rank().id << "], start_send to " << dest.to_string() << "\n";
    //       << "]::start_send(" << count << ", " << type << ", " << tag << ", "
    //       << dest << ", " << content << ")\n";
    mpirequest_t req = mpirequest::construct(key::construct());
    eventhandler::ptr snd = sender::construct(mpicollective::ptr(this), status_,
        count, type, tag, dest);
    queue_->send(req, count, type, dest, tag, comm_, false, false,
        mpimessage::collective, snd, content);
  }

  //
  // Request a recv.
  //
  void
  mpicollective::start_recv(int count, const mpitype &type, const mpitag &tag,
      const mpiid &source)
  {
    if (dbg_.is_active())
      dbg_ << "mpicollective[" << comm_.rank().id << ", start_recv from " << source.to_string() << "\n";
    //       << "]::start_recv(" << count << ", " << type << ", " << tag << ", "
    //       << source << ")\n";
    mpirequest_t req = mpirequest::construct(key::construct());
    eventhandler::ptr rcv = recver::construct(mpicollective::ptr(this), count, type,
        tag, source);

    queue_->recv(req, count, type, source, tag, comm_, mpimessage::collective,
        rcv);
  }

  //
  // Request a funneled copy.
  //
  void
  mpicollective::start_funneled_copy(int count, const mpitype &type,
      const mpitag &tag, const mpiid &partner)
  {
    /** DEPRECATED JJW 01/15/2012 
        Using funneled copies creates a mass memory leak
    */
    throw unimplementederror(
        "mpi collective funneled copy is a deprecated feature and "
        "requires a refactoring before being used."
    );

    if (funnel_ == 0)
      {
        throw nullerror(
            "mpicollective::start_funneled_copy: Null funnel object. "
              "Did you forget to call mpicollective::start?");
      }
    funnel_->copy(count, type, tag, partner);
  }

  //
  // Request an immediate copy.
  //
  void
  mpicollective::start_immediate_copy(int count, const mpitype &type,
      const mpitag &tag, const mpiid &partner)
  {
    // eventhandler::ptr copier = immediatecopy::construct(
    //     this->get_self(), count, type, tag, partner);
    //  node_->copy(node_->now(), count * type.size, copier);
  }

  //
  // Goodbye.
  //
  mpicollective::~mpicollective() throw ()
  {
  }

  //
  // Get busy child.
  //
  void
  mpicollective::start()
  {
    SSTMAC_DEBUG << "mpicollective[" << comm_.rank().id << "]::start \n";
    //       << my_node << ")\n";
    //  if (self_.lock() == 0)
    //   {
    //     throw nullerror("mpicollective::start:  Null self pointer.  "
    //       "Did you forget to call set_self()?");
    //   }

    /** DEPRECATED JJW 01/15/2012
        The funneled copy currently never gets used and it is 
        creating a circular reference the keeps the collective
        from getting deleted.  It should not be up to us
        to keep the funnel alive.  It should be up to the 
        pending messages for the funnel to keep it alive

        funnel_ = funneledcopy::construct(mpicollective::ptr(this));
    */

  }

  void
  mpicollective::complete(const payload::const_ptr& content)
  {
    SSTMAC_DEBUG << "Completing mpicollective on rank " << comm_.rank().id << " \n";
    status_->set_content(content);
    mpimessage::ptr fake = mpimessage::construct(content);

    req_->complete(fake);
    completion_->handle(fake);
  }

  const payload::const_ptr
  mpicollective::combine_content(const payload::const_ptr &p1,
      const payload::const_ptr &p2, mpiop &op)
  {
    payload::const_ptr ret;

    if (op.label == "sum")
      {
        ret = p1->add(p2);
      }
    else if (op.label == "max")
      {
        ret = (p1->less_than(p2)) ? p2 : p1;
      }
    else if (op.label == "min")
      {
        ret = (p1->greater_than(p2)) ? p2 : p1;
      }
    else
      {
        std::cout << "mpi op name: " << op.label << std::endl;
        throw valueerror("mpicollective: mpiop not supported yet.");
      }

    return ret;
  }

} } // end of namespace sstmac
