/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPIAPI_PERSISTENT_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPIAPI_PERSISTENT_H_INCLUDED

#include <sstmac/software/libraries/mpi/mpiapi.h>
#include <sstmac/software/libraries/mpi/mpirequest.h>

namespace sstmac { namespace sw {

  /**
   * Base for persistent operations.
   */
  class mpiapi::persistent : public mpirequest {
  public:
    persistent();
    virtual ~persistent() throw();
    virtual void start() = 0;

    typedef boost::intrusive_ptr<persistent> ptr;
  };

  /**
   * Persistent send operations (send, bsend, rsend, ssend)
   */
  class mpiapi::persistentsend : public mpiapi::persistent {
  public:
    /// The function type used to decide what kind of send we do.
    typedef timestamp (mpiapi::*sendfunc_t)
    (int count, const mpitype &type, const mpiid &target,
     const mpitag &tag, const mpicomm &comm, const mpirequest_t &req,
     const payload::const_ptr& content);

  private:
    /// The parent object.
    mpiapi::ptr parent_;
    sendfunc_t func_;

    /// The arguments.
    int count_;
    mpitype type_;
    mpiid target_;
    mpitag tag_;
    mpicomm comm_;
    payload::const_ptr content_;

  protected:
    persistentsend(mpiapi::ptr api,
                   sendfunc_t func, int count, const mpitype &type,
                   const mpiid &target, const mpitag &tag, const mpicomm &comm,
                   const payload::const_ptr& content);




  public:
    typedef boost::intrusive_ptr<persistentsend> ptr;

    static persistentsend::ptr
    construct(mpiapi::ptr api,
              sendfunc_t func, int count, const mpitype &type,
              const mpiid &target, const mpitag &tag, const mpicomm &comm,
              const payload::const_ptr& content);

    virtual ~persistentsend() throw();

    /// Start this request.
    virtual void start();
  };

  /**
   * Persistent recv operations
   */
  class mpiapi::persistentrecv : public mpiapi::persistent {
  private:
    /// Parent object.
    mpiapi::ptr parent_;

    /// The arguments.
    int count_;
    mpitype type_;
    mpiid dest_;
    mpitag tag_;
    mpicomm comm_;

  protected:
    persistentrecv(mpiapi::ptr api, int count, const mpitype &type,
                   const mpiid &target, const mpitag &tag, const mpicomm &comm);



  public:
    typedef boost::intrusive_ptr<persistentrecv> ptr;

    static persistentrecv::ptr
    construct(mpiapi::ptr api, int count, const mpitype &type,
              const mpiid &target, const mpitag &tag, const mpicomm &comm);

    virtual ~persistentrecv() throw();

    /// Start this request.
    virtual void start();
  };

} } // end of namespace sstmac

#endif
