/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_SSTMPI_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_SSTMPI_H_INCLUDED

struct MPI_Status
{
  long MPI_SOURCE;
  long MPI_TAG;
  long MPI_ERROR;
  long BYTES_RECV;
};

typedef struct MPI_Status MPI_Status;

typedef long MPI_Request;
typedef long MPI_Datatype;
typedef long MPI_Comm;
typedef long MPI_Op;
typedef long MPI_Group;

/** Define the error codes for use as return values */
#define   MPI_SUCCESS         0   
#define   MPI_ERR_BUFFER      1  // Invalid buffer pointer
#define   MPI_ERR_COUNT       2  // Invalid count argument
#define   MPI_ERR_TYPE        3  // Invalid datatype argument
#define   MPI_ERR_TAG         4  // Invalid tag argument
#define   MPI_ERR_COMM        5  // Invalid communicator
#define   MPI_ERR_RANK        6  // Invalid rank 
#define   MPI_ERR_REQUEST     7  // Invalid request 
#define   MPI_ERR_ROOT        8  // Invalid root 
#define   MPI_ERR_GROUP       9  // Invalid group 
#define   MPI_ERR_OP          10 // Invalid operation 
#define   MPI_ERR_TOPOLOGY    11 // Invalid topology 
#define   MPI_ERR_DIMS        12 // Invalid dimensions argument 
#define   MPI_ERR_ARG         13 // Invalid argument of some other kind 
#define   MPI_ERR_UNKNOWN     14 // Unknown error 
#define   MPI_ERR_TRUNCATE    15 // Message truncated on receive
#define   MPI_ERR_OTHER       16 // Known error not in this list   
#define   MPI_ERR_INTERN      17 // Internal MPI error   
#define   MPI_ERR_IN_STATUS   18 // Error code is in status  
#define   MPI_ERR_PENDING     19 // Pending request 
#define   MPI_ERR_LASTCODE    20 // Last error code

/** Define a null payload */
#define MPI_PAYLOAD_IGNORE NULL

#define MPI_

//ops
enum _mpi_ops_
{
  MPI_MAX = 1,
  MPI_MIN,
  MPI_SUM,
  MPI_PROD,
  MPI_LAND,
  MPI_BAND,
  MPI_LOR,
  MPI_BOR,
  MPI_LXOR,
  MPI_BXOR,
  MPI_MINLOC,
  MPI_MAXLOC
};

// useful flags
#define MPI_ANY_SOURCE -1
#define MPI_ANY_TAG -1

// datatypes
enum _mpi_datatypes_
{
  MPI_CHAR = 1,
  MPI_BYTE,
  MPI_SHORT,
  MPI_INT,
  MPI_LONG,
  MPI_FLOAT,
  MPI_DOUBLE,
  MPI_UNSIGNED_CHAR,
  MPI_UNSIGNED_SHORT,
  MPI_UNSIGNED,
  MPI_UNSIGNED_LONG,
  MPI_LONG_DOUBLE,
  MPI_LONG_LONG_INT
};

#define MPI_COMM_WORLD MPI_Comm_World()

#ifdef __cplusplus
extern "C"
{
#endif

  int
  MPI_Init(int *, char ***);

  int
  MPI_Finalize(void);

  int
  MPI_Wait(MPI_Request *request, struct MPI_Status *status);
  int
  MPI_Waitany(int count, MPI_Request array_of_requests[], int *index,
      struct MPI_Status *status);
  int
  MPI_Waitsome(int incount, MPI_Request array_of_requests[], int *outcount,
      int array_of_indices[], struct MPI_Status array_of_statuses[]);

  int
  MPI_Get_count(struct MPI_Status *status, MPI_Datatype datatype, int *count);

  int
  MPI_Test(MPI_Request *request, int *flag, struct MPI_Status *status);

  /**
   * MPI_Cancel also frees the request, and it should not be used again.
   * @param request
   * @return
   */
  int
  MPI_Cancel(MPI_Request *request);

  int
  MPI_Request_free(MPI_Request *request);

  int
  MPI_Comm_dup(MPI_Comm comm, MPI_Comm *newcomm);
  int
  MPI_Comm_rank(MPI_Comm comm, int *rank);
  int
  MPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm *newcomm);
  int
  MPI_Comm_size(MPI_Comm comm, int *size);
  int
  MPI_Comm_free(MPI_Comm *comm);
  int
  MPI_Comm_create(MPI_Comm comm, MPI_Group group, MPI_Comm *newcomm);
  int
  MPI_Comm_group(MPI_Comm comm, MPI_Group *group);

  int
  MPI_Group_incl(MPI_Group group, int n, int *ranks, MPI_Group *newgroup);
  int
  MPI_Group_free(MPI_Group *group);

  int
  MPI_Send(void *buf, int count, MPI_Datatype datatype, int dest, int tag,
      MPI_Comm comm);

  int
  MPI_SendRecv(void *sendbuf, int sendcount, MPI_Datatype sendtype, int dest,
      int sendtag, void *recvbuf, int recvcount, MPI_Datatype recvtype,
      int source, int recvtag, MPI_Comm comm, struct MPI_Status* stat);

  int
  MPI_Isend(const void *buf, int count, MPI_Datatype datatype, int dest,
      int tag, MPI_Comm comm, MPI_Request *request);

  int
  MPI_Recv(void *buf, int count, MPI_Datatype datatype, int source, int tag,
      MPI_Comm comm, struct MPI_Status *status);

  /**
   * Irecv  -- MPI_Recover_Payload must be used to set the payload buffer
   * @param buf Not used, pass NULL
   */
  int
  MPI_Irecv(void *buf, int count, MPI_Datatype datatype, int source, int tag,
      MPI_Comm comm, MPI_Request *request);

  /**
   * --- MPI_Recover_Payload - used to recover the payload on an Irecv
   * @param buf
   * @param request
   * @return 0 if successful, 1 if request is not complete yet.
   */
  int
  MPI_Recover_Payload(void *buf, MPI_Request *request);

  int
  MPI_Allreduce(const void *sendbuf, void *recvbuf, int count,
      MPI_Datatype datatype, MPI_Op op, MPI_Comm comm);

  int
  MPI_Reduce(const void *sendbuf, void *recvbuf, int count,
      MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm);

  int
  MPI_Barrier(MPI_Comm comm);

  int
  MPI_Bcast(void *buffer, int count, MPI_Datatype datatype, int root,
      MPI_Comm comm);

  int
  MPI_Gather(void *sendbuf, int sendcnt, MPI_Datatype sendtype, void *recvbuf,
      int recvcnt, MPI_Datatype recvtype, int root, MPI_Comm comm);

  int
  MPI_Allgather(void *sendbuf, int sendcount, MPI_Datatype sendtype,
      void *recvbuf, int recvcount, MPI_Datatype recvtype, MPI_Comm comm);

  int
  MPI_Scatter(void *sendbuf, int sendcnt, MPI_Datatype sendtype, void *recvbuf,
      int recvcnt, MPI_Datatype recvtype, int root, MPI_Comm comm);

  int
  MPI_Alltoall(void *sendbuf, int sendcount, MPI_Datatype sendtype,
      void *recvbuf, int recvcount, MPI_Datatype recvtype, MPI_Comm comm);

  MPI_Comm
  MPI_Comm_World();

  double
  MPI_Wtime();

  void
  MPI_Disable_Payloads();
  void
  MPI_Enable_Payloads();

#ifdef __cplusplus
}
#endif

#endif
