/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <stdlib.h>

#include <sstmac/common/sstmac-config.h>

#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/software/process/app.h>
#include <sstmac/software/process/key.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/common/messages/sst_message.h>

#include <sstmac/software/threading/threading_pth.h>
#include <sstmac/software/threading/threading_pthread.h>
#include <sstmac/software/threading/threading_ucontext.h>

#include <sstmac/software/services/service.h>

#include <sstream>

namespace sstmac
{
  namespace sw
  {

    thread::ptr OS_API::current_thread_ = NULL;

    stackalloc operatingsystem::stackalloc_(1 << 20, 1 << 24, false);

    std::list<thread::ptr> operatingsystem::to_delete_;

    debug operatingsystem::dbg_;

    operatingsystem::operatingsystem(hardware_descriptor::ptr &hw)
    {

      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> os | operatingsystem");
          logger::register_logger(&dbg_);
        }

      self_ = ptr(this);

      restarting_ = false;

      const char *threading_pchar = getenv("SSTMAC_THREADING");
      if (threading_pchar)
        {
          // Different variants of getenv return different strings.
          // Adjust the threading_string to point to just the value.
          if (!strncmp("SSTMAC_THREADING=", threading_pchar, 17))
            {
              threading_pchar = strchr(threading_pchar, '=');
            }
          if (*threading_pchar == '=')
            threading_pchar++;
          std::string threading_string(threading_pchar);
          if (threading_string == "pth")
            {
#       if defined(SSTMAC_HAVE_PTH)
              des_context_ = new threading_pth();
#       else
              throw valueerror("operatingsystem: SSTMAC_THREADING=pth is not supported");
#       endif
            }
          else if (threading_string == "pthread")
            {
#       if defined(SSTMAC_HAVE_PTHREAD)
              des_context_ = new threading_pthread();
#       else
              throw valueerror("operatingsystem: SSTMAC_THREADING=pthread is not supported");
#       endif
            }
          else if (threading_string == "ucontext")
            {
#       if defined(SSTMAC_HAVE_UCONTEXT)
              des_context_ = new threading_ucontext();
#       else
              throw valueerror(
                  "operatingsystem: SSTMAC_THREADING=ucontext is not supported");
#       endif
            }
          else
            {
              throw valueerror(
                  "operatingsystem: invalid value for SSTMAC_THREADING environmental variable");
            }
        }
      else
        {
#     if defined(SSTMAC_USE_PTH)
          des_context_ = new threading_pth();
#     elif defined(SSTMAC_USE_PTHREAD)
          des_context_ = new threading_pthread();
#     elif defined(SSTMAC_USE_UCONTEXT)
          des_context_ = new threading_ucontext();
#     else
#       error "no default threading library specified"
#     endif
        }

      des_context_->init_context();
      //current_ = &caller_context_;

      threadstack_.push(thread_data_t(des_context_, NULL));

    }

    operatingsystem::~operatingsystem()
    {

    }

    //
    // Allocate an object of this type.
    //
    operatingsystem::ptr
    operatingsystem::construct(hardware_descriptor::ptr &hw)
    {
      operatingsystem::ptr retval(new operatingsystem(hw));

      return retval;
    }

    void
    operatingsystem::execute_kernel(AMI::AMI_CAT cat, int kernelName,
        sst_message::ptr data)
    {
      AMI* ami = getAMI(cat);

      if (ami == NULL)
        {
          std::stringstream ss;
          ss << "operatingsystem - can't find AMI " << cat;
          throw valueerror(ss.str());
        }
      ami->execute_kernel(kernelName, data);
    }

    ptr_type::ptr
    operatingsystem::execute_kernel(AMI::AMI_CAT cat, int kernelName)
    {
      AMI* ami = getAMI(cat);

      if (ami == NULL)
        {
          std::stringstream ss;
          ss << "operatingsystem - can't find AMI " << cat;
          throw valueerror(ss.str());
        }
      return ami->execute_kernel(kernelName);
    }

    bool
    operatingsystem::kernel_supported(AMI::AMI_CAT cat, int kernelName)
    {
      AMI* ami = getAMI(cat);
      if (ami == NULL)
        {
          return false;
        }
      return ami->kernel_supported(kernelName);
    }

    AMI*
    operatingsystem::getAMI(AMI::AMI_CAT cat)
    {
      if (amis_.find(cat) == amis_.end())
        {
          return NULL;
        }

      return amis_[cat];
    }

    void
    operatingsystem::setAMI(AMI::AMI_CAT cat, AMI* a)
    {
      amis_[cat] = a;
    }

    // ------- THREADING functions ----------

    void
    operatingsystem::switch_to_thread(thread_data_t tothread)
    {
      SSTMAC_DEBUG << "os::switch_thread : begin " << tothread.first << "\n";
      //threadcontext_t *saved_current = current_;
      //threadcontext_t *saved_caller = &caller_context_;
      // current_ = tothread;
      // caller_context_ = *saved_current;
      thread_data_t from = threadstack_.top();
      threadstack_.push(tothread);

      SSTMAC_DEBUG << "os: this: " << (long) this << "\n";
      SSTMAC_DEBUG << "os: node address: " << get_my_addr()->to_string()
            << "\n";
      SSTMAC_DEBUG << "os: size of threadstack: " << threadstack_.size()
            << "\n";

      current_thread_ = tothread.second;
      from.first->swap_context(tothread.first);
      SSTMAC_DEBUG << "os::switch_thread : back " << tothread.first << "\n";
      //current_ = saved_current;
      //caller_context_ = *saved_caller;

    }

    void
    operatingsystem::cleanup()
    {
      std::list<thread::ptr>::iterator it, end = to_delete_.end();
      std::list<thread::ptr>::iterator it2, end2 = threads_.end();
      for (it = to_delete_.begin(); it != end; it++)
        {
          (*it)->kill();
          for (it2 = threads_.begin(); it2 != end2; it2++)
            {
              if (*it2 == *it)
                {
                  threads_.erase(it2);
                  break;
                }
            }
        }

      to_delete_.clear();
    }

    timestamp
    operatingsystem::block(const key::ptr &req)
    {

      if (req == 0)
        {
          throw nullerror("operatingsystem::block:  Refusing to block "
            "pending a null key pointer.");
        }
      // OK, we're ready to roll.
      //blocked_.insert(std::make_pair(req, current_));
      if (threadstack_.size() == 1)
        {
          throw ssterror(
              "OS: trying to block the DES thread (bottom of the stack)");
        }
      thread_data_t top = threadstack_.top();
      req->blocking_.insert(top);
      // blocked_count_++;
      SSTMAC_DEBUG << "operatingsystem::block: "
            << "Blocking context with req->blocking_.size() == "
            << req->blocking_.size() << "\n";
      //<< ", leaving " << active_.size() << " contexts pending.\n";
      // swap_context(*current_, caller_context_);

      threadstack_.pop();
      current_thread_ = threadstack_.top().second;
      top.first->swap_context(threadstack_.top().first);

      SSTMAC_DEBUG << "os: this: " << (long) this << "\n";
      SSTMAC_DEBUG << "os: node address: " << get_my_addr()->to_string()
            << "\n";
      SSTMAC_DEBUG << "operatingsystem::block: " << "Done Blocking context "
            << top.first << ", size of threadstack: " << threadstack_.size()
            << "\n";

      return now();
    }

    timestamp
    operatingsystem::unblock(const key::ptr &req)
    {
      if (dbg_.is_active())
        dbg_ << "operatingsystem::unblock: " << " unblocking "
            << req->blocking_.size() << " threads \n";

      key::blocking_t::iterator it, end = req->blocking_.end();
      key::blocking_t templist;

      //save to a temp list first, so we can clear the key before we start switching to threads
      for (it = req->blocking_.begin(); it != end; ++it)
        {
          templist.insert(*it);
        }

      req->blocking_.clear();

      for (it = templist.begin(); it != templist.end(); ++it)
        {
          switch_to_thread((*it));
        }

      return now();
    }

    void
    operatingsystem::start_thread(const thread::ptr &t)
    {
      add_thread(t);

      /*dbg_ << "os::start_thread : begin " << t.get() << "\n";
       threadcontext_t *saved_current = current_;
       threadcontext_t *saved_caller = &caller_context_;
       current_ = &(t->context_);
       caller_context_ = *saved_current;
       swap_context(caller_context_, *current_);

       dbg_ << "os::start_thread : back " << t.get() << "\n";
       current_ = saved_current;
       caller_context_ = *saved_caller;*/
      switch_to_thread(thread_data_t(t->context_, t));

    }
    /*  void
     operatingsystem::join_thread(const thread::ptr &t,
     threadcontext_t con)
     {
     // scheduler_.join_thread(t, con);
     }*/

    void
    operatingsystem::complete_thread()
    {

      thread_data_t top = threadstack_.top();

      //SSTMAC_DEBUG << "killing thread: " << t << "\n";
      SSTMAC_DEBUG << "os::complete: " << "Context " << (long) top.first
            << "\n";
      //  << ", is complete, now " << active_.size() << " contexts pending.\n";

      threadstack_.pop();
      current_thread_ = threadstack_.top().second;
      //top.second->kill();

      //top.second->kill();
      //stackalloc_.free(top.second->stack_);
      to_delete_.push_back(top.second);

      top.first->complete_context(threadstack_.top().first);
      // This thread is not permitted to start again.
      throw illformederror(
          "operatingsystem::complete:  Thread switched in again "
            "after calling complete and relinquishing stack.");
    }

    // ---------------------

    nodeaddress::ptr
    operatingsystem::get_my_addr()
    {
      if (amis_.find(AMI::AMI_QUERY) == amis_.end())
        {
          throw unimplementederror(
              "operatingsystem::get_my_addr can't find query ami");
        }

      ptr_type::ptr p = amis_[AMI::AMI_QUERY]->execute_kernel(
          AMI::AMI_QUERY_ADDRESS);
      nodeaddress::ptr casted = boost::dynamic_pointer_cast<nodeaddress>(p);
      if (!p)
        {
          throw nullerror(
              "operatingsystem::get_my_addr - error casting to nodeaddress");
        }

      return casted;
    }

    nodeaddress::ptr
    operatingsystem::get_task_addr(softwareid sid)
    {
      if (env_.find(sid.app_) == env_.end())
        {
          std::cout << "sid: " << sid.to_string() << std::endl;
          throw oserror(
              "operatingsystem::get_task_addr(..) - no environment entry for this app");
        }

      return env_[sid.app_]->get_node_address(sid.task_);
    }

    void
    operatingsystem::register_lib(const library::ptr &lib)
    {
      if (libs_.find(lib->get_lib_name()) != libs_.end())
        {
          std::cout << "lib name: " << lib->get_lib_name() << std::endl;
          std::cout << "node: " << get_my_addr()->to_string() << "\n";
          throw oserror(
              "operatingsystem: trying to register a lib that is already registered");
        }

      libs_[lib->get_lib_name()] = lib;
      lib->init_os(self_);
      lib->register_child_libs();
    }

    void
    operatingsystem::unregister_lib(const library::ptr &lib)
    {
      SSTMAC_DEBUG << "OS: trying to unregister lib " << lib->get_lib_name()
            << "\n";
      if (libs_.find(lib->get_lib_name()) == libs_.end())
        {
          /* std::cout << "lib name: " << lib->get_lib_name() << std::endl;
           std::cout << "but we do have: \n";
           std::map<std::string, library::ptr>::iterator it, end = libs_.end();
           for(it = libs_.begin(); it != end; it++){
           std::cout << it->first << ", ";
           }
           throw oserror(
           "operatingsystem: trying to unregister a lib that is not registered");*/
        }

      libs_.erase(lib->get_lib_name());
    }

    void
    operatingsystem::register_ser(const service::ptr &lib)
    {
      register_lib(lib);
      lib->init_full_os(self_);

    }

    library::ptr
    operatingsystem::get_lib(std::string name)
    {
      if (libs_.find(name) == libs_.end())
        return library::ptr();
      else
        return boost::dynamic_pointer_cast<library>(libs_[name]);
    }

    timestamp
    operatingsystem::now()
    {
      return amis_[AMI::AMI_COMPUTE]->now();
    }

    void
    operatingsystem::add_thread(const thread::ptr &t)
    {
      t->init_thread(des_context_, stackalloc_.alloc(),
          stackalloc_.stacksize(), operatingsystem::ptr(this), NULL);

      threads_.push_back(t);
    }

    void
    operatingsystem::add_application(app::ptr &a,
        const sim_parameters::ptr &params)
    {

      a->consume_params(params);

      add_thread(a);
    }

    void
    operatingsystem::start_apps()
    {
      std::list<thread::ptr>::iterator it;

      for (it = threads_.begin(); it != threads_.end(); it++)
        {
          if (!(*it)->is_started())
            {
              // current_ = &((*it)->context_);
              //swap_context(caller_context_, (*it)->context_);
              switch_to_thread(thread_data_t((*it)->context_, *it));
            }
        }

      //check pending messages
      int psize = pending_messages_.size();
      for (int i = 0; i < psize; i++)
        {
          sst_message::ptr msg = pending_messages_.front();
          pending_messages_.pop_front();
          handle_message(msg);
        }
    }

    void
    operatingsystem::handle_message(const sst_message::ptr &msg)
    {
      std::string libn = msg->get_lib_name();
      if (libs_.find(libn) == libs_.end())
        {
          SSTMAC_DEBUG << "OS[" << get_my_addr()->to_string()
                << "]: didn't find lib named " << libn << "\n";
          /* if (pending_messages_.size() > 10)
           std::cout << "WARNING: OS storing message for lib name: " << libn
           << ", size of pending_messages queue is "
           << (pending_messages_.size() + 1) << std::endl;

           pending_messages_.push_back(msg);
           std::stringstream sstr;
           sstr << "os: can't find library "
           << libn << " when handling a message";
           throw oserror(sstr.str());*/

          if (!restarting_)
            {
              SSTMAC_DEBUG << "OS[" << get_my_addr()->to_string()
                    << "]: queueing up \n";
              pending_messages_.push_back(msg);
            }
        }
      else
        {
          SSTMAC_DEBUG << "OS[" << get_my_addr()->to_string()
                << "]: delivering to lib " << libn << "\n";
          libs_[libn]->incoming_message(msg);
        }
    }

    void
    operatingsystem::apps_kill()
    {
      std::list<thread::ptr>::iterator it;

      if (dbg_.is_active())
        dbg_ << "OS: killing threads \n";
      for (it = threads_.begin(); it != threads_.end(); it++)
        {
          (*it)->unregister_all_libs();

          to_delete_.push_back(*it);
        }

    }

    void
    operatingsystem::kill_app(app::ptr app)
    {
      app->unregister_all_libs();
      to_delete_.push_back(app);
    }

    std::list<app::ptr>
    operatingsystem::get_app_ptrs(appid aid)
    {
      cleanup(); //just in case

      std::list<app::ptr> ret;
      std::list<thread::ptr>::iterator iter, end = threads_.end();
      for (iter = threads_.begin(); iter != end; iter++)
        {
          app::ptr a = boost::dynamic_pointer_cast<app>(*iter);
          if (a)
            {
              if (a->get_sid().app_ == aid)
                {
                  ret.push_back(a);
                }
            }
        }

      return ret;
    }

    app::ptr
    operatingsystem::get_app_ptr(softwareid sid)
    {
      cleanup(); //just in case

      app::ptr ret;
      std::list<thread::ptr>::iterator iter, end = threads_.end();
      for (iter = threads_.begin(); iter != end; iter++)
        {
          app::ptr a = boost::dynamic_pointer_cast<app>(*iter);
          if (a)
            {
              if (a->get_sid() == sid)
                {
                  return a;
                }
            }
        }

      return ret;
    }

  }
} //end of namespace sstmac
