/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/services/launch/allocation/dumpiallocation.h>
#include <sstmac/common/fileio.h>
#include <sstmac/common/config.h>

#include <sstmac/hardware/network/congestion/interconnect.h>

namespace sstmac
{

  sstmac::hw::topology::ptr construct_topology(sim_parameters::ptr sim_params);

  namespace sw
  {

    dumpiallocation::nodemap_t dumpiallocation::hostnamemap_;
    sstmac::hw::topology::ptr dumpiallocation::omnet_top_;
    std::map<long, std::string> dumpiallocation::nodenum_to_host_map_;

    dumpiallocation::dumpiallocation(std::string mapname,
        sstmac::hw::topology::ptr top) :
        allocationstrategy(0), mapfile_(mapname), top_(top)
    {

    }

    dumpiallocation::ptr
    dumpiallocation::construct(std::string mapname, sstmac::hw::topology::ptr top)
    {
      return ptr(new dumpiallocation(mapname, top));
    }

    dumpiallocation::ptr
    dumpiallocation::construct(sim_parameters::ptr params, sstmac::hw::topology::ptr top)
    {
      std::string mapname = params->get_param("launch_dumpi_mapname");
      if (!top) /** This must be omnet. Build a topology from the parameters */
      {
          if (!omnet_top_)
          {
              omnet_top_ = construct_topology(params);
          }
          return construct(mapname, omnet_top_);
      }
      else
      {
          return construct(mapname, top);
      }
    }

    void
    dumpiallocation::read_map_file(const std::string &here,
        const std::string &mapfile,
        std::map<std::string, std::vector<int> >&hostmap)
    {
      SSTMAC_DEBUG << "dumpiallocation: reading map file \n";
      hostmap.clear();
      std::ifstream in;
      SSTFileIO::open_file(in, mapfile);

      if (!in.is_open())
        {
          throw valueerror(here + "could not find node map file in current folder or configuration include path");
        }
      SSTMAC_DEBUG << "dumpiallocation: found map file \n";
      int nnode = -1;
      in >> nnode;
      if (nnode <= 0)
        throw valueerror(here + "bad nnode in node map file");
      int ncoor = -1;
      in >> ncoor;
      if (ncoor <= 0)
        throw valueerror(here + "bad ncoor in node map file");

      SSTMAC_DEBUG << "dumpiallocation: looks good to go \n";
      for (int i = 0; i < nnode; i++)
        {

          std::string hostname;
          in >> hostname;

          if (hostname.size() == 0)
            throw valueerror(here + "bad hostname in map file");
          std::vector<int> coor(ncoor);
          for (int j = 0; j < ncoor; j++)
            {
              coor[j] = -1;
              in >> coor[j];
              if (coor[j] < 0)
                throw valueerror(here + "bad coordinate in map file");
            }

          if (dbg_.is_active())
            {
              SSTMAC_DEBUG << "inserting " << hostname << " into map at coor";
              for (int j = 0; j < ncoor; j++)
                dbg_ << " " << coor[j] ;
              dbg_ << "\n";
            }
          hostmap[hostname] = coor;
        }
    }

    void
    dumpiallocation::allocate(int nnode_requested,
        std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> &allocation)
    {
      static std::string here("dumpiallocation::allocate: ");

      std::map<std::string, std::vector<int> > hostmap;
      read_map_file(here, mapfile_, hostmap);

      if(!top_){
          throw valueerror("dumpiallocation: needs a topology pointer to convert coordinates to addresses");
      }

      std::map<std::string, std::vector<int> >::iterator it, end = hostmap.end();

      for (it = hostmap.begin(); it != end; it++)
        {
          std::vector<int> coor = it->second;
          //std::copy(coor.begin(), coor.end(),
          //    std::ostream_iterator<int>(std::cout, " "));

          // find node index for this vertex
          int64_t nodeindex;
          sstmac::hw::vertex hostvert(coor);


          try
          {
              nodeindex = top_->node_to_index(hostvert);
          }
          catch (invalidkeyerror e)
          {
              std::cout << e.what() << "\n" << here
                  << "possible mismatch between network geometry and hostmap\n";
              throw;
          }
          nodeaddress::ptr nid = nodeaddress::convert_from_uniqueid(nodeindex);
          hostnamemap_[it->first] = nid;
          nodenum_to_host_map_[nid->unique_id()] = it->first;
          allocation.insert(nid);
        }

    }

  }
}
