/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/skeletons/test/mpi_testall.h>

#include <sstmac/common/messages/valuepayload.h>

#include <math.h>

namespace sstmac
{
  namespace sw
  {
    using namespace std;

    int mpi_testall::errors_ = 0;

    mpi_testall::~mpi_testall()
    {

    }

    void
    mpi_testall::consume_params(const sim_parameters::ptr &params)
    {
      if(params->has_param("testmpi_stop_at_errors"))
        stop_at_errors_ = params->get_int_param("testmpi_stop_at_errors");
      else
        stop_at_errors_ = false;
    }

    void
    mpi_testall::skeleton_main()
    {
      //std::cout << "starting appMPI_pingpong" << endl;



      timestamp start = mpi()->init();
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();
      //dbg << "App " << rank << " initialized" << "\n";

      if(size.id < 8){
          throw rangeerror("mpi_testall needs at least 8 nodes to test on");
      }

      //----------first, lets do a ping pong with payload
      test_sendrecv();

      // ---- test nonblocking recv
      test_asynch();

      // ---- alright, let's try a broadcast
      test_bcast();

      // --- try a reduce
      test_reduce();

      // ---- sync up with barrier
      test_barrier();

      // ---- test allreduce
      test_allreduce();

      // ------ test scatter
      test_scatter();

      // ------ test gather
      test_gather();

      //------ test communicator functions
      test_comms();

      // ------ test wait functions
      test_wait();

      // ------ test allgather
      test_allgather();

      // ------ test alltoall
      test_alltoall();

      // ------ test probing
      test_probe();

      // ------- test persistent
      test_persistent();



      // ------ test reduce_scatter
      // test_reducescatter();

      // ----- finalize
      mpi()->barrier(world);

      if (rank.id == 0 && errors_ == 0)
        {
          SSTMAC_DEBUG << "- Finished testing! test successful \n";
        }
      else if (rank.id == 0 && errors_ != 0)
        {
          SSTMAC_DEBUG << "- Finished testing! " << errors_
              << " ERRORS - test not successful \n";
        }

      mpi()->finalize();

    }

    void
    mpi_testall::test_sendrecv()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();

      SSTMAC_DEBUG << "testall[" << rank.id << "]-- Testing payload send/recv \n";

      int count = 100; //arbitrary

      mpitag tag(0);
      bool participant = true;
      if ((size.id % 2) && (rank.id + 1 >= size.id))
        {
          // This is the odd-node-out -- communicating with no-one.
          participant = false;
        }
      if (participant)
        {
          mpiid buddy(rank.id ^ 1); // 0<=>1, 2<=>3, etc.

          if ((rank.id) & 1)
            { // even values of half-cycle plus rank.

              //dbg_ << "testall[" << rank.id << "]: sending payload \n";
              valuepayload<int>::const_ptr senddata = valuepayload<int>::construct(
                  rank.id * 1000);
              timestamp st = mpi()->send(count, mpitype::mpi_double, buddy, tag,
                  world, senddata);

            }
          else
            {
              // dbg_ << "testall[" << rank.id << "]: receiving payload \n";
              const_mpistatus_t stat;
              mpi()->recv(count, mpitype::mpi_double, buddy, tag, world, stat);

              const valuepayload<int>::const_ptr recvdata =
                  boost::dynamic_pointer_cast <const valuepayload<int>
              > (stat->content());

              if (!recvdata)
                {
                  errors_++;
                  SSTMAC_DEBUG << "ERROR at rank " << rank.id
                      << ": SEND-RECV: NULL payload recv'd \n";

                  if(stop_at_errors_) throw ssterror("an error occurred in the application");
                }
              else
                {
                  if (recvdata->data() != (buddy.id * 1000))
                    {
                      errors_++;
                      SSTMAC_DEBUG << "ERROR at rank " << rank.id
                          << ": SEND-RECV: received value " << recvdata->data()
                          << " and we should have got "
                          << buddy.id * 1000 << "\n";

                      if(stop_at_errors_) throw ssterror("an error occurred in the application");
                    }

                }

            }

        }
    }

    void
    mpi_testall::test_barrier()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing barrier " << " \n";

      if (rank.id % 2 == 0)
        {
          timestamp t(2e-3);
          compute(t);
        }

      mpi()->barrier(world);

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Through barrier " << " \n";
    }

    void
    mpi_testall::test_reduce()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();
      int count = 100; //arbitrary

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing reduce " << " \n";

      payload::const_ptr result;

      payload::const_ptr load = valuepayload<int>::construct((1 << rank.id));
      mpi()->reduce(count, mpitype::mpi_double, mpiop::sum, mpiid(0), world, load,
          result);

      if (rank.id == 0)
        {
          const valuepayload<int>::const_ptr recvdata = boost::dynamic_pointer_cast
              <const valuepayload<int> > (result);

          if (!recvdata)
            {
              errors_++;
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": REDUCE: NULL payload recv'd \n";
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }
          else
            {
              int expected = (int) pow(2.0, world.size().id) - 1;
              if (recvdata->data() != expected)
                {
                  errors_++;
                  SSTMAC_DEBUG << "ERROR at rank " << rank.id
                      << ": REDUCE: received value " << recvdata->data()
                      << " and we should have got " << expected << "\n";
                  if(stop_at_errors_) throw ssterror("an error occurred in the application");
                }

            }
        }
    }

    void
    mpi_testall::test_asynch()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();
      int count = 100; //arbitrary

      const_mpistatus_t stat;
      if (rank.id == 0)
        {
          SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing nonblocking recv/wait \n";
          mpiid r(1);
          mpitag t(1);
          mpi()->send(count, mpitype::mpi_double, r, t, world);

          SSTMAC_DEBUG << "testall[" << rank.id
              << "]---- This should come SECOND -- \n\n\n";
        }
      else if (rank.id == 1)
        {
          mpirequest_t req;
          mpiid r(0);
          mpitag t(1);
          mpi()->irecv(count, mpitype::mpi_double, r, t, world, req);
          SSTMAC_DEBUG << "testall[" << rank.id << "]---- This should come FIRST -- \n";
          mpi()->wait(req, stat);

          SSTMAC_DEBUG << "testall[" << rank.id << "]---- This should come THIRD -- \n";

        }
    }

    void
    mpi_testall::test_allreduce()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();
      int count = 100; //arbitrary

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing allreduce " << " \n";

      payload::const_ptr load = valuepayload<int>::construct((1 << rank.id));
      payload::const_ptr result;

      mpi()->allreduce(count, mpitype::mpi_double, mpiop::sum, world, load, result);

      const valuepayload<int>::const_ptr allrecvdata = boost::dynamic_pointer_cast
          <const valuepayload<int> > (result);

      if (!allrecvdata)
        {
          errors_++;
          SSTMAC_DEBUG << "ERROR at rank " << rank.id
              << ": ALLREDUCE: NULL payload recv'd \n";
          if(stop_at_errors_) throw ssterror("an error occurred in the application");
        }
      else
        {
          int expected = (int) pow(2.0, world.size().id) - 1;
          if (allrecvdata->data() != (expected))
            {
              errors_++;
              SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": ALLREDUCE: received value "
                  << allrecvdata->data() << " and we should have got " << expected
                  << "\n";
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }

        }
    }

    void
    mpi_testall::test_bcast()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();
      int count = 100; //arbitrary

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing payload broadcast \n";

      payload::const_ptr load;

      if (rank.id == 0)
        {
          load = valuepayload<int>::construct(1234);
        }
      mpiid root(0);
      mpi()->bcast(count, mpitype::mpi_double, root, world, load);

      const valuepayload<int>::const_ptr recvdata = boost::dynamic_pointer_cast
          <const valuepayload<int> > (load);

      if (!recvdata)
        {
          errors_++;
          SSTMAC_DEBUG << "ERROR at rank " << rank.id
              << ": NULL payload recv'd for BCAST \n";
          if(stop_at_errors_) throw ssterror("an error occurred in the application");
        }
      else
        {
          if (recvdata->data() != (1234))
            {
              errors_++;
              SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": BCAST: received value "
                  << recvdata->data() << " and we should have got " << 1234 << "\n";
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }

        }
    }

    void
    mpi_testall::test_scatter()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();
      int count = 100; //arbitrary

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing scatter " << " \n";

      std::vector<payload::const_ptr> vals;
      std::vector<int> vcounts;

      if (rank.id == 0)
        {
          for (int i = 0; i < size.id; i++)
            {
              int val = 1 << i;
              vals.push_back(valuepayload<int>::construct(val));
              vcounts.push_back(count);
            }
        }

      payload::const_ptr result;

      mpi()->scatter(count, mpitype::mpi_double, count, mpitype::mpi_double, mpiid(0),
          world, vals, result);

      const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast
          <const valuepayload<int> > (result);

      if (!scatterdata)
        {
          errors_++;
          SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": NULL payload recv'd for SCATTER \n";
          if(stop_at_errors_) throw ssterror("an error occurred in the application");
        }
      else
        {
          int expected = (int) (1 << rank.id);
          if (scatterdata->data() != (expected))
            {
              errors_++;
              SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": SCATTER: received value "
                  << scatterdata->data() << " and we should have got " << expected
                  << "\n";
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }
        }




      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing scatterv " << " \n";
      mpi()->scatterv(vcounts, mpitype::mpi_double, count, mpitype::mpi_double, mpiid(0),
          world, vals, result);

      const valuepayload<int>::const_ptr scatterdata2 = boost::dynamic_pointer_cast
          <const valuepayload<int> > (result);

      if (!scatterdata2)
        {
          errors_++;
          SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": NULL payload recv'd for SCATTERV \n";
          if(stop_at_errors_) throw ssterror("an error occurred in the application");
        }
      else
        {
          int expected = (int) (1 << rank.id);
          if (scatterdata2->data() != (expected))
            {
              errors_++;
              SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": SCATTERV: received value "
                  << scatterdata2->data() << " and we should have got " << expected
                  << "\n";
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }
        }

    }

    void
    mpi_testall::test_gather()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();
      int count = 100; //arbitrary

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing gather " << " \n";

      std::vector<payload::const_ptr> vals;

      payload::const_ptr load = valuepayload<int>::construct(1 << rank.id);

      mpi()->gather(count, mpitype::mpi_double, count, mpitype::mpi_double, mpiid(0),
          world, load, vals);

      if (rank.id == 0)
        {
          std::vector<payload::const_ptr>::iterator it, end = vals.end();

          if (vals.size() != world.size().id)
            {
              SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": GATHER: result vector size ("
                  << vals.size() << ") does not match world comm size (" << world.size().id
                  << ") \n";
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }

          int i = 0;
          for (it = vals.begin(); it != end; it++)
            {
              if (*it)
                {
                  int expected = (int) (1 << i);
                  const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast
                      <const valuepayload<int> > (*it);

                  if (scatterdata->data() != (expected))
                    {
                      errors_++;
                      SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": GATHER: received value "
                          << scatterdata->data() << " and we should have got " << expected
                          << "\n";
                      if(stop_at_errors_) throw ssterror("an error occurred in the application");
                    }
                }
              else
                {
                  errors_++;
                  SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": NULL payload recv'd for GATHER \n";
                  if(stop_at_errors_) throw ssterror("an error occurred in the application");
                }

              i++;

            }

        }



      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing gatherv " << " \n";

      std::vector<payload::const_ptr> vals2;

      payload::const_ptr load2 = valuepayload<int>::construct(1 << rank.id);
      std::vector<int> vcounts;

      for(int i = 0; i < size.id; i++){
          vcounts.push_back(count);
      }

      mpi()->gatherv(count, mpitype::mpi_double, vcounts, mpitype::mpi_double, mpiid(0),
          world, load2, vals2);

      if (rank.id == 0)
        {
          std::vector<payload::const_ptr>::iterator it, end = vals2.end();

          if (vals2.size() != world.size().id)
            {
              SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": GATHERV: result vector size ("
                  << vals.size() << ") does not match world comm size (" << world.size().id
                  << ") \n";
            }

          int i = 0;
          for (it = vals2.begin(); it != end; it++)
            {
              if (*it)
                {
                  int expected = (int) (1 << i);
                  const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast
                      <const valuepayload<int> > (*it);

                  if (scatterdata->data() != (expected))
                    {
                      errors_++;
                      SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": GATHERV: received value "
                          << scatterdata->data() << " and we should have got " << expected
                          << "\n";
                      if(stop_at_errors_) throw ssterror("an error occurred in the application");
                    }
                }
              else
                {
                  errors_++;
                  SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": NULL payload recv'd for GATHERV \n";
                  if(stop_at_errors_) throw ssterror("an error occurred in the application");
                }

              i++;

            }

        }
    }

    void
    mpi_testall::test_allgather()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();
      int count = 100; //arbitrary

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing allgather " << " \n";

      std::vector<payload::const_ptr> vals;

      payload::const_ptr load = valuepayload<int>::construct(1 << rank.id);

      mpi()->allgather(count, mpitype::mpi_double, count, mpitype::mpi_double, world,
          load, vals);

      std::vector<payload::const_ptr>::iterator it, end = vals.end();

      if (vals.size() != world.size().id)
        {
          errors_++;
          SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": ALLGATHER: result vector size ("
              << vals.size() << ") does not match world comm size (" << world.size().id
              << ") \n";
          if(stop_at_errors_) throw ssterror("an error occurred in the application");
        }

      int i = 0;
      for (it = vals.begin(); it != end; it++)
        {
          if (*it)
            {
              int expected = (int) (1 << i);
              const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast
                  <const valuepayload<int> > (*it);

              if (scatterdata->data() != (expected))
                {
                  errors_++;
                  SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": ALLGATHER: received value "
                      << scatterdata->data() << " and we should have got " << expected << "\n";
                  if(stop_at_errors_) throw ssterror("an error occurred in the application");
                }
            }
          else
            {
              errors_++;
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": NULL payload recv'd for ALLGATHER \n";
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }

          i++;

        }


      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing allgatherv " << " \n";

      std::vector<payload::const_ptr> vals2;

      payload::const_ptr load2 = valuepayload<int>::construct(1 << rank.id);
      std::vector<int> vcounts;

      for(int i = 0; i < size.id; i++){
          vcounts.push_back(count);
      }

      mpi()->allgatherv(count, mpitype::mpi_double, vcounts, mpitype::mpi_double, world,
          load2, vals2);

      std::vector<payload::const_ptr>::iterator it2, end2 = vals2.end();

      if (vals2.size() != world.size().id)
        {
          errors_++;
          SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": ALLGATHERV: result vector size ("
              << vals.size() << ") does not match world comm size (" << world.size().id
              << ") \n";
          if(stop_at_errors_) throw ssterror("an error occurred in the application");
        }

      i = 0;
      for (it2 = vals2.begin(); it2 != end2; it2++)
        {
          if (*it2)
            {
              int expected = (int) (1 << i);
              const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast
                  <const valuepayload<int> > (*it2);

              if (scatterdata->data() != (expected))
                {
                  errors_++;
                  SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": ALLGATHERV: received value "
                      << scatterdata->data() << " and we should have got " << expected << "\n";
                  if(stop_at_errors_) throw ssterror("an error occurred in the application");
                }
            }
          else
            {
              errors_++;
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": NULL payload recv'd for ALLGATHERV \n";
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }

          i++;

        }

    }

    void
    mpi_testall::test_alltoall()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();
      int count = 100; //arbitrary

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing alltoall " << " \n";

      std::vector<payload::const_ptr> vals;

      for (int i = 0; i < size.id; i++)
        {
          int val = 1 << i;
          vals.push_back(valuepayload<int>::construct(val));
        }

      std::vector<payload::const_ptr> result;

      mpi()->alltoall(count, mpitype::mpi_double, count, mpitype::mpi_double, world,
          vals, result);

      std::vector<payload::const_ptr>::iterator it, end = result.end();

      if (result.size() != world.size().id)
        {
          errors_++;
          SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": ALLTOALL: result vector size ("
              << result.size() << ") does not match world comm size (" << world.size().id
              << ") \n";
          if(stop_at_errors_) throw ssterror("an error occurred in the application");
        }

      int i = 0;
      for (it = result.begin(); it != end; it++)
        {
          if (*it)
            {
              int expected = (int) (1 << rank.id);
              const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast
                  <const valuepayload<int> > (*it);

              if (scatterdata->data() != (expected))
                {
                  errors_++;
                  SSTMAC_DEBUG << "ERROR at rank " << rank.id << ": ALLTOALL: received value "
                      << scatterdata->data() << " and we should have got " << expected << "\n";
                  if(stop_at_errors_) throw ssterror("an error occurred in the application");
                }
            }
          else
            {
              errors_++;
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": NULL payload recv'd for ALLTOALL for rank " << i << " \n";
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }

          i++;

        }

    }

    void
    mpi_testall::test_comms()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing comm create \n";

      std::vector<mpiid> ids;
      std::vector<mpiid> idodd;
      for (int i = 0; i < size.id; i++)
        {
          if (i % 2 == 0)
            ids.push_back(mpiid(i));
          else
            idodd.push_back(mpiid(i));
        }

      mpicomm evens;
      mpi()->comm_create(world, ids, evens);


      if (rank.id % 2 == 0)
        {
          if (evens == mpicomm::comm_null)
            {
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": create_comm returned null comm when it should be a real thing \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }else{

                if(evens.rank().id != rank.id / 2){
                    errors_++;
                    SSTMAC_DEBUG << "testall[" << rank.id << "] -- my new comm rank is " << evens.rank().id << " and it should be " << (rank.id / 2) << "\n";
                    if(stop_at_errors_) throw ssterror("an error occurred in the application");
                }

                if(evens.id().id != 1){
                    SSTMAC_DEBUG << "ERROR at rank " << rank.id
                        << ": create produced id " << evens.id().id << " and it should have been " << 1 << " \n";
                    errors_++;
                    if(stop_at_errors_) throw ssterror("an error occurred in the application");
                }

            }

          SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing comm dup \n";

          mpicomm duped;
          mpi()->comm_dup(evens, duped);
          mpicommid shouldbe((rank.id % 2 == 0) ? evens.id().id + 1 : evens.id().id + 2);

          if(duped.id().id != shouldbe.id){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": create_dup produced id " << duped.id().id << " and it should have been " << shouldbe.id << " \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }

          if(duped.size() != evens.size()){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": create_dup produced comm with size " << duped.size().id << " and it should have been " << evens.size().id << " \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }

          if(duped.rank() != evens.rank()){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": create_dup produced comm where my rank is " << duped.rank().id << " and it should have been " << evens.rank().id << " \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }



          SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing comm split \n";

          mpicomm mynewsplit;
          int newkey = duped.rank().id;
          mpi()->comm_split(duped, ((rank.id == 2) ? 0 : 1), newkey, mynewsplit);
          int sizecheck  = (rank.id == 2) ? 1 : ((duped.size().id) - 1);

          if(mynewsplit == mpicomm::comm_null){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                                    << ": create_split returned null comm \n";
                                errors_++;
          }else{

              if(mynewsplit.size().id != sizecheck){
                  SSTMAC_DEBUG << "ERROR at rank " << rank.id
                      << ": create_split produced comm with size " << mynewsplit.size().id << " and it should have been " << sizecheck << " \n";
                  errors_++;
                  if(stop_at_errors_) throw ssterror("an error occurred in the application");
              }

          }


        }
      else
        {
          if (evens != mpicomm::comm_null)
            {
              SSTMAC_DEBUG
              << "ERROR at rank "
              << rank.id
              << ": create_comm returned a real comm when it should have returned the null comm \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
            }


        }


    }

    void
    mpi_testall::test_wait()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();

      int count = 100;
      mpitag tag(698);


      mpiid sender(5);

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing wait any \n";

      if(rank.id == 0){

          std::vector<mpirequest_t> reqs;

          for(int i = 1; i < size.id; i++){
              mpirequest_t req;
              mpi()->irecv(count, mpitype::mpi_double, mpiid(i), tag, world, req);
              reqs.push_back(req);
          }

          const_mpistatus_t stat;
          int index;
          mpi()->waitany(reqs, index, stat);

          if(index != sender.id - 1){  //subtract 1 because I didn't push a mpirequest in the vector for myself
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": waitany unblocked with index " << index << " and it should have been sent from " << sender.id << "\n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }

      }else if(rank == sender){
          mpi()->send(count, mpitype::mpi_double, mpiid(0), tag, world);
      }

      mpi()->barrier(world);


      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing wait some \n";

      mpitag tag2(699);

      if(rank.id == 0){
          sleep(timestamp(1));  //lag me, so the others have a chance to send

          std::vector<mpirequest_t> reqs;

          for(int i = 1; i < size.id; i++){
              mpirequest_t req;
              mpi()->irecv(count, mpitype::mpi_double, mpiid(i), tag2, world, req);
              reqs.push_back(req);
          }

          std::vector<const_mpistatus_t> stat;
          std::vector<int> index;
          mpi()->waitsome(reqs, index, stat);

          if(index.size() != (size.id / 2) - 1){  //subtract 1 for myself
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": waitsome unblocked with number of indeces " << index.size() << " and it should have been " << ((size.id / 2) - 1) << "\n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }

      }else if(rank.id % 2 == 0){
          mpi()->send(count, mpitype::mpi_double, mpiid(0), tag2, world);
      }

      mpi()->barrier(world);

    }



    void
    mpi_testall::test_reducescatter(){
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing reduce_scatter \n";

      std::vector<int> recvcounts;

      for(int i = 0; i < size.id; i++){
          recvcounts.push_back(10 * rank.id);
      }

      mpi()->reduce_scatter(recvcounts, mpitype::mpi_double, mpiop::sum, world);

      SSTMAC_DEBUG << "testall[" << rank.id << "]: reduce_scatter: recv counts: ";
      for(int i = 0; i < recvcounts.size(); i++){
          SSTMAC_DEBUG << recvcounts[i] << ", ";
      }

      SSTMAC_DEBUG << "\n";


    }


    void
    mpi_testall::test_probe()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();

      int count = 100;
      mpitag tag(713);
      mpitag tag2(714);

      mpi()->barrier(world);


      mpiid sender(5);

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing probe \n";

      if(rank.id == 0){

          std::vector<mpirequest_t> reqs;

          for(int i = 1; i < size.id; i++){
              mpirequest_t req;
              mpi()->irecv(count, mpitype::mpi_double, mpiid(i), tag, world, req);
              reqs.push_back(req);
          }

          const_mpistatus_t stat;
          int index;
          bool flag;
          std::vector<int> indices;
          std::vector<const_mpistatus_t> statuses;

          mpi()->test(reqs.at(sender.id), flag, stat);

          if(flag){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": test returned completed, and it probably should not have \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }

          mpi()->testsome(reqs, indices, statuses);

          if(indices.size() > 0){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": testsome returned completed, and it probably should not have \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }


          mpi()->testany(reqs, index, flag, stat);

          if(flag){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": testany returned completed, and it probably should not have \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }

          mpi()->iprobe(sender, tag, world, flag, stat);

          if(flag){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": iprobe returned completed, and it probably should not have \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }


          // -------  here is the blocking probe ----------- //
          mpi()->probe(sender, tag, world, stat);

          if(!stat){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": blocking probe did not return a valid mpistatus \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }


          mpi()->test(reqs.at(sender.id - 1), flag, stat);

          if(!flag){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": test returned NOT completed, and it should not have \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }

          mpi()->testsome(reqs, indices, statuses);

          if(indices.size() != 1){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": testsome returned indices with size " << indices.size() << ", and it should have been 1 \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }


          mpi()->testany(reqs, index, flag, stat);

          if(index != (sender.id - 1)){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": testany returned index " << index << ", and it should not have been " << (sender.id - 1) << " \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }

          sleep(timestamp(1));

          mpi()->iprobe(sender, tag2, world, flag, stat);

          if(!flag){
              SSTMAC_DEBUG << "ERROR at rank " << rank.id
                  << ": iprobe returned NOT completed, and it should not have \n";
              errors_++;
              if(stop_at_errors_) throw ssterror("an error occurred in the application");
          }

          mpi()->recv(count, mpitype::mpi_double, sender, tag2, world, stat);



      }else if(rank == sender){
          compute(timestamp(2));
          mpi()->send(count, mpitype::mpi_double, mpiid(0), tag, world);

          mpi()->send(count, mpitype::mpi_double, mpiid(0), tag2, world);
      }

      mpi()->barrier(world);



    }



    void
    mpi_testall::test_persistent()
    {
      mpicomm world = mpi()->comm_world();

      mpiid rank = world.rank();
      mpiid size = world.size();

      int count = 100;
      mpitag tag(715);

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Testing persistent actions \n";

      mpiid sender(6);
      mpiid recver(1);

      mpirequest_t asend;
      mpirequest_t arecv;

      if(rank == sender){
          mpi()->send_init(count, mpitype::mpi_double, recver, tag, world, asend);
      }

      if(rank == recver){
          mpi()->recv_init(count, mpitype::mpi_double, sender, tag, world, arecv);
      }

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Persistent: Haven't started them yet \n";


      if(rank == sender){
          mpi()->start(asend);
      }

      if(rank == recver){
          mpi()->start(arecv);
      }

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Persistent: OK, I started them \n";

      if(rank == sender){
          const_mpistatus_t stat;
         mpi()->wait(asend, stat);
      }

      if(rank == recver){
          const_mpistatus_t stat;
          mpi()->wait(arecv, stat);
      }

      SSTMAC_DEBUG << "testall[" << rank.id << "] -- Persistent: I'm through \n";

    }

  }
} //end of namespace sstmac
