/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_THREADING_STACKALLOC_H_INCLUDED
#define SSTMAC_SOFTWARE_THREADING_STACKALLOC_H_INCLUDED


#include <deque>
#include <vector>
#include <set>

namespace sstmac { namespace sw {

  /**
   * A management type to handle dividing mmap-ed memory for use
   * as ucontext stack(s).  This is basically a very simple malloc
   * which allocates uniform-size chunks (with the NX bit unset)
   * and sets guard pages on each side of the allocated stacks.
   *
   * This allocator does not return memory to the system until it is
   * deleted, but regions can be allocated and free-d repeatedly.
   */
  class stackalloc {
  private:
    /// The memory regions get allocated in chunks.
    class chunk;
    /// This is where we store the memory regions.
    typedef chunk* chunkptr_t;
    typedef std::vector<chunkptr_t> chunk_t;
    chunk_t chunk_;
    /// Each chunk is of this suggested size.
    size_t suggested_chunk_;
    /// Each stack request is of this size:
    size_t stacksize_;
    /// Do we want stacks separated by an mprot region?
    bool use_mprot_;
    /// This is our list of un-allocated chunks:
    typedef std::vector<void*> available_t;
    available_t available_;
    /// And this is our list of already-issued allocations.
    //typedef std::set<void*> allocated_t;
    //allocated_t allocated_;

  public:
    /// Build.
    stackalloc(size_t stacksize, size_t alloc_unit, bool use_mprot);

    /// Goodbye.
    virtual ~stackalloc();

    /// Access the stack size.
    size_t stacksize() const;

    /// Get a stack memory region.
    void* alloc();

    /// Return the given memory region.
    void free(void*);

    /// Clear -- return all memory (since any use of the stacks
    /// should be illegal from now on).
    void clear();
  };

} } // end of namespace sstmac

#endif
