/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_THREADING_THREADING_PTH_H_INCLUDED
#define SSTMAC_SOFTWARE_THREADING_THREADING_PTH_H_INCLUDED

#include <sstmac/software/threading/threading_interface.h>

#ifdef SSTMAC_HAVE_PTH
#  ifdef SSTMAC_HAVE_PTH_UCTX_SWITCH_IGNORE_SIGMASK
#    include <pth-sstmac.h>
#  else // SSTMAC_HAVE_PTH_UCTX_SWITCH_IGNORE_SIGMASK
#    include <pth.h>
#  endif // SSTMAC_HAVE_PTH_UCTX_SWITCH_IGNORE_SIGMASK
#endif

namespace sstmac { namespace sw {

#ifdef SSTMAC_HAVE_PTH

class threading_pth : public threading_interface
{
  typedef pth_uctx_t threadcontext_t;

private:
  threadcontext_t context_;

public:

  virtual threading_interface* copy(){
    return new threading_pth();
  }

  // Initializing a context.
  virtual void
  init_context()
  {
    const std::string here("init_context(pth_uctx_t &context):  ");
    if (pth_uctx_create(&context_) != TRUE)
      {
        throw oserror(here + strerror(errno));
      }
  }

  virtual void
  destroy_context()
  {
    const std::string here("destroy_context(pth_uctx_t&):  ");
    if (pth_uctx_destroy(context_) != TRUE)
      {
        throw oserror(here + strerror(errno));
      }
  }

  /// Start a new context.
  virtual void
  start_context(void *stack, size_t stacksize, void
  (*func)(void*), void *args, threading_interface *yield_to)
  {
    const std::string here("start_context(pth_uctx_t, ...):  ");
    if (stacksize < (16384))
      throw valueerror(here + "PTH does not accept stacks smaller than 16KB.");
    init_context();
    threading_pth* yield_pth = (threading_pth*)yield_to;
    int retval = pth_uctx_make(context_, (char*) stack, stacksize, NULL, func,
        args, (yield_to ? yield_pth->context_ : NULL));
    if (retval != TRUE)
      {
        throw oserror(here + strerror(errno));
      }
  }

  /// Swap context.
  virtual void
  swap_context(threading_interface *to)
  {
    static const std::string here("switch_context(ucontext_t&, ucontext_t&): ");

    threading_pth* topth = (threading_pth*)to;
    if (pth_uctx_switch(context_, topth->context_) != TRUE)
      {
        throw oserror(here + strerror(errno));
      }

  }

  /// This is called when we have completed running the thread. It is
  /// called in the from context.
  virtual void
  complete_context( threading_interface *to)
  {
    swap_context(to);
  }

};

#endif
} } // end of namespace sstmac
#endif
