#ifndef SDMA_ANALYSIS_NODE_CONFIG_H_
#define SDMA_ANALYSIS_NODE_CONFIG_H_

#include <sstmac/common/ptr_type.h>

#include <vector>
#include <sstmac/sstmacro.h>
#include <sstmac/common/sstmac-config.h>

#ifdef SSTMAC_HAVE_BOOST_1_37
#include <boost/unordered_map.hpp>
#else
#include <map>
#endif

namespace sdma
{

  class sdma_analysis_node_config : public sstmac::ptr_type
  {
  protected:
    static sstmac::debug dbg_;

#ifdef SSTMAC_HAVE_BOOST_1_37
    boost::unordered_map<sstmac::sw::taskid, sstmac::nodeaddress::ptr> taskmap_;
    boost::unordered_map<sstmac::nodeaddress::ptr, sstmac::nodeaddress::ptr>
    stages_;
#else
    std::map<sstmac::sw::taskid, sstmac::nodeaddress::ptr> taskmap_;
    std::map<sstmac::nodeaddress::ptr, sstmac::nodeaddress::ptr> stages_;
#endif

    sdma_analysis_node_config(const std::vector<int> &levs,
        const std::list<sstmac::nodeaddress::ptr> &nodes, long appsize)
    {

      if (!dbg_.is_real())
        {
          dbg_ = sstmac::debug("<debug> sdma | sdma_config",
              sstmac::logger::cyan);
          sstmac::logger::register_logger(&dbg_);
        }

      std::list<sstmac::nodeaddress::ptr>::const_iterator nodeiter =
          nodes.begin();

      std::set<sstmac::nodeaddress::ptr, sstmac::nodeaddress::nodeaddresscomp>
          prev;

      //first, build task map
      int appsPerNode = appsize / levs[levs.size() - 1];
      int appnodes = 1;
      int i;
      bool lastloop = false;
      for (i = 0; i < appsize; i++)
        {
          lastloop = false;
          prev.insert(*nodeiter);
          sstmac::sw::taskid tid(i);
          taskmap_[tid] = *nodeiter;

          SSTMAC_DEBUG << "sdma_config: task " << tid << " is using node "
              << (*nodeiter)->to_string() << "\n";

          if (i % appsPerNode == appsPerNode - 1)
            {
              appnodes++;
              nodeiter++;

              lastloop = true;
            }
        }

      if (lastloop)
        {
          nodeiter--;
        }

      //nodeiter = nodes.begin();
      nodeiter++;

      for (int j = levs.size() - 2; j >= 0; j--)
        {
          SSTMAC_DEBUG << "sdma_config: setting up stage " << j << "\n";
          int n = levs[j];
          int per =  prev.size() / n;
          int ind = 0;
          std::set<sstmac::nodeaddress::ptr,
              sstmac::nodeaddress::nodeaddresscomp> prevtemp = prev;
          prev.clear();
          std::set<sstmac::nodeaddress::ptr,
              sstmac::nodeaddress::nodeaddresscomp>::iterator iter, end =
              prevtemp.end();

          int k = 0;
          for (iter = prevtemp.begin(); iter != end; iter++)
            {
              prev.insert(*nodeiter);
              sstmac::nodeaddress::ptr from = *iter;
              sstmac::nodeaddress::ptr to = *nodeiter;
              SSTMAC_DEBUG << "sdma_config: setting stage["
                  << from->to_string() << "] to " << to->to_string() << "\n";
              stages_[from] = to;

              if (k % per == per - 1)
                {
                  nodeiter++;
                }

              k++;
            }
        }

    }

  public:
    typedef boost::intrusive_ptr<sdma_analysis_node_config> ptr;

    static ptr
    construct(const std::vector<int> &levs,
        const std::list<sstmac::nodeaddress::ptr> &nodes, long appsize)
    {
      return ptr(new sdma_analysis_node_config(levs, nodes, appsize));
    }

    virtual
    ~sdma_analysis_node_config()
    {
    }

    virtual std::string
    to_string() const
    {
      return "sdma_analysis_node_config";
    }

    /**
     * This function is called by the library used by the application
     * to find the first analysis node in the tree.
     */
    sstmac::nodeaddress::ptr
    get_my_analysis_node(sstmac::sw::taskid tid)
    {
      if (taskmap_.find(tid) == taskmap_.end())
        {
          std::cout << "task: " << tid.id << "\n";
          throw ssterror(
              "sdma_analysis_node_config::get_my_analysis_node -- couldn't find task");
        }
      return taskmap_[tid];
    }

    /**
     * This function is called by analysis nodes to find out if they should
     * send data along to another node
     */
    sstmac::nodeaddress::ptr
    get_next_stage(sstmac::nodeaddress::ptr node)
    {
      if (stages_.find(node) == stages_.end())
        {
          return sstmac::nodeaddress::ptr();
        }
      return stages_[node];
    }
  };

}

#endif
