#include "sdma_analysis_server.h"
#include "sdma_message.h"
#include <sstmac/sstmacro.h>
#include "sdma_lib.h"

using namespace sstmac;
using namespace sstmac::sw;

namespace sdma
{

  debug sdma_analysis_server::dbg_;

  void
  sdma_analysis_server::handle(const sst_message::ptr &msg)
  {
    if (!myaddr_)
      {
        myaddr_ = os_->get_my_addr();
      }

    if (msg == work_msg_)
      {
        work();
      }
    else if (msg->get_type() == sst_message::SDMA)
      {
        sdma_message::ptr smsg = boost::dynamic_pointer_cast<sdma_message>(msg);

        work_queue_.push(smsg);

        SSTMAC_DEBUG << "sdma_server[" << myaddr_->to_string()
              << "]: got a message to work on, work queue is of size "
              << work_queue_.size() << "\n";
        work();
      }
    else
      {
        std::cout << "type: " << msg->get_type() << "\n";
        throw ssterror("sdma_analysis_server::handle - got a non-SDMA message");
      }
  }

  /**
   * This gets called to handle messages from the work library
   * @param msg
   */
  void
  sdma_analysis_server::work_handler::handle(const sst_message::ptr &msg)
  {
    debug dbg_ = parent_->dbg_;
    SSTMAC_DEBUG << "sdma_analysis_server[" << parent_->myaddr_->to_string()
          << "]::work_handler \n";

    if (msg->get_type() == sst_message::COMPUTE)
      {

        sdma_message::ptr smsg = parent_->work_queue_.front();
        if (parent_->work_->doing_mem())
          {
            //the memory access is done, finish the computation
            parent_->work_->compute(smsg->get_work_bytes());
            return;
          }
        SSTMAC_DEBUG << "sdma_analysis_server["
              << parent_->myaddr_->to_string()
              << "]: work is done, schedule more \n";

        buildup_++;

        if (buildup_ == factor_)
          {
            buildup_ = 0;

            //now send along to the next stage, if necessary
            nodeaddress::ptr me = parent_->myaddr_;
            nodeaddress::ptr next = sdma_lib::config_->get_next_stage(me);
            if (next)
              {
                SSTMAC_DEBUG << "sdma_analysis_server[" << me->to_string()
                      << "]: sending along to " << next->to_string() << "\n";

                parent_->send_along(factor_ * smsg->get_work_bytes(), next);

              }
            else
              {
                SSTMAC_DEBUG << "sdma_analysis_server[" << me->to_string()
                      << "]: last in the chain, writing to disk \n";

                if (!parent_->resultfile_)
                  {
                    parent_->to_write_ += factor_ * smsg->get_work_bytes();
                    parent_->resultfile_ = parent_->pfs_->create("sdma-result");
                  }
                else
                  {
                    parent_->pfs_->write(parent_->resultfile_,
                        factor_ * smsg->get_work_bytes(), false); //don't wait for this one, pipelined write
                  }
              }
          }else{
              parent_->work_done();
          }

      }
  }

  /**
   * This gets called to handle messages from the file system library
   * @param msg
   */
  void
  sdma_analysis_server::pfs_handler::handle(const sst_message::ptr &msg)
  {
    debug dbg_ = parent_->dbg_;
    SSTMAC_DEBUG << "sdma_disk_server[" << parent_->myaddr_->to_string()
          << "]::work_handler \n";

    if (msg->get_type() == sst_message::FILEDATA)
      {
        file_metadata::ptr meta = boost::dynamic_pointer_cast<file_metadata>(
            msg);
        parent_->handle_pfs(meta);
      }
    else
      {
        throw ssterror(
            "sdma_disk_server::pfs_handler -- got a non-file system message");
      }
  }

  /**
   * This is made virtual so it can be overridden
   * @param msg
   */
  void
  sdma_analysis_server::handle_pfs(const file_metadata::ptr &meta)
  {

    SSTMAC_DEBUG << "sdma_analysis_server[" << myaddr_->to_string()
          << "]::handle_pfs \n";

    if (meta->op_ == file_metadata::CREATE_RESP) //this was to create the results file

      {
        SSTMAC_DEBUG << "sdma_disk_server[" << myaddr_->to_string()
              << "]: got a file creation resp, can now write \n";
        resultfile_ = pfs_->process_openfile_meta(meta);
        pfs_->write(resultfile_, to_write_, false);
      }

  }

  /**
   * This should be called to advance work
   */
  void
  sdma_analysis_server::work_done()
  {
    //we're done, pop the queue and schedule another work
    work_queue_.pop();
    working_ = false;

    os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, work_msg_);
  }

}
